package siva;

import java.io.File;
import java.util.ArrayList;
import java.util.Hashtable;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.SAXParser;
import org.sanskritlibrary.SAXParserHandler;
import org.sanskritlibrary.TextFormatter;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class CollectComments implements SAXParserHandler
{
    private Regex re = new Regex("", "^(f.\\s*\\d+[rv])");
    private ArrayList<String> comments = new ArrayList<String>();
    private Locator locator;
    private String filename;
    private String shortname;
    private String currentManuscript;
    private String currentElement;
    private int annotationNumber;
    private SAXParser saxParser;
    private StringBuilder buffer = new StringBuilder();
    private String currentFolio = null;

    public CollectComments () {
        saxParser = new SAXParser(this);
    }

    public void parse(String filename) throws Exception {
        this.filename = filename;
        annotationNumber = 0;
        currentManuscript = null;
        currentElement = null;
        currentFolio = "";
        buffer.setLength(0);
        shortname = new File(filename).getName();
        shortname = shortname.substring(0, shortname.indexOf("."));
        try {
            saxParser.parse(filename);
        } catch (Exception ex) {

        }
    }

    public ArrayList<String> getComments () {
        return comments;
    }

    public void warning(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void error(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void fatalError(SAXParseException exception) throws SAXException {
        System.out.format("%s:%d.%d %s\n", filename, locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        throw new SAXException("Boom!");
    }

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    public void startDocument() throws SAXException {
    }

    public void endDocument() throws SAXException {
    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    public void endPrefixMapping(String prefix) throws SAXException {
    }

    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        currentElement = qName;
        if (qName.equals("annotations")) {
            String page = atts.getValue("page");
            if (page == null) {
                System.out.format("%s missing annotations.page attribute\n", filename);
                currentManuscript = "?";
            } else {
                currentManuscript = page.substring(page.lastIndexOf("/")+1);
            }
        } else if (qName.equals("annotation")) {
            annotationNumber++;
        }
    }


    public void endElement(String uri, String localName, String qName) throws SAXException {
        if (qName.equals("annotation")) {
            String s = buffer.toString().toLowerCase();
            if (s.startsWith("f.")) {
                re.reset(s);
                if (re.find()) {
                    String ss = re.getGroup(1);
                    if (!ss.equals(currentFolio)) {
                        currentFolio = ss;
                        String comment = String.format("%s/%s/%s::%s",
                                currentManuscript, shortname, TextFormatter.intToString(annotationNumber, 4), ss);
                        comments.add(comment);
                    }
                }
            }
            buffer.setLength(0);
        }
    }

    public void characters(char[] ch, int start, int length) throws SAXException {
        if (currentElement.equals("annotation")) {
            String s = new String(ch, start, length);
            if (s.contains("&")) {
                s = fixup(s);
            }
            buffer.append(s.trim());
        }
    }

    private String fixup (String s) {
        return s;                           // write this if necessary
    }
}
