package siva;

import java.util.Hashtable;
import org.sanskritlibrary.SAXParser;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class CollectReferenceTargets implements SAXParserHandler
{
    private Locator locator;
    private String filename;
    private SAXParser saxParser;
    private Hashtable<String, String> targets = new Hashtable<String, String>();
    private Hashtable<String, Boolean> references;
    private String currentID;
    private String currentElement;
    private StringBuilder buffer;

    public CollectReferenceTargets (Hashtable<String, Boolean> references) {
        this.references = references;
        saxParser = new SAXParser(this);
    }

    public void parse(String filename) throws Exception {
        this.filename = filename;
        try {
            saxParser.parse(filename);
        } catch (Exception ex) {

        }
    }

    public Hashtable<String, String> getTargets () {
        return targets;
    }

    public void warning(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void error(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void fatalError(SAXParseException exception) throws SAXException {
        System.out.format("%s:%d.%d %s\n", filename, locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        throw new SAXException("Boom!");
    }

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    public void startDocument() throws SAXException {
    }

    public void endDocument() throws SAXException {
    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    public void endPrefixMapping(String prefix) throws SAXException {
    }

    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        String value = atts.getValue("xml:id");
        if (value != null) {
            if (references.get(value) != null) {
                if (currentElement == null) {
                    currentID = value;
                    currentElement = qName;
                    buffer = new StringBuilder();
                    copyElement(buffer, qName, atts);
                } else {
                    throw new SAXException("Nested xml:id");
                }
            }
        } else if (currentElement != null) {
            copyElement(buffer, qName, atts);
        }
    }

    private void copyElement(StringBuilder sb, String qName, Attributes atts) {
        sb.append(String.format("<%s", qName));
        int numatts = atts.getLength();
        for (int i=0; i<numatts; i++) {
            String value = atts.getValue(i);
            if (value.contains("'")) {
                sb.append(String.format(" %s=\"%s\"", atts.getQName(i), value));           // assume value does not also contain a "
            } else {
                sb.append(String.format(" %s='%s'", atts.getQName(i), value));           // assume value does not also contain a '
            }
        }
        sb.append(">");
    }

    public void endElement(String uri, String localName, String qName) throws SAXException {
        if (currentElement != null && qName.equals(currentElement)) {
            buffer.append(String.format("</%s>", currentElement));
            targets.put(currentID, buffer.toString());
            currentElement = null;
            buffer = null;
        }
    }

    public void characters(char[] ch, int start, int length) throws SAXException {
        if (buffer != null) {
            String s = new String(ch, start, length);
            if (s.contains("&")) {
                s = fixup(s);
            }
            buffer.append(s);
        }
    }

    private String fixup (String s) {
        return s;                           // write this if necessary
    }
}
