package siva;

import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.SAXParser;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class CompareVersions implements SAXParserHandler
{
    static BufferedWriter htmlFile;

    private String shortname;
    private StringBuilder buffer = new StringBuilder();
    private Locator locator;
    private SAXParser saxParser;
    private String elementName;
    private String elementPath;
    private int currentLevel;
    private boolean copying = false;

    static void setFile (BufferedWriter htmlFile) {
        CompareVersions.htmlFile = htmlFile;
    }

    public CompareVersions () {
        saxParser = new SAXParser(this);
    }

    public void setElementName (String elementName){
        this.elementName = elementName;
    }

    public void setElementPath (String elementPath) {
        this.elementPath = elementPath;
    }

    private void writeHtml(String col) {
        int lineno = 0;
        String[] lines = null;
        StringBuilder sb = new StringBuilder("<td valign='top'><pre>");
        StringBuilder sb02 = new StringBuilder();
        int adjustIndent = 0;
        String line02 =  null;
        try {
        col = col.replaceAll("&", "&amp;");
        col = col.replaceAll("<", "&lt;");
        col = col.replaceAll(">", "&gt;");
        lines = col.split("\n");
        sb = new StringBuilder("<td valign='top'><pre>");
        sb02 = new StringBuilder();
        adjustIndent = 0;
        for (String line : lines) {
            lineno++;
            line02 = line.replaceFirst("^\\s*", "");
            int indent = line.length() - line02.length();
            if (indent == 0) {
                adjustIndent = 0;
            } else if (adjustIndent == 0) {
                adjustIndent = indent;
            }

            int newIndent = 0;
            if (adjustIndent != 0) {
                newIndent = indent - adjustIndent + 1;
            }
            if (newIndent < 0) {
                newIndent = 0;
            }
            if (line02.length() < 90) {
                sb.append("                                                           ".substring(0, newIndent));
                sb.append(line02);
                sb.append("\n");
            } else {
                String[] words = line02.split("\\s+");
                for (int i=0; i<words.length; i++) {
                    if (sb02.length() + words[i].length() > 90) {
                        sb.append("                                                           ".substring(0, newIndent));
                        sb.append(sb02.toString());
                        sb.append("\n");
                        sb02.setLength(0);
                    }
                    sb02.append(words[i]);
                    sb02.append(" ");
                }
                if (sb02.length() > 0) {
                    sb.append("                                                           ".substring(0, newIndent));
                    sb.append(sb02.toString());
                    sb.append("\n");
                    sb02.setLength(0);
                }
            }
        }
        } catch (Exception ex) {
           System.out.format("%s, lineno=%d, %s\n", shortname, lineno, ex.getMessage());
        }
        sb.append("</pre></td>");
        try {
            htmlFile.write(sb.toString());
            htmlFile.write("\n");
        } catch (Exception ex) {
            System.out.println(sb.toString());
        }
    }

    public void parse(String filename) throws Exception {
        this.shortname = new File(filename).getName();
        try {
            saxParser.parse(filename);
        } catch (Exception ex) {
            // error should have been reported elsewhere
        }
    }

    private ArrayList<ElementInfo> stack = new ArrayList<ElementInfo>();
    private void pushElement (String qName, Attributes atts) {
        stack.add(new ElementInfo(qName, atts));
    }

    private ElementInfo popElement () {
        int top = stack.size() - 1;
        ElementInfo ei = stack.get(top);
        stack.remove(top);
        return ei;
    }

    private ElementInfo peekElement () {
        int top = stack.size() - 1;
        return stack.get(top);
    }
    
    private String getPath () {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (ElementInfo ei : stack) {
            sb.append(String.format("%s%s", delim, ei.qName));
            delim = ".";
        }
        return sb.toString();
    }

    public void warning(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void error(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void fatalError(SAXParseException exception) throws SAXException {
        String msg = String.format("%s:%d.%d %s\n", shortname, locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        System.out.println(msg);
        throw new SAXException("Boom!");
    }

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    public void startDocument() throws SAXException {
    }

    public void endDocument() throws SAXException {
        if (buffer.length() > 0) {
            writeHtml(buffer.toString());
        } else {
            writeHtml(String.format("No %s elements found.", elementName != null ? elementName : elementPath));
        }
    }


    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    public void endPrefixMapping(String prefix) throws SAXException {
    }


    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        pushElement(qName, atts);
        if (!copying) {
            if (elementName != null) {
                if (qName.equals(elementName)) {
                    copying = true;
                    currentLevel = stack.size();
                }
            } else {
                if (getPath().equals(elementPath)) {
                    copying = true;
                    currentLevel = stack.size();
                }
            }
        }

        if (copying) {
            buffer.append(String.format("<%s", qName));
            int numatts = atts.getLength();
            for (int i = 0; i < numatts; i++) {
                buffer.append(String.format(" %s='%s'", atts.getQName(i), atts.getValue(i)));
            }
            buffer.append(">");
        }
    }

    public void endElement(String uri, String localName, String qName) throws SAXException {
        if (copying) {
            buffer.append(String.format("</%s>", qName));
            boolean emptyBuffer = false;
            if (stack.size() == currentLevel) {
                if (elementName != null) {
                    emptyBuffer = qName.equals(elementName);
                } else {
                    emptyBuffer = getPath().equals(elementPath);
                }
            }

            if (emptyBuffer) {
                copying = false;
                ElementInfo ei = peekElement();
                Enumeration<String> en = ei.h.keys();
                boolean foundxmlid = false;
                while (en.hasMoreElements()) {
                    String attname = en.nextElement();
                    String attvalue = ei.h.get(attname);
                    String refkey = String.format("%s.%s", qName, attname);
                    Integer status = References.referenceAttributes.get(refkey);
                    if (status != null && status.intValue() == References.INPLACE) {
                        String[] fields = attvalue.split(" ");
                        for (String field : fields) {
                            buffer.append("\n+++++++++++++++++++++++\n");
                            String xmlidkey = String.format("%s%s", shortname, field);
                            buffer.append(References.xmlidContent.get(xmlidkey));
                        }
                        buffer.append("\n-----------------------\n");
                        foundxmlid = true;
                    }
                }
                if (!foundxmlid) {
                    buffer.append("\n\n");
                }
            }
        }
        popElement();
    }

    public void characters(char[] ch, int start, int length) throws SAXException {
        if (copying) {
            buffer.append(new String(ch, start, length));
        }
    }

    class Location {
        int row;
        int col;
        public Location(Locator locator) {
            this.row = locator.getLineNumber();
            this.col = locator.getColumnNumber();
        }

        public int getLineNumber () {
            return row;
        }

        public int getColumnNumber () {
            return col;
        }

    }

    class ElementInfo {
        public String qName;
        public Hashtable<String, String> h = new Hashtable<String, String>();

        public ElementInfo (String qName, Attributes atts) {
            this.qName = qName;
            int numatts = atts.getLength();
            for (int i=0; i<numatts; i++) {
                h.put(atts.getQName(i), atts.getValue(i));
            }
        }
    }
}
