/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package siva;

import java.util.ArrayList;
import java.util.Hashtable;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.XML;
import org.w3c.dom.Node;

public class FormatInfo {
    private static Regex formatMap = new Regex("", "\\s*map\\s*=\\s*['\"]([^'\"]+)");
    private static Regex formatRE = new Regex("", "(?:<value(.*?)>(.*?)</value>)|(?:<macro id='(.*?)'/>)");
    TestValueInfo test;
    ArrayList<FormatItem> items;

    FormatInfo(FormatInfoRaw fir) {
        items = new ArrayList<FormatItem>();
        formatRE.reset(fir.buffer);
        int n = 0;
        while (formatRE.find()) {
            int offset = formatRE.getOffset(0);
            if (offset > n) {
                items.add(new FormatItem(false, fir.buffer.substring(n, offset)));
            }
            String match = formatRE.getGroup(0);
            if (match.startsWith("<value")) {
                formatMap.reset(formatRE.getGroup(1));
                String map = null;
                if (formatMap.find()) {
                    map = formatMap.getGroup(1);
                    if (map.length() == 0) {
                        map = null;
                    }
                }
                String value = formatRE.getGroup(2);
                items.add(new FormatItem(true, value, map));
            } else if (match.startsWith("<macro")) {
                String id = formatRE.getGroup(3);
                items.add(new FormatItem(id));
            }
            n = offset + match.length();
        }
        if (n < fir.buffer.length()) {
            items.add(new FormatItem(false, fir.buffer.substring(n)));
        }

        if (fir.test != null) {
            test = new TestValueInfo(fir.test);
        }
    }

    static String format(PathInfo pi, Style style, ArrayList<FormatInfo> fis, Node node, String context) {
        StringBuilder sb = new StringBuilder();
        int len = fis.size();
        if (len > 0) {
            for (int i = 0; i < len; i++) {
                boolean matchFound = true;
                sb.setLength(0);
                FormatInfo fi = fis.get(i);
                if (fi.test != null && !fi.test.evaluate(node)) {
                    continue;
                }
                for (FormatItem item : fi.items) {
                    if (item.isMacro) {
                        MacroInfo mi = ParseLayout.getMacro(style, item.id);
                        sb.append(mi.execute(pi, style, node, context));
                    } else if (!item.evaluate) {
                        sb.append(item.value);
                    } else {            // must be a value
                        if (item.isAttribute) {
                            Node att = node.getAttributes().getNamedItem(item.value);
                            if (att == null) {
                                matchFound = false;
                            } else {
                                matchFound = append(pi, sb, att.getNodeValue(), item);
                            }
                        } else {
                            try {
                                Node child = XML.getRelativeNode(node, item.value);
                                if (child == null) {
                                    matchFound = false;
                                }
                                matchFound = append(pi, sb, child.getTextContent(), item);
                            } catch (Exception ex) {
                                sb.append("?");
                                System.out.println(ex.getMessage());
                            }
                        }
                        if (!matchFound) {
                            break;
                        }
                    }
                }
                if (matchFound) {
                    return sb.toString();
                }
            }
            return "";      // reach here means that no format could be applied
        }

        return "";      // Reach here if no formats were given. This is probably an error.
    }

    private static boolean append (PathInfo pi, StringBuilder sb, String text, FormatItem item) {
        boolean matchFound = true;
        MapInfo mi = null;
        if (item.map != null) {          // if we should attempt to map text to default encoding (i.e. SLP1 to roman)
            Hashtable<String, MapInfo> mis = ParseLayout.maps.get(item.map);
            mi = mis.get(text);     // should this text be mapped?
            if (mi !=  null) {      // yes
                matchFound = true;
                if (mi.capitalizeFirstLetter && mi.sourceEncoding.equals("SLP1")) {
                    text = String.format("{%c}%s", text.charAt(0), text.substring(1));
                }
                text = Main.transcode(text, mi.sourceEncoding);
            } else {
                mi = mis.get("?");
                matchFound = mi.identityMapForMissing;
                if (matchFound && mi.capitalizeFirstLetter) {
                    text = String.format("%c%s", Character.toUpperCase(text.charAt(0)), text.substring(1));
                }
            }
        }

        if (matchFound) {
            sb.append(text);
        }

        return matchFound;
    }
}

