package siva;

// This does nothing. It is a template for writing a class that parses XML

import java.io.File;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Hashtable;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
//DefaultHandler implements ContentHandler, DTDHandler, EntityResolver, ErrorHandler
//import org.xml.sax.helpers.DefaultHandler;

//DefaultHandler2 implements ContentHandler, DTDHandler, EntityResolver, ErrorHandler, DeclHandler, EntityResolver2, LexicalHandler
//http://docs.oracle.com/javase/1.5.0/docs/api/org/xml/sax/ext/DefaultHandler2.html
import org.xml.sax.ext.DefaultHandler2;         // this allows processing of comments because it implements LexicalHandler


// SAXParserHandler is an interface that forces us to include required methods.
// Serializable is optional. Have made local variables that don't need to be serialized transient.
public class FormatWithLayout extends DefaultHandler2 implements SAXParserHandler, Serializable {
    transient Locator locator;
    transient ArrayList<String> pathStack = new ArrayList<String>();
    transient Hashtable<String, StringBuilder> buffers = new Hashtable<String, StringBuilder>();
    private ParseBiblLayout layout;
    private String prefix;
    private String base;
    private boolean firstEntry = true;
    private Hashtable<String, String> pathmap;
    private Hashtable<String, String> lang = new Hashtable<String, String>();
    private StringBuilder sb = new StringBuilder();
    private ArrayList<ArrayList<String>> sequence; 


    public FormatWithLayout(String fileName, ParseBiblLayout layout) throws Exception
    {
        this.layout = layout;

        prefix = layout.getPrefix();
        base = layout.getBase();
        pathmap = layout.getPathMap();
        
        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();

        // Parse the input
        SAXParser saxParser = factory.newSAXParser();
        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
        // Uncomment the following to turn on comment processing.
        // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
        // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        try {
            saxParser.parse( new File(fileName), this);
        } catch (Exception ex) {
            throw ex;
        }
    }

    public String getHTML () {
        return sb.toString();
    }

	// Maintains a stack of element names (qName)
    private void push (String qName) {
        pathStack.add(qName);
    }

	// Pop a qName from the stack
    private String pop () {
        int last = pathStack.size()-1;
        String top = null;
        if (last >= 0) {
            top = pathStack.get(last);
            pathStack.remove(last);
        }

        return top;
    }

	// Peek at a qName without popping stack
    private String peek () {
        int last = pathStack.size()-1;
        String top = null;
        if (last >= 0) {
            top = pathStack.get(last);
        }

        return top;

    }

	// Serialize the qName stack. The string that is returned represent the path to the current qName
    private String buildPath () {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (String name : pathStack) {
            sb.append(String.format("%s%s", delim, name));
            delim = ".";
        }
        return sb.toString();
    }

    private void displayError(String msg) {
        System.out.format("Line %d: %s\n", locator.getLineNumber(), msg);
    }

    @Override
    public void warning(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void error(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    @Override
    public void startDocument() throws SAXException {
        sb.append("<html><head><title>Bibliography</title></head><body>\n");
    }

    @Override
    public void endDocument() throws SAXException {
        sb.append("</table></html>");
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException {
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        push(qName);
        String path = buildPath();
        if (path.equals(base)) {
            sb.append(String.format("<a name='%s'/>\n", atts.getValue("xml:id")));
            if (firstEntry) {
                firstEntry = false;
            } else {
                sb.append("<br/>");
            }
            sb.append("<table>\n");
        } else {
            int length = atts.getLength();
            for (int i=0; i<length; i++) {
                String name = atts.getQName(i);
                String key = String.format("%s@%s", path, name);
                String label = pathmap.get(key);
                if (label != null) {
                    sb.append(String.format("<tr><td>%s:</td><td>%s</td></tr>\n", label, atts.getValue(i)));
                }
                if (name.equals("xml:lang")) {
                    lang.put(path, atts.getValue(i));
                }
            }
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        String path = buildPath();
        StringBuilder buffer = buffers.get(path);
        String value = "";
        if (buffer != null) {
            value = buffer.toString().trim();
            buffer.setLength(0);
        }
        if (path.equals(base)) {
            sb.append("</table>\n");
        } else if (value.trim().length() > 0) {
            String label = pathmap.get(path);
            if (label != null) {
                if (lang.get(path) != null) {
                    value = Main.transcode(value, lang.get(path));
                    lang.remove(path);
                }
                sb.append(String.format("<tr><td>%s:</td><td>%s</td></tr>\n", label, value));
            }
        }

        pop();
    }

    @Override
    // Collects the innerText in a StringBuilder keyed by the current path. It is endElement's responsibility
    // to retrieve the StringBuilder and set its length to 0.
    // This is optional. It is not specified in SAXParserHandler
    public void characters(char[] ch, int start, int length) throws SAXException {
        String path = buildPath();
        StringBuilder sb = buffers.get(path);
        if (sb == null) {
            sb = new StringBuilder();
            buffers.put(path, sb);
        }
        sb.append(new String(ch, start, length));
    }

    @Override
    // Optional if you want to be notified of comments
    public void comment(char[] ch, int start, int length) throws SAXException {

    }

    class ValueInfo {
        String value;
        int idx;
    }

}
