package siva;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.Config;
import org.sanskritlibrary.FSM;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.ResourceInputStream;
import org.sanskritlibrary.TextFormatter;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;


public class Main implements ResourceInputStream
{
    static Hashtable<String, FSM[]> fsmh = new Hashtable<String, FSM[]>();
    static ResourceInputStream ris;
    static public String htmlpath = null;
    static public String textpath = null;
    static public String textSubdirectory = null;
    static public String xmlSubdirectory = "xml/";      // used by ParseTEI to detect references to XML in a TEI file
    static public String xmlpath = null;
    //static public String imagesFull = null;
    //static public String imagesCropped = null;
    static public String manuscriptServer = "../LoadImage";        // used to build links in generated HTML
    static public String htmlServer = "../LoadHTML";   // used to build links in generated HTML
    final static Hashtable<String, Boolean> ignore = new Hashtable<String, Boolean>();
    static Regex reXMLID = new Regex("", "xml:id='([^']+)'");

    // Do not process this xml files
    static {
        ignore.put("IndicSubjectClassification.xml", Boolean.TRUE);
        ignore.put("listbiblfile.xml", Boolean.TRUE);
        ignore.put("listpersonfile.xml", Boolean.TRUE);
        ignore.put("neh2008.xml", Boolean.TRUE);
    }

    Config config;

    // f=name (in XML directory)
    // p=path
    public static void main(String[] args) throws Exception
    {
        Main m = new Main();
        m.parse(args);
    }


    // Convenience method to convert -value- encoded as -lang- to ParseTEI.target (probably roman or deva)
    // Expect -lang- to be either "sa-Latn-x-SLP1" or "sa-Latn-x-KH"
    public static String transcode (String value, String lang) {
        try
        {
            String map = null;
            if (lang.endsWith("SLP1"))
                map = String.format("*:slp1->%s", ParseTEI.target);
            else if (lang.endsWith("KH"))
                map = String.format("*:hk->%s", ParseTEI.target);
            if (map != null)
            {
                TransformMap tm = new TransformMap(map, ris, fsmh);
                value = WebServices.transformString(value, tm);
                value = TextFormatter.charToEntity(value);
            }
        }
        catch (Exception ex)
        {
            System.out.format("TransformMap failed with %s\n", ex.getMessage());
        }

        return value;
    }

    @SuppressWarnings("ResultOfObjectAllocationIgnored")
    public void parse(String[] args) throws Exception {
        try {
            ris = this;
            if (args.length != 1) {
                printUsage();
                return;
            }

            new ParseLayout("layout.xml");
            ParseTEI.sivaErrorReport = IOAssistance.openWriter("sivaErrorReport.txt");


            boolean pathsRead = true;
            config = new Config("config.xml");

            xmlpath = config.catalogEntriesWithFacsimiles;
            pathsRead = pathsRead && fileExists(xmlpath);

            htmlpath = config.catalogEntriesHTML;
            pathsRead = pathsRead && fileExists(htmlpath);

            textpath = config.textsDirectory;
            pathsRead = pathsRead && fileExists(textpath);
/*
            imagesFull = config.imagesSitaAnnotated;
            pathsRead = pathsRead && fileExists(imagesFull);

            imagesCropped = config.croppedXMLRepaired;
            pathsRead = pathsRead && fileExists(imagesCropped);
*/
            if (!pathsRead) {
                System.out.println("Please fix errors and run again");
                return;
            }

            Regex re = new Regex(textpath, "([^\\\\/]+)[\\\\/]$");
            if (re.find()) {
                textSubdirectory = re.getGroup(1) + "/";        // this is the last component of the path, e.g. "texts/"
            } else {
                System.out.format("Illegal textpath: %s\n", textpath);
                return;
            }

            createReferenceHTML();  // create listbiblfile.html and listpersonfile.html

            // See what the command line says
            re = new Regex(args[0], "^([pfFhi]).*=.*?([^\\s]+)$");
            if (re.find()) {
                String command = re.getGroup(1);
                String arg = re.getGroup(2);
                if (command.equals("F")) {
                    xmlpath = "XML-fixedup/";
                    command = "f";
                }
                if (command.equals("f")) {
                    if (arg.equals("all")) {
                        formatPennCatalogValidate(null);
                    } else {
                        arg = arg.toLowerCase();
                        if (!arg.endsWith(".xml")) {
                            arg += ".xml";
                        }
                        formatPennCatalogValidate(xmlpath + arg);
                    }
                } else if (command.equals("p")) {
                    extractFromPennCatalog(arg);
                } else if (command.equals("i")) {
                    preprocessPennCatalog();
                } else if (command.equals("h")) {
                    displayHTML();
                }
            } else {
                printUsage();
            }

        } catch (Exception ex) {
            System.out.format("Failed with %s\n", ex.getMessage());
        } finally {
            if (ParseTEI.sivaErrorReport != null) {
                ParseTEI.sivaErrorReport.close();
            }
        }
    }

    private void createReferenceHTML () throws Exception {
        ParseBiblLayout pbf = new ParseBiblLayout("biblLayout.xml");
        FormatWithLayout fwl = new FormatWithLayout(String.format("%s/listbiblfile.xml", xmlpath), pbf);
        BufferedWriter out = IOAssistance.openWriter(String.format("%s/listbiblfile.html", htmlpath));
        out.write(fwl.getHTML());
        out.close();

        pbf = new ParseBiblLayout("personLayout.xml");
        fwl = new FormatWithLayout(String.format("%s/listpersonfile.xml", xmlpath), pbf);
        out = IOAssistance.openWriter(String.format("%s/listpersonfile.html", htmlpath));
        out.write(fwl.getHTML());
        out.close();
    }


    private String convertToHTML (String xml, String title) {
        StringBuilder sb = new StringBuilder();
        sb.append("<html><head><title>");
        sb.append(title);
        sb.append("</title></head><body><pre>");
        String[] lines = xml.split("\n");
        for (String line : lines) {
            line = line.replace(" ", "&nbsp;");
            line = line.replace("<", "&lt;");
            line = line.replace(">", "&gt;");
            reXMLID.reset(line);
            if (reXMLID.find()) {
                String anchor = reXMLID.getGroup(1);
                sb.append(String.format("<a name='%s'></a>\n", anchor));
            }
            sb.append(line);
            sb.append("\n");
        }

        sb.append("</pre></body></html>");

        return sb.toString();
    }

    private boolean fileExists (String name) {
        File f = new File(name);
        if (f.exists()) {
            return true;
        } else {
            System.out.format("%s does not exist\n", name);
            return false;
        }
    }

    public static boolean isCatalogEntry (String filename) {
        return filename.endsWith(".xml") && filename.charAt(0) != '.' && ignore.get(filename) == null;
    }

    Hashtable<String, String> directoryPaths = new Hashtable<String, String>();

    private boolean testExistence(String type, String dir) {
        boolean rc = false;
        return rc;
    }

    private boolean compareVersions () throws Exception {
        boolean tabDelimited = false;
        Hashtable<String, LocationInfo> v1 = new PathCollection("XML").getInfo();
        Hashtable<String, LocationInfo> v2 = new PathCollection("XMLnew").getInfo();
/*
        Enumeration<String> en = v2.keys();
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            if (v1.get(key) == null) {
                LocationInfo li = v2.get(key);
                System.out.format("[%s] %s:%d.%d %d\n", key, new File(li.filename).getName(), li.row, li.col, li.count);
            }
        }
*/
        BufferedWriter bw = IOAssistance.openWriter("pathcomp.html");
        bw.write("<html><head></head><body><table border='1'><tr><th>Row</th><th>Old</th><th>New</th></tr>\n");
        ArrayList<String> a1 = sortPaths(v1);
        ArrayList<String> a2 = sortPaths(v2);

        int len1 = a1.size();
        int len2 = a2.size();
        int len = len1;
        if (len < len2) {
            len = len2;
        }

        int rownum = -1;
        int idx1 = 0;
        int idx2 = 0;
        while (idx1 < len1 || idx2 < len2) {
            rownum++;
            String p1 = idx1 < len1 ? a1.get(idx1) : "";
            String p2 = idx2 < len2 ? a2.get(idx2) : "";
            if (p1.equals(p2)) {
                if (tabDelimited) {
                    bw.write(String.format("%s\t\t%s\t\n", p1, p2));
                } else {
                    bw.write(String.format("<tr><td>%d</td><td>%s</td><td>%s</td></tr>\n", rownum, p1, p2));
                }
                idx1++;
                idx2++;
            } else if (p1.length() > 0 && p1.compareTo(p2) < 0) {
                String loc = String.format("%s", v1.get(p1).toString());
                if (tabDelimited) {
                    bw.write(String.format("%s\t%s\t\t\n", p1, loc));
                } else {
                    bw.write(String.format("<tr><td>%d</td><td>%s<br/><span style='color:red'>%s</span></td><td>&nbsp;</td></tr>\n",
                        rownum, p1, loc));
                }
                idx1++;
            } else if (p2.length() > 0){
                String loc = String.format("%s", v2.get(p2).toString());
                if (tabDelimited) {
                    bw.write(String.format("\t\t%s\t%s\n", p2, loc));
                } else {
                    bw.write(String.format("<tr><td>%d</td><td>&nbsp;</td><td>%s<br/><span style='color:red'>%s</span></td></tr>\n",
                        rownum, p2, loc));
                }
                idx2++;
            }
        }
        bw.write("</tr></body></html>");
        bw.close();
        return true;
    }

    private ArrayList<String> sortPaths(Hashtable<String, LocationInfo> h) {
        ArrayList<String> a = new ArrayList<String>();
        Enumeration<String> en = h.keys();
        while (en.hasMoreElements()) {
            a.add(en.nextElement());
        }

        Collections.sort(a, new LengthSort());

        return a;
    }

    private void copyfile(File f1, String dtFile){
    try
    {
      File f2 = new File(dtFile);

      InputStream in = new FileInputStream(f1);
      OutputStream out = new FileOutputStream(f2);  // (f2, true) to append in to out

      byte[] buf = new byte[1024];
      int len;
      while ((len = in.read(buf)) > 0)
      {
        out.write(buf, 0, len);
      }
      in.close();
      out.close();
    }
    catch(FileNotFoundException ex)
    {
      System.out.println(ex.getMessage() + " in the specified directory.");
    }
    catch(IOException e){
      System.out.println(e.getMessage());
    }
  }

    void printUsage()
    {
        System.out.println("Usage:");
        System.out.println("Assumes that all .xml file names in the XML directory are lower case!");
        System.out.println("f=<name of XML file in XML directory>, e.g. f=penn0349 converts XML/penn0394.xml to HTML/penn0394.html. (Note: the .xml is optional)");
        System.out.println("f=* will convert all files in the XML directory to html files in the HTML directory.");
        System.out.println("p=<path>, e.g. p=TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier will print out all msIdentifier elements found in the files in the XML directory.");
        System.out.println("h=*, create siva/kitchenSink.html from all .html files in sconiva/HTML");
    }



    void formatPennCatalogValidate(String filename) throws Exception {
        Hashtable<String, Boolean> h = new Hashtable<String, Boolean>();
        //h.put("br33.xml", Boolean.TRUE);        // for debugging

        String teiDir = xmlpath;
        validate();
//        generateHTMLComparison("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents", false);

//        if (teiDir != null) return;

//        String teiDir = "C:/Users/ralph/Documents/PeterScharf/SanskritLibrary/siva/XML/";       // review how old was done

        if (filename != null) {
            formatPennCatalog(new File(filename));
        } else {
            File dir = new File(teiDir);
            File[] files = dir.listFiles();

            for (File file : files) {
                String fn = file.getName();
                if (h.get(fn) == null) {
                    //continue;                           // remove
                }

                if (!fn.equals("index.html") && 
                    !fn.equals("listbiblfile.xml") &&
                    !fn.equals("listpersonfile.xml") && fn.endsWith(".xml")) {
                    formatPennCatalog(file);
//                break;
                }
            }
        }

        /*
        BufferedWriter paths = IOAssistance.openWriter("sivapaths.txt");
        displayPaths(paths, "lb", ParseTEI.lb);
        displayPaths(paths, "lg", ParseTEI.lg);
        displayPaths(paths, "pb", ParseTEI.pb);
        paths.close();
         */
    }

    void displayPaths (BufferedWriter bw, String type, Hashtable<String, String> h)  throws Exception {
        bw.write(String.format("*****type=%s****\n", type));
        Enumeration<String> en = h.keys();
        ArrayList<String> a = new ArrayList<String>();
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            a.add(key);
        }

        Collections.sort(a);

        for (String key : a) {
            String fn = h.get(key);
            key = key.replace("/", ".");
            key = key.replace("&#160;", " ");
            bw.write(String.format("%s (%s)\n", key, fn));
        }

        bw.write("\n\n");
    }

    void validate() throws Exception {
        References.checkURL = false;
        String teiDir = xmlpath;
        File dir = new File(teiDir);
        File[] files = dir.listFiles();
        BufferedWriter bw = IOAssistance.openWriter("validationReport.txt");
        References.setFile(bw);

        // first pass, collected all xml:ids and verify that reference ids appear to be references
        int numfiles = 0;
        for (File file : files) {
            String fn = file.getName();
            if (isCatalogEntry(fn)) {
                numfiles++;
                validateFile(file, 1);
            }
        }

        Enumeration<String> en = ignore.keys();
        while (en.hasMoreElements()) {
            String filename = String.format("%s%s", this.xmlpath, en.nextElement());
            validateFile(new File(filename), 1);
        }

        if (References.wellFormednessError) {
            bw.write(String.format("Note: %d/%d files had well-formedness errors. These files will be ignored in the remaining passes. Their absence may cause other errors.\n",
                    References.notWellFormed.size(), numfiles));
        }

        // second pass, collect inplace xml and validate references
        for (File file : files) {
            String fn = file.getName();
            if (isCatalogEntry(fn)) {
                validateFile(file, 2);
            }
        }

        if (References.numErrors != 0) {
            System.out.format("%d error found during validation. See file validationReport.txt\n", References.numErrors);
        }

        bw.write(String.format("*****%d errors found.*****\n", References.numErrors));

        bw.close();
    }

    void validateFile (File file, int passNumber) throws Exception {
        References r = new References(passNumber);
        r.parse(file.getAbsolutePath());
    }

    void generateHTMLComparison(String elementName, boolean isName) throws Exception {
        String teiDirOld = "C:\\Users\\ralph\\Documents\\PeterScharf\\SanskritLibrary\\siva\\XML\\";
        String teiDirNew = "C:\\Users\\ralph\\Documents\\PeterScharf\\SanskritLibrary\\siva\\XMLnew\\";
        File dir = new File(teiDirOld);
        File[] files = dir.listFiles();
        BufferedWriter bw = IOAssistance.openWriter(String.format("compare_%s.html", elementName));
        bw.write(String.format("<html><head></head><body>"));
        CompareVersions.setFile(bw);
        int count = 0;
        for (File file : files) {
            String fn = file.getName();
            if (isCatalogEntry(fn)) {
                String newfn = file.getAbsolutePath();
                newfn = newfn.replace(teiDirOld, teiDirNew);
                File newf = new File(newfn);
                if (newf.exists()) {
                    bw.write(String.format("<h2>%s</h2>\n", fn));
                    bw.write("<table border='1px' style='border-collapse:collapse;width:1200px'><tr>\n");
                    if (isName) {
                        generateHTMLFromName(file, elementName);
                        generateHTMLFromName(newf, elementName);
                    } else {
                        generateHTMLFromPath(file, elementName);
                        generateHTMLFromPath(newf, elementName);
                    }
                    bw.write("</tr></table>\n");
                }
                count++;
            }
        }

        bw.write("</body></html>\n");
        bw.close();
    }

    void generateHTMLFromName (File file, String elementName) throws Exception {
        CompareVersions cv = new CompareVersions();
        cv.setElementName(elementName);
        cv.parse(file.getAbsolutePath());
    }

    void generateHTMLFromPath (File file, String elementPath) throws Exception {
        CompareVersions cv = new CompareVersions();
        cv.setElementPath(elementPath);
        cv.parse(file.getAbsolutePath());
    }

    void displayHTML() throws Exception
    {
        SectionLabel[] sls = new SectionLabel[]
        {
            new SectionLabel("Manuscript Identifiers", "Collection"),
            new SectionLabel("Manuscript Identifiers", "Repository"),
            new SectionLabel("Manuscript Identifiers", "Institution"),
            new SectionLabel("Manuscript Identifiers", "Location"),
            new SectionLabel("Manuscript Identifiers", "Catalog"),
            new SectionLabel("Manuscript Identifiers", "Item"),
            new SectionLabel("Contents", "Work"),
            new SectionLabel("Contents", "Locus"),
            new SectionLabel("Contents", "Author"),
            new SectionLabel("Contents", "Title"),
            new SectionLabel("Contents", "Part"),
            new SectionLabel("Contents", "Cover rubric"),
            new SectionLabel("Contents", "Incipit"),
            new SectionLabel("Contents", "Explicit"),
            new SectionLabel("Contents", "Final rubric"),
            new SectionLabel("Contents", "Colophon"),
            new SectionLabel("Contents", "Filiation"),
            new SectionLabel("Contents", "Second folio"),
            new SectionLabel("Contents", "Language"),
            new SectionLabel("Physical description", "Form"),
            new SectionLabel("Physical description", "Material"),
            new SectionLabel("Physical description", "Watermark"),
            new SectionLabel("Physical description", "Extent"),
            new SectionLabel("Physical description", "Dimension"),
            new SectionLabel("Physical description", "Foliation"),
            new SectionLabel("Physical description", "Collation"),
            new SectionLabel("Physical description", "Condition"),
            new SectionLabel("Physical description", "Binding"),
            new SectionLabel("Physical description", "Seal"),
            new SectionLabel("Physical description", "Accompanying matter"),
            new SectionLabel("Physical description", "Layout"),
            new SectionLabel("Hands", "Summary"),
            new SectionLabel("Hands", "Hand"),
            new SectionLabel("Hands", "Additions"),
            new SectionLabel("Decoration", "Cover"),
            new SectionLabel("Decoration", "Border"),
            new SectionLabel("Decoration", "Illustration"),
            new SectionLabel("History", "Origin"),
            new SectionLabel("History", "Provenance"),
            new SectionLabel("History", "Acquisition"),
            new SectionLabel("Subject headings", "SubjectLC"),
            new SectionLabel("Subject headings", "SubjectSL"),
            new SectionLabel("Facsimile", "*"),
            new SectionLabel("Transcription", "*"),
        };
        BufferedWriter bw = IOAssistance.openWriter("kitchenSink.html");
        bw.write("<html><head><title>Kitchen Sink</title><meta http-equiv='Content-Type' content='text/html; charset=utf-8'/><body>");
        File dir = new File("HTML");
        File[] files = dir.listFiles();
        String header = "";
        boolean firstTime = true;
        for (SectionLabel sl : sls)
        {
            int n = 0;
            for (File file : files)
            {
                String fn = file.getName().toLowerCase();
                if (fn.startsWith("br") || fn.startsWith("penn"))
                {
                    String s = String.format("<h3>%s[%s]</h3>", sl.section, sl.label);
                    if (!header.equals(s))
                    {
                        if (!firstTime)
                            bw.write("</table>");
                        else
                            firstTime = false;
                        bw.write(s);
                        bw.write("<table>");
                        header = s;
                    }
                    new ParseHTML(file.getAbsolutePath(), sl.section, sl.label, bw);
                }
            }
        }

        bw.write("</table></body></html>");
        bw.close();
    }

    @SuppressWarnings("ResultOfObjectAllocationIgnored")
    void preprocessPennCatalog() throws Exception
    {
        String originalTEIDir = "XML-matthias/";
        String tempXMLDir = "XML-renamed/";
        String marcDir = "MARC3/";
        String[] marcfiles = new String[] {"mbh_bhp.xml", "penn_skt_2records.xml", "up2618.xml"};
        Hashtable<String, MarcRecord> marcRecords = new Hashtable<String, MarcRecord>();

        renameFiles(originalTEIDir, tempXMLDir);

        Hashtable<String, String> h = mapItemNumbersToFiles(tempXMLDir);

        for (String marcfile : marcfiles)
        {
            ParseMARC pm = new ParseMARC(marcDir + marcfile, h, tempXMLDir, this, marcRecords);
        }

        File dir = new File(tempXMLDir);
        File[] xmlfiles = dir.listFiles();
        for (File xmlfile: xmlfiles)
        {
            String fn = xmlfile.getName();
            String key = String.format("%s%s", tempXMLDir, fn);
            ParseTEIInsert pti = new ParseTEIInsert(key, this, marcRecords.get(key), config);
            pti.parseTEI();
        }

        BufferedWriter bw = IOAssistance.openWriter("empty.txt");
        PathInfo.writeEmptyElements(bw);
        bw.close();
    }

    Hashtable<String, String> renameFiles(String originalTEIDir, String tempXMLDir) throws Exception
    {
        Hashtable h = new Hashtable<String, String>();
        String[] directories = new String[]
        {
            "MSS-BhG",
            "MSS-BhP",
            "MSS-Br-Skt-Mbh",
            "Mss-Br-Te-BhP",
            "Mss-Br-Te-Mbh",
            "MSS-MBh",
            "MSS-misc",
            "MSS-stotra-as",
            "MSS-stotra-gm",
            "MSS-stotra-misc",
            "MSS-stotra-sr",
            "MSS-stotra-vsn",
        };

        for (String directory : directories)
        {
            renameXMLFiles(originalTEIDir + directory, tempXMLDir);
        }

        return h;
    }

    void renameXMLFiles(String directory, String tempXMLDir) throws Exception
    {
        File dir = new File(directory);
        File[] files = dir.listFiles();
        Regex re = new Regex("", "(\\d+)");
        for (File file : files)
        {
            String name = file.getName();
            if (isCatalogEntry(name))
            {
                re.reset(name);
                if (re.find())
                {
                    String prefix = "?";
                    String itemNumber = re.getGroup(1);
                    if (name.startsWith("MsDescPenn"))
                        prefix = "penn";
                    else if (name.startsWith("BrMs"))
                        prefix = "br";

                    String target = String.format("%s/%s%s.xml", tempXMLDir, prefix, itemNumber);
                    copyfile(file, target);
                }
            }
        }
    }

    static Hashtable<String, String> mapItemNumbersToFiles(String teiDir) throws Exception
    {
        Hashtable h = new Hashtable<String, String>();

        getXMLFiles(teiDir, h);

        return h;
    }


    static void getXMLFiles(String directory, Hashtable<String, String> h) throws Exception
    {
        File dir = new File(directory);
        File[] files = dir.listFiles();
        Regex re = new Regex("", "(\\d+)");
        for (File file : files)
        {
            String filename = file.getName();
            if (isCatalogEntry(filename))
            {
                re.reset(file.getAbsolutePath());
                if (re.find())
                {
                    String itemNumber = re.getGroup(1);
                    h.put(itemNumber, String.format("%s%s", directory, file.getName()));
                }
            }
        }
    }



    void formatPennCatalog(File f) throws Exception
    {
        if (f.exists()) {
            ParseTEIFormat ptf = new ParseTEIFormat(f.getAbsolutePath(), this, config);
            ptf.parseTEI();
        } else
            System.out.format("System cannot find file: %s\n", f.getAbsolutePath());
    }


    void extractFromPennCatalog(String extraction) throws Exception
    {
        String teiDir = "XML";
        BufferedWriter bw = null;
        bw = IOAssistance.openWriter("extract.xml");
        bw.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        bw.write(String.format("<extractions path='%s'>\n", extraction));
        File dir = new File(teiDir);
        File[] files = dir.listFiles();
        for (File file : files)
        {
            String fn = file.getName();
            if (isCatalogEntry(fn))
            {
                bw.write(String.format("\n\n<file name='%s'/>\n\n", fn));
                ParseTEIExtract pte = new ParseTEIExtract(file.getAbsolutePath(), extraction, bw, this, config);
                pte.parseTEI();
            }
        }
        bw.write("<allpaths>");
        ParseTEI.writeAllPaths(bw);
        bw.write("</allpaths>");
        bw.write("\n</extractions>\n");

        bw.close();
    }

    public InputStream[] getInputStreams(String sourceEncoding, String targetEncoding)
    {
        String[] fileNames = FSM.createFileNames("", sourceEncoding, targetEncoding);
        InputStream[] iss = new InputStream[fileNames.length];
        int i = 0;

        for (String name : fileNames)
        {
            String fname = String.format("%s%s", config.transcoderDirectory, name);
            try
            {
                iss[i++] = new FileInputStream(fname);
            } catch (Exception ex)
            {
                System.out.println(String.format("In getInputStreams: %s", ex.getMessage()));
            }
        }

        return iss;
    }

    class LengthSort implements Comparator<String> {

        public int compare(String o1, String o2) {
            int len1 = o1.length();
            int len2 = o1.length();
            if (len1 != len2) {
                return o1.length() - o2.length();
            } else {
                return o1.compareTo(o2);
            }
        }

    }
}
