package siva;

/* Copied to Utilities.tei.manuscript*/

import java.io.File;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.SAXParserHandler;
import org.sanskritlibrary.StackEntry;
import org.w3c.dom.Node;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
//DefaultHandler implements ContentHandler, DTDHandler, EntityResolver, ErrorHandler
//import org.xml.sax.helpers.DefaultHandler;

//DefaultHandler2 implements ContentHandler, DTDHandler, EntityResolver, ErrorHandler, DeclHandler, EntityResolver2, LexicalHandler
//http://docs.oracle.com/javase/1.5.0/docs/api/org/xml/sax/ext/DefaultHandler2.html
import org.xml.sax.ext.DefaultHandler2;         // this allows processing of comments because it implements LexicalHandler


// SAXParserHandler is an interface that forces us to include required methods.
// Serializable is optional. Have made local variables that don't need to be serialized transient.
public class ParseLayout extends DefaultHandler2 implements SAXParserHandler, Serializable, Comparator<MarkupInfo> {
    transient Locator locator;
    transient ArrayList<StackEntry> pathStack = new ArrayList<StackEntry>();
    transient Hashtable<String, StringBuilder> buffers = new Hashtable<String, StringBuilder>();
    public static Hashtable<String, Boolean> expectedTags = new Hashtable<String, Boolean>();
    private static Hashtable<String, Boolean> markupHash = new Hashtable<String, Boolean>();
    private static ArrayList<MarkupInfo> markup = new ArrayList<MarkupInfo>();
    private static Hashtable<String, Boolean> ignoreHash = new Hashtable<String, Boolean>();
    private static ArrayList<MarkupInfo> ignore = new ArrayList<MarkupInfo>();
    private static Hashtable<String, MacroInfo> macros = new Hashtable<String, MacroInfo>();
    private static String root;
    private static Hashtable<String, Style> layoutTypes = new Hashtable<String, Style>();
    public static Hashtable<String, Style> styleOverrides = new Hashtable<String, Style>();
    public static Hashtable<String, Hashtable<String, MapInfo>> maps = new Hashtable<String, Hashtable<String, MapInfo>>();
    static {
        layoutTypes.put("text", Style.TEXT_STRUCTURE);
        layoutTypes.put("manuscript", Style.MANUSCRIPT_LAYOUT);
        layoutTypes.put("header", Style.HEADER);
    }
    static boolean styleOverridden = false;

    Style currentStyle;
    ArrayList<String> currentPrefixStack = new ArrayList<String>();
    
    String currentMarkupPath = null;
    ArrayList<FormatInfo> currentMarkupBeforeFormat = new ArrayList<FormatInfo>();
    ArrayList<FormatInfo> currentMarkupAfterFormat = new ArrayList<FormatInfo>();
    ArrayList<FormatInfo> currentMarkupFormat = new ArrayList<FormatInfo>();
    ArrayList<FormatInfo> currentMacroFormat = new ArrayList<FormatInfo>();
    ArrayList<TestExistenceInfo> currentMacroTests = new ArrayList<TestExistenceInfo>();
    Hashtable<String, MapInfo> currentMap;
    MapInfo currentMapInfoValue;

    FormatInfoRaw currentFormatRaw = null;

    String currentStylePath = null;
    String currentStyleValue = null;
    TestExistenceInfo currentMacroTest = null;
    String currentMacroID = null;

    String currentFormatBuffer = null;

    String currentMapType;
    String currentMissing;
    String currentCapitalizeFirstLetter;

    String currentValueMap = null;
    
    public ParseLayout(String fileName) throws Exception
    {
        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();

        // Parse the input
        SAXParser saxParser = factory.newSAXParser();
        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
        // Uncomment the following to turn on comment processing.
        // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
        // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        try {
            saxParser.parse( new File(fileName), this);
        } catch (Exception ex) {
            throw ex;
        }
    }

	// Maintains a stack of element names (qName)
    private void push (String qName, Attributes att) {
        pathStack.add(new StackEntry(qName, att));
    }

	// Pop a qName from the stack
    private StackEntry pop () {
        int last = pathStack.size()-1;
        StackEntry top = null;
        if (last >= 0) {
            top = pathStack.get(last);
            pathStack.remove(last);
        }

        return top;
    }

	// Peek at a qName without popping stack
    private StackEntry peek () {
        int last = pathStack.size()-1;
        StackEntry top = null;
        if (last >= 0) {
            top = pathStack.get(last);
        }

        return top;

    }

    private StackEntry peek (int n) {
        int last = pathStack.size()-1 - n;
        StackEntry top = null;
        if (last >= 0) {
            top = pathStack.get(last);
        }

        return top;

    }
	// Serialize the qName stack. The string that is returned represent the path to the current qName
    private String buildPath () {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (StackEntry se : pathStack) {
            sb.append(String.format("%s%s", delim, se.qName));
            delim = ".";
        }
        return sb.toString();
    }

    private void displayError(String msg) {
        System.out.format("Line %d: %s\n", locator.getLineNumber(), msg);
    }

    @Override
    public void warning(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void error(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    @Override
    public void startDocument() throws SAXException {
    }

    @Override
    public void endDocument() throws SAXException {

    }


    public static boolean ignoreElement(Node node, String context, Style style) {
        MarkupInfo result = getMarkup(node, ignore, context, style);

        if (result == null) {
            result = getMarkup(node, ignore, context, Style.ANY);
            if (result != null) {
                style = style.ANY;
            }
        }

        return result != null;
    }

    static Style overrideStyle (String key, Style style) {
        int idx = key.indexOf('[');
        styleOverridden = false;
        if (idx != -1) {
            key = key.substring(0, idx);
        }
        if (ParseLayout.styleOverrides.get(key) != null) {
            style = ParseLayout.styleOverrides.get(key);
            styleOverridden = true;
        }

        return style;
    }

    public static MarkupInfo getMarkup(Node node, String context, Style style) {
        MarkupInfo result = getMarkup(node, markup, context, style);

        if (result == null) {
            result = getMarkup(node, markup, context, Style.ANY);
            if (result != null) {
                style = style.ANY;
            }
        }

        return result;
    }

    public static MacroInfo getMacro(Style style, String id) {
        return macros.get(String.format("%s/%s", style.name(), id));
    }

    private static MarkupInfo getMarkup (Node node, ArrayList<MarkupInfo> mis, String context, Style style) {
        MarkupInfo result = null;
        String s_style = style.name();


        for (MarkupInfo mi : mis) {
            if (mi.debugging == 45) {
                if (context.endsWith(".div.gap")) {
                    mi.debugging = 45;
                }
            }
            if (!mi.style.equals(s_style)) {
                continue;
            }

            if (mi.path.re != null) {
                mi.path.re.reset(context);
                if (mi.path.re.find()) {
                    result = mi;
                    if (mi.debugging == 44) {
                        mi.debugging = 44;
                    }
                }
            } else if (mi.path != null) {
                String path = mi.path.toString();
                if (path.startsWith(ParseLayout.root)) {
                    if (context.equals(path)) {
                        result = mi;
                    }
                } else if (context.endsWith(String.format(".%s", mi.path.toString()))) {
                    result = mi;
                }
            }

            if (result != null) {
                if (result.path.attributeName == null) {
                    break;          // no attribute in path so we are done.
                } else {
                    Node att = node.getAttributes().getNamedItem(result.path.attributeName);
                    if (att != null) {
                        if (att.getNodeValue().equals(result.path.attributeValue)) {
                            break;
                        } else {
                            result = null;
                        }
                    } else {
                        result = null;          // not it, need the right attribute and value
                    }
                }
            }
        }
        return result;
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException {
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        if (locator.getLineNumber() == 158) {
            localName = null;
        }
        push(qName, atts);
        String path = buildPath();
        if (qName.equals("format")) {
            currentFormatBuffer = path;
            StringBuilder bfr = buffers.get(currentFormatBuffer);
            if (bfr == null) {
                buffers.put(currentFormatBuffer, new StringBuilder());
            }
            currentFormatRaw = new FormatInfoRaw();
            if (atts.getValue("test") != null) {
                currentFormatRaw.test = atts.getValue("test");
            }
            currentFormatRaw.buffer = null;
        }
        String prefix = atts.getValue("prefix");
        pushPrefix(prefix);
        if (path.equals("layouts")) {
            root = atts.getValue("root");
            if (root == null) {
                System.out.println("No root attribute of layouts element. Root is set to TEI");
                root = "TEI";
            }
            root += ".";            
        } else if (path.equals("layouts.layout")) {
            currentStyle = layoutTypes.get(atts.getValue("style"));
        } else if (path.endsWith("macro")) {
            currentMacroID = atts.getValue("id");
        } else if (path.endsWith("macro.test")) {
            currentMacroTest = new TestExistenceInfo();
            if (atts.getValue("outcome") != null) {
                currentMacroTest.outcome = atts.getValue("outcome").equals("true");
            }
        } else if (path.endsWith("maps.map")) {
            currentMap = new Hashtable<String, MapInfo>();
            String id = atts.getValue("id");
            currentMapType = atts.getValue("type");
            currentMissing = atts.getValue("missing");
            currentCapitalizeFirstLetter = atts.getValue("capitalizeFirstLetter");
            maps.put(id, currentMap);
        } else if (path.endsWith("maps.map.value")) {
            currentMapInfoValue = new MapInfo(currentMissing, currentMapType, currentCapitalizeFirstLetter);
        } else if (path.endsWith("format.value")) {
            currentValueMap = atts.getValue("map");
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        String path = buildPath();
        StringBuilder buffer = buffers.get(path);
        String value = "";
        if (locator.getLineNumber() == 408) {
            localName = null;
        }
        if (buffer != null) {
            value = buffer.toString();
            buffer.setLength(0);
        }

        if (path.equals("layouts.layout")) {
            currentStyle = null;
        } else if (path.equals("layouts.expectedTags.tag")) {
            expectedTags.put(value, Boolean.TRUE);
        } else if (path.endsWith("ignore.path")) {
            MarkupInfo mi = new MarkupInfo(value, null, null, null, currentStyle, locator.getLineNumber());
            if (ignoreHash.get(mi.getKey()) == null) {
                ignoreHash.put(mi.getKey(), Boolean.TRUE);
                ignore.add(mi);
            } else {
                System.out.format("Duplicate ignore: %d: %s\n", locator.getLineNumber(), mi.getKey());
            }
        } else if (path.endsWith("markup")) {
            MarkupInfo mi = new MarkupInfo(currentMarkupPath,
                    currentMarkupFormat,
                    currentMarkupBeforeFormat,
                    currentMarkupAfterFormat,
                    currentStyle,
                    locator.getLineNumber());
            String key = null;
            if (markupHash.get(mi.getKey()) == null) {
                markupHash.put(mi.getKey(), Boolean.TRUE);
                markup.add(mi);
            } else {
                System.out.format("Duplicate markup: %d: %s\n", locator.getLineNumber(), mi.getKey());
            }

            currentMarkupPath = null;
            currentMarkupBeforeFormat = new ArrayList<FormatInfo>();
            currentMarkupAfterFormat = new ArrayList<FormatInfo>();
            currentMarkupFormat = new ArrayList<FormatInfo>();
        } else if (path.endsWith("markup.path")) {
            currentMarkupPath = value;
        } else if (path.endsWith("markup.before")) {
            if (currentMarkupBeforeFormat.size() == 0) {
                currentMarkupBeforeFormat.add(new FormatInfo(new FormatInfoRaw(null, value)));
            }
        } else if (path.endsWith("markup.after")) {
            if (currentMarkupAfterFormat.size() == 0) {
                currentMarkupAfterFormat.add(new FormatInfo(new FormatInfoRaw(null, value)));
            }
        } else if (path.endsWith("markup.format")) {
            currentFormatRaw.buffer = value;
            currentMarkupFormat.add(new FormatInfo(currentFormatRaw));
        } else if (path.endsWith("markup.before.format")) {
            currentFormatRaw.buffer = value;
            currentMarkupBeforeFormat.add(new FormatInfo(currentFormatRaw));
        } else if (path.endsWith("markup.after.format")) {
            currentFormatRaw.buffer = value;
            currentMarkupAfterFormat.add(new FormatInfo(currentFormatRaw));
        } else if (path.endsWith("format.value")) {
            StringBuilder formatBuffer = buffers.get(currentFormatBuffer);
            if (currentValueMap != null) {
                formatBuffer.append(String.format("<value map='%s'>%s</value>", currentValueMap, value));
                currentValueMap = null;
            } else {
                formatBuffer.append(String.format("<value>%s</value>", value));
            }
        } else if (path.endsWith("format.macro")) {
            StringBuilder formatBuffer = buffers.get(currentFormatBuffer);
            StackEntry se = peek();
            String id = se.atthash.get("id");
            formatBuffer.append(String.format("<macro id='%s'/>", id));
        } else if (path.equals("layouts.styles.style.path")) {
            currentStylePath = value;
        } else if (path.equals("layouts.styles.style.value")) {
            currentStyleValue = value;
        } else if (path.equals("layouts.styles.style")) {
            styleOverrides.put(currentStylePath, layoutTypes.get(currentStyleValue));
        } else if (path.endsWith("macro")) {
            macros.put(String.format("%s/%s", currentStyle.name(), currentMacroID), new MacroInfo(currentMacroID, currentMacroFormat, currentMacroTests));
        } else if (path.endsWith("macro.test")) {
            currentMacroTests.add(currentMacroTest);
        } else if (path.endsWith("macro.test.path")) {
            currentMacroTest.path = new Path(value);
        } else if (path.endsWith("macro.format")) {
            currentFormatRaw.buffer = value;
            currentMacroFormat.add(new FormatInfo(currentFormatRaw));
        } else if (path.endsWith("maps.map")) {
            currentMap.put("?", currentMapInfoValue);               // so that information will be available for what to do with non-mapped values
        } else if (path.endsWith("maps.map.value")) {
            currentMap.put(value, currentMapInfoValue);
        }


        pop();
        currentPrefixStack.remove(currentPrefixStack.size() - 1);
    }


    private void pushPrefix (String prefix) {
        if (prefix != null) {
            currentPrefixStack.add(prefix.trim());
        } else if (currentPrefixStack.size() == 0) {
            currentPrefixStack.add("");
        } else {
            currentPrefixStack.add(currentPrefixStack.get(currentPrefixStack.size()-1));
        }
    }


    @Override
    // Collects the innerText in a StringBuilder keyed by the current path. It is endElement's responsibility
    // to retrieve the StringBuilder and set its length to 0.
    // This is optional. It is not specified in SAXParserHandler
    public void characters(char[] ch, int start, int length) throws SAXException {
        String path = buildPath();
        StringBuilder sb = buffers.get(path);
        if (sb == null) {
            sb = new StringBuilder();
            buffers.put(path, sb);
        }
        sb.append(new String(ch, start, length));
    }

    @Override
    // Optional if you want to be notified of comments
    public void comment(char[] ch, int start, int length) throws SAXException {

    }

    public int compare(MarkupInfo mi1, MarkupInfo mi2) {
        if (mi1.re != null && mi2.re != null) {
            return mi1.re.getPattern().length() - mi2.re.getPattern().length();
        } else if (mi1.path != null && mi2.path != null) {
            return mi1.path.toString().length() - mi2.path.toString().length();
        } else if (mi1.re != null) {
            return -1;          // choose mi1
        } else {
            return 1;           // choose mi2
        }
    }

}
