package siva;

import java.util.ArrayList;
import java.io.File;
import java.io.IOException;
import java.io.OutputStreamWriter;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;
import org.sanskritlibrary.ResourceInputStream;
import java.util.Hashtable;

import javax.xml.parsers.SAXParserFactory;
//import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;

public class ParseMARC extends DefaultHandler //implements ErrorHandler, ContentHandler, DTDHandler, EntityResolver
{
    String currentDirectory;
    Locator locator;
    OutputStreamWriter out;
    ArrayList<String> path = new ArrayList<String>();
    Hashtable<String, MarcRecord> marcRecords = new Hashtable<String, MarcRecord>();
    MarcRecord mr = new MarcRecord();
    Hashtable<String, String> h;
    String teiDir;
    ResourceInputStream ris;

    public ParseMARC(String fileName, Hashtable<String, String> h, String teiDir, ResourceInputStream ris, Hashtable<String, MarcRecord> marcRecords)
    {
        this.h = h;
        this.teiDir = teiDir;
        this.ris = ris;
        this.marcRecords = marcRecords;
        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Set up output stream
            out = new OutputStreamWriter (System.out, "UTF8");

            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse( new File(fileName), this);

        }
        catch (Throwable t)
        {
            t.printStackTrace ();
        }
    }


    private String getPath()
    {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (String name : path)
        {
            sb.append(String.format("%s%s", delim, name));
            delim = ".";
        }

        return sb.toString();
    }


    private void pushPath(String qname, Attributes atts)
    {
        String append = "";
        if (atts.getValue("tag") != null)
            append = String.format("[tag='%s']", atts.getValue("tag"));
        else if (atts.getValue("code") != null)
            append = String.format("[code='%s']", atts.getValue("code"));

        path.add(qname + append);
    }


    // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
    }

    @Override
    public void endDocument() throws SAXException
    {
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        if (qName.substring(0, 5).equals("marc:"))
            qName = qName.substring(5);         // skip over marc:
        pushPath(qName, atts);
        if (getPath().equals("collection.record"))
        {
            mr.clear();
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (qName.substring(0, 5).equals("marc:"))
            qName = qName.substring(5);         // skip over marc:
        String p = getPath();

        if (p.equals("collection.record"))
        {
            boolean ok = true;
            String fn = h.get(mr.getDatafield773());
            if (fn != null)
            {
                File f = new File(fn);
                if (!f.exists())
                {
                    ok = false;
                    System.out.format("%s does not exist\n", fn);
                }
            }
            else
            {
                ok = false;
                System.out.format("%s is not in hash table\n", mr.getDatafield773());
            }

            if (ok)         // insert marc info into given file
            {
                String filename = h.get(mr.getDatafield773());
                try
                {
                    marcRecords.put(filename, mr);
                    //new ParseTEI(filename, ParseTEI.ACTION_INSERT, ris, mr);
                }
                catch (Exception ex)
                {
                    throw new SAXException(ex.getMessage());
                }
            }
        }
        path.remove(path.size() - 1);

    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        String p = getPath();
        String s = new String(ch, start, length);

        if (p.equals("collection.record.leader"))
        {
            mr.setLeader(s);
        }
        else if (p.equals("collection.record.controlfield[tag='001']"))
        {
            mr.setControl001(s);
        }
        else if (p.equals("collection.record.controlfield[tag='008']"))
        {
            mr.setControl008(s);
        }
        else if (p.equals("collection.record.datafield[tag='035'].subfield[code='a']"))
        {
            mr.setDatafield035(s);
        }
        else if (p.equals("collection.record.datafield[tag='079'].subfield[code='a']"))
        {
            mr.setDatafield079(s);
        }
        else if (p.equals("collection.record.datafield[tag='773'].subfield[code='g']"))
        {
            mr.setDatafield773(s);
        }
    }

    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void processingInstruction(String target, String data) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void skippedEntity(String name) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    // For DTDHandler
    @Override
    public void notationDecl(String name, String publicId, String systemId) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void unparsedEntityDecl(String name, String publicId, String systemId, String notationName) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    // For EntityHandler
    @Override
    public InputSource resolveEntity(String publicId, String systemId) throws SAXException, IOException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }
}
