/*
http://www.tei-c.org/release/doc/tei-p5-doc/en/html/index-toc.html
*/
/*
 * Currently only penn0452, penn0488, penn0489, penn0490 have zones
 */

package siva;

import java.io.BufferedWriter;
import java.util.ArrayList;
import java.io.File;
import java.io.StringReader;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.XML;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.ext.DefaultHandler2;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.Config;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.ResourceInputStream;
import org.sanskritlibrary.SubjectClassifications;
import org.sanskritlibrary.TextFormatter;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public abstract class ParseTEI extends DefaultHandler2 //implements ErrorHandler, ContentHandler, DTDHandler, EntityResolver
{
    static Regex reTD = new Regex("", "<td.*?>\\s*$");
    static Regex reTable = new Regex("", "</table>\\s*$");
    static Regex reHeader = new Regex("", "</h3>\\s*$");
    static Regex reHTML = new Regex("", "^xml/(.+)\\.xml(#.*)?$");
    static Regex reNoExtension = new Regex("", "^(.*)\\.");
    static Regex reText = new Regex("", Main.textSubdirectory + "([^#]+)#([^\\s]*)");
    static Regex reTextNoType = new Regex("", "([^#]+)#([^\\s]*)");

    public enum RenderType {
        COLOR, SMALLCAP, ITALIC, SMALL, BOLD;  //; is optional
    }


    Style style = Style.HEADER;


    //    static public final String textDirectory = "texts/";
//    static public final String textURL = "http://sanskrit1d.ccv.brown.edu/tomcat/Announce/";
    static public final String indent = "&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;";
    static public final String xmlindent = "                                                                                                              ";
    static public final String rowTemplate = "<tr><td valign='top'>%s%s:</td><td>%s</td></tr>\n";
    static public final String rowTemplateNoLabel = "<tr><td valign='top'></td><td>%s</td></tr>\n";
    static public final String rowTemplateHeader = "<tr><th valign='top'>%s:</th><th>%s</th></tr>\n";
    static public final String rowTemplateNoLabelHeader = "<tr><th valign='top'></th><th>%s</th></tr>\n";
    static public final int ACTION_INSERT = 1;
    static public final int ACTION_FORMAT = 2;
    static public final int ACTION_EXTRACT = 3;
    static public String target = "roman";
    static private SubjectClassifications sc = null;
    static private Hashtable<String, String> classifications;
    final static Hashtable<String, Boolean>onetag = new Hashtable<String, Boolean>();
    final static Hashtable<String, Boolean> embeddedTags = new Hashtable<String, Boolean>();
    final static Hashtable<String, RenderType> rendAttributeValues = new Hashtable<String, RenderType>();
    static Hashtable<String, Boolean> unexpected = new Hashtable<String, Boolean>();
    static Hashtable<String, Boolean> missingPoundSign = new Hashtable<String, Boolean>();
    static Hashtable<String, Boolean> errorTags = new Hashtable<String, Boolean>();
    static public Hashtable<String, StringBuilder> action = new Hashtable<String, StringBuilder>();
    static public BufferedWriter sivaErrorReport = null;
    static protected String script;
    static
    {
        try {
            script = IOAssistance.readEntireFile("script.js");
        } catch (Exception ex) {
            System.out.println(String.format("Error reading script.js: ", ex.getMessage()));
        }
        rendAttributeValues.put("smallcap", RenderType.SMALLCAP);
        rendAttributeValues.put("red", RenderType.COLOR);
        rendAttributeValues.put("italic", RenderType.ITALIC);
        rendAttributeValues.put("small", RenderType.SMALL);
        rendAttributeValues.put("bold", RenderType.BOLD);
        
        onetag.put("lb", Boolean.TRUE);
        onetag.put("pb", Boolean.TRUE);
        onetag.put("catRef", Boolean.TRUE);
        onetag.put("graphic", Boolean.TRUE);
        onetag.put("cb", Boolean.TRUE);
        onetag.put("milestone", Boolean.TRUE);
        onetag.put("refState", Boolean.TRUE);
        onetag.put("caesura", Boolean.TRUE);
        onetag.put("link", Boolean.TRUE);
        onetag.put("anchor", Boolean.TRUE);

        embeddedTags.put("add", Boolean.TRUE);
        embeddedTags.put("biblScope", Boolean.TRUE);
        embeddedTags.put("catRef", Boolean.TRUE);
        embeddedTags.put("corr", Boolean.TRUE);
        embeddedTags.put("persName", Boolean.TRUE);
        embeddedTags.put("date", Boolean.TRUE);
        embeddedTags.put("desc", Boolean.TRUE);
        embeddedTags.put("del", Boolean.TRUE);
        embeddedTags.put("gap", Boolean.TRUE);
        embeddedTags.put("graphic", Boolean.TRUE);
        embeddedTags.put("sic", Boolean.TRUE);
        embeddedTags.put("unclear", Boolean.TRUE);
        embeddedTags.put("foreign", Boolean.TRUE);
    }

    private static int debugging = 0;
    //File fn;
    String fileName;
    String shortName;

    protected int tagLevel = 0;
    protected PathInfo pi;
    ResourceInputStream ris;
    String currentDirectory;
    Locator locator;
    ArrayList<String> path = new ArrayList<String>();
    int debug = 0;
    int ignoreUntil = -1;
    BufferedWriter bw;
    boolean endCommentWithLineFeed = true;      // for comments before root element
    String extractionPath;
    StringBuilder contents = new StringBuilder();
    StringBuilder manuscriptIdentifiers = new StringBuilder();
    StringBuilder physicalDescription = new StringBuilder();
    private Hashtable<String, Boolean> references = null;
    private Hashtable<String, String> targets = null;

    StringBuilder hands = new StringBuilder();

    StringBuilder decoration = new StringBuilder();
    StringBuilder history = new StringBuilder();

    StringBuilder subjectHeadings = new StringBuilder();

    StringBuilder facsimile = new StringBuilder();

    StringBuilder transcription = new StringBuilder();
    StringBuilder changes = new StringBuilder();

    ArrayList<ChangeInfo>changeItems = new ArrayList<ChangeInfo>();

    String theFileName;

    ArrayList<String> languageStack = new ArrayList<String>();
    StringBuilder currentBuffer = new StringBuilder();
    boolean biblLink = false;
    String newElement = null;
    MarcRecord mr = null;
    boolean insertMARC = true;;
    String measureQuantity = null;
    ArrayList<Integer> inList = new ArrayList<Integer>();
    int columnCount = 0;
    int choiceLevel = -1;
    boolean firstChoice = false;
    public static Hashtable<String, String> lb = new Hashtable<String, String>();
    public static Hashtable<String, String> pb = new Hashtable<String, String>();
    public static Hashtable<String, String> lg = new Hashtable<String, String>();

    StringBuilder insertBuffer = new StringBuilder();
    int currentFacsimile = 1;
    String lastAppend = "";
    int levelNoLastEndTable = -1;
    Hashtable<String, String> zoneMap = null;
    
    SAXParser saxParser;
    Config config;

    /*
    public ParseTEI(String fileName, ResourceInputStream ris, MarcRecord mr) throws Exception
    {
        this(fileName, null, null, ris, mr);
    }

    public ParseTEI(String fileName, ResourceInputStream ris) throws Exception
    {
        this(fileName, null, null, ris, null);
    }

    public ParseTEI(String fileName, String extractionPath, BufferedWriter bw, ResourceInputStream ris) throws Exception
    {
        this(fileName, extractionPath, bw, ris, null);
    }
    */

    public ParseTEI(String fileName, String extractionPath, BufferedWriter bw, ResourceInputStream ris, MarcRecord mr, Config config) throws Exception
    {
        this.ris = ris;
        this.mr = mr;
        this.fileName = fileName;
        this.config = config;
        File fn = new File(fileName);
        shortName = fn.getName();
        theFileName = shortName;             // for debugging
        this.extractionPath = extractionPath;
        this.bw = bw;


        if (sc == null)
        {
            sc = new SubjectClassifications(new File("IndicSubjectClassification.xml"), config);
            classifications = sc.getClassifications();
        }


        pi = new PathInfo();


        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();
        saxParser = factory.newSAXParser();
    }
    
    
    public void parseTEI() throws Exception {        
        prepareParse(saxParser);
        zoneMap = new ParseFacsimiles(fileName).getZoneMap();
        System.out.println(fileName);
        saxParser.parse(fileName, this);
    }
    
    abstract void prepareParse(SAXParser saxParser) throws Exception;    
    /*
        try
        {
            // Parse the input
            if (actionType == ACTION_INSERT)
            {
            }
            else if (actionType == ACTION_FORMAT)
            {
            }
            else if (actionType == ACTION_EXTRACT)
            {
                // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description,
                endCommentWithLineFeed = true;
                saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
            }

        }
        catch (Throwable t)
        {
            displayError(t.getMessage());
            //t.printStackTrace ();
        }
        
    }
    */
    
    protected void setKeys () {
            pi.key("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote", "type");

            // Set handlers (will be called when end tag of path is reached)
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc", new Handler_physDesc());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier", new Handler_msIdentifierStruct());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents", new Handler_msContents());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions", new Handler_handDesc());  // wait until additions to process handDesc
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc", new Handler_decoDesc());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history", new Handler_history());
            pi.handler("TEI.teiHeader.profileDesc", new Handler_profileDesc());
            pi.handler("TEI.facsimile", new Handler_facsimile());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist", new Handler_recordHist());
            pi.handler("TEI.text", new Handler_transcription());
        
    }
    
    protected void displayError(String message)
    {
        try {
            sivaErrorReport.write(String.format("%s (%d, %d): %s\n",
                        theFileName, locator.getLineNumber(), debugging, message));
        } catch (Exception ex) {
            System.out.format("Error in displayError: %s (%d, %d): %s\n",
                        theFileName, locator.getLineNumber(), debugging, message);

        }
    }

    public static void writeAllPaths(BufferedWriter bw) throws Exception
    {
        PathInfo.writeStatistics(bw);
    }


    protected String getPath()
    {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (String name : path)
        {
            sb.append(String.format("%s%s", delim, name));
            delim = ".";
        }

        return sb.toString();
    }


    protected void pushPath(String qname)
    {
        path.add(qname);
    }

    protected String popPath()
    {
        int top = path.size()-1;
        String s = path.get(top);
        path.remove(top);
        return s;
    }

    // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        String message = String.format("%d:%d %s\n", locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        displayError(message);
        throw exception;
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
    }

    @Override
    public abstract void endDocument() throws SAXException;
    {
    }

    protected String transactionFixup (String s) {
        while (true) {
            int len = s.length();
            s = s.replaceAll("\\.\\.\\.<br/>\\s*\\.\\.\\.<br/>", "...<br/>");
            if (s.length() == len) {
                break;
            }
        }

        return s;
    }

    protected String doMappings() {
        StringBuilder sb = new StringBuilder();
        Enumeration<String> en = zoneMap.keys();
        ArrayList<String> a = new ArrayList<String>();
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            a.add(key);
        }
        Collections.sort(a);
        sb.append("<script type='text/javascript'>");
        sb.append("var facsmap = {\n");
        String delim = "\t";
        for (String key : a) {
            sb.append(String.format("%s'%s':'%s'", delim, key, zoneMap.get(key)));
            delim = ",\n\t";
        }
        sb.append("};\n");
        sb.append("</script>");
        return sb.toString();
    }

    protected String fixup(String s)
    {
        s = s.replaceAll("& ", "&amp; ");
        s = s.replaceAll("--", "&ndash;");
        return s;

    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public abstract void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException;

    @Override
    public abstract void endElement(String uri, String localName, String qName) throws SAXException;


    @Override
    public abstract void characters(char[] ch, int start, int length) throws SAXException;

    protected void writeMARCIdentifier(String type, String idno) throws Exception
    {
        String margin = getXMLIndentation(tagLevel-1);
        String margin2 = margin + "  ";
        insertBuffer.append(String.format("\n%s<altIdentifier type='%s'>\n%s<idno>%s</idno>\n%s</altIdentifier>",
                            margin, type, margin2, idno, margin));
    }

    protected void trimRight()
    {
        int n = insertBuffer.length() - 1;
        while (n >= 0)
        {
            if (insertBuffer.charAt(n) <= 0x20)
                insertBuffer.deleteCharAt(n--);
            else
                break;
        }
    }

    protected boolean containsEOL(String qName)
    {
        int namelen = qName.length();
        int n = insertBuffer.length() - 1;
        while (n > 0)
        {
            char ch = insertBuffer.charAt(n);
            if (ch == '\n')
                return true;
            else if (ch == '<')         // check if we are at the start tag <qName>
            {
                if (n > namelen)        // just to be safe
                {
                    int i = n+1;          // start of tag name
                    boolean found = true;
                    for (int j=0; j<namelen; j++, i++)
                    {
                        if (insertBuffer.charAt(i) != qName.charAt(j))      // we haven't found the start tag
                        {
                            found = false;
                            break;
                        }
                    }
                    if (found)              
                    {
                        ch = insertBuffer.charAt(i);
                        if (ch == ' ' || ch == '>')
                            return false;    // if we have found the start tag, this means that we did not encounter any \n chars
                        // otherwise, keep going.
                    }
                }
                else
                    return false;           // shouldn't ever reach here.
            }
            n--;
        }
        
        return false;       // to make the compiler happy
    }

    protected String getXMLIndentation(int level)
    {
        return xmlindent.substring(0, level*2);
    }
    

    public void charactersExtract(char[] ch, int start, int length) throws Exception
    {
    }

    @Override
    public abstract void comment(char[] ch, int start, int length) throws SAXException;
    

    @Override
    public abstract void processingInstruction(String target, String data) throws SAXException;

    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
    }



    protected void reserveEntries()
    {
        // Identifiers (handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier is reached)
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier");         // mixed bag

        // Contents (handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents is reached)
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents");

	// handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions is reached.
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions");      // specialPara

	// handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc is reached
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='color']");    // specialPara
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='border']");   // specialPara
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='illustration']"); // specialPara

        // Physical description (handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc is reached)
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc");        // mixed bag
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.bindingDesc");        // bunch of conditions
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.sealDesc");        // bunch of conditions
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.accMat");            // specialPara
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc");       // bunch


        // History (handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.history is reached)
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.origin");          // specialPara
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.provenance");      // specialPara
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.acquisition");     // specialPara

        // Subject headings (handler called when end element TEI.teiHeader.profileDesc is reached)
        pi.put("TEI.teiHeader.profileDesc");

        // Facsimile (handler called when end element TEI.facsimile is reached)
        pi.put("TEI.facsimile");

        // Transcription (handler called when end element TEI.text is reached)
        pi.put("TEI.text");         // kitchen sink

        // Change (Handler called when end element of TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist is reached)
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist");
    }


    public void writeRow(StringBuilder sb, String value)
    {
        sb.append(value);
    }

    public void writeRow(StringBuilder sb, String type, String value, int level)
    {
        if (type.length() == 0)
            sb.append(String.format(rowTemplateNoLabel, value));
        else if (value.trim().length() > 0) {
            String indentation = getHTMLIndentation(level);
            sb.append(String.format(rowTemplate, indentation, type, value));
        }
    }

    private String getHTMLIndentation(int level)
    {
        return indent.substring(0, 6*3*level);
    }

    public void writeRow(StringBuilder sb, String type, String value)
    {
        writeRow(sb, type, value, 0);
    }

    public void writeRowHeader(StringBuilder sb, String type, String value)
    {
        if (type.length() == 0)
            sb.append(String.format(rowTemplateNoLabelHeader, value));
        else if (value.trim().length() > 0)
            sb.append(String.format(rowTemplateHeader, type, value));
    }

    public void writeXML(Node node, StringBuilder sb) throws Exception
    {
        String tagName = writeStartTag(node, sb);
        NodeList list = node.getChildNodes();
        int len = list.getLength();
        if (len > 0)
            sb.append("&gt;");
        for (int i=0; i<len; i++)
        {
            writeXML(list.item(i), sb);
        }
        if (tagName != null)
        {
            if (len == 0)
                sb.append("/&gt;");
            else
                sb.append(String.format("&lt;/%s&gt;", tagName));
        }
    }

    public void writeXML (Node node, StringBuilder sb, String backgroundColor, String textColor) throws Exception {
        sb.append(String.format("<span style='background-color:%s; color:%s'>", backgroundColor, textColor));
        writeXML(node, sb);
        sb.append("</span>");
    }


    public String writeStartTag(Node node, StringBuilder sb)
    {
        if (node.getNodeType() == Node.TEXT_NODE)
        {
            sb.append(node.getTextContent());
            return null;
        }
        else
        {
            sb.append(String.format("&lt;%s", node.getNodeName()));
            NamedNodeMap atts = node.getAttributes();
            int len = atts.getLength();
            for (int i=0; i<len; i++)
            {
                Node att = atts.item(i);
                sb.append(String.format(" %s='%s'",
                        att.getNodeName(), att.getNodeValue()));
            }
            return node.getNodeName();
        }
    }


    String getAttributeValue(XML xml, String path, String attname)
    {
        String value = null;
        try
        {
            Node node = xml.getNode(path);
            value = getAttributeValue(node, attname);
        }
        catch (Exception ex)
        {
            displayError(ex.getMessage());
        }
        
        return value;
    }

    String getAttributeValue(Node node, String attname)
    {
        String value = null;
        try
        {
            if (node != null)
            {
                NamedNodeMap atts = node.getAttributes();
                Node att = atts.getNamedItem(attname);
                if (att != null)
                    value = att.getNodeValue();
            }
        }
        catch (Exception ex)
        {
            displayError(String.format("Error in getAttributeValue: %s", ex.getMessage()));
        }

        return value;
    }

    void append (StringBuilder sb, String s) {
        if (s.trim().startsWith("<br/>")) {
            if (lastAppend.endsWith("<br/>")) {
                s = s.substring(5);     // remove break (may result in empty string but this will be caught later)
            } else {
                reTD.reset(lastAppend);

                if (reTD.find()) {          // don't do a break at the beginning of <td>
                    s = s.substring(5);     // remove break (may result in empty string but this will be caught later)
                } else {
                    reTable.reset(lastAppend);
                    if (reTable.find()) {
                        s = s.substring(5);     // remove break (may result in empty string but this will be caught later)
                    } else {
                        reHeader.reset(lastAppend);
                        if (reHeader.find()) {
                            s = s.substring(5);
                        }
                    }
                }
            }
        }

        if (s.equals("!!!")) {           // get rid of trailing white space before a lb with break='no'
            int last = sb.length() - 1;
            int n = 0;
            for (int j=last; j>=0; j--) {
                if (Character.isWhitespace(sb.charAt(j))) {
                    n++;
                } else {
                    break;
                }
            }
            if (n > 0) {
                sb.setLength(sb.length()-n);
                if (sb.length() > 0 && sb.charAt(sb.length()-1) == ':') {  // hack, don't get rid of space after Line n:
                    sb.append(" ");
                }
            }
        } else if (s.length() > 0) {
            if (Character.isWhitespace(s.charAt(0))) {
                String bfr = sb.toString();
                if (bfr.length() > 6) {
                    if (bfr.endsWith("&nbsp;")) {
                        s = (s + '.').trim();
                        s = s.substring(0, s.length() - 1);
                    }
                }
            }
            sb.append(s);
            s = s.trim();
            if (s.length() > 0 && !s.startsWith("<a name")) {           // don't let name anchor affect breaks
                lastAppend = s;
            }
        }
    }

    void emitBreak (StringBuilder sb) {
        if (lastAppend.length() > 0 && !lastAppend.endsWith("<br/>")) {
            char ch = lastAppend.charAt(lastAppend.length()-1);
            if (ch != '.' && ch != ',') {
                append(sb, "<br/>");
            }
        }
    }

    private boolean checkDebugging (int n) {
        if (n == 84) {
            n = 84;
            return true;
        } else {
            return false;
        }
    }

    String getMarkedText(Node node, String parentlang, int level, String context) {
        if (node == null) {
            return "";
        }

        String closingElement = null;
        String nn = node.getNodeName();
        if (ParseLayout.expectedTags.get(nn) == null) {
            if (errorTags.get(nn) == null) {
                System.out.format("Unexpected tag: %s\n", nn);
                errorTags.put(nn, Boolean.TRUE);
            }
            return "";
        }

        boolean isFragment = false;
        String subContext = "";
        String lastseg = "";
        StringBuilder sb = new StringBuilder();
        MarkupInfo markupInfo = null;

        try {
            checkDebugging(++debugging);

            isFragment = node.getNodeName().equals("#document");        // if called from a handler; otherwise it is a recursive call.

            // emit span if node has a rend attribute that we respond to
            NamedNodeMap atts = node.getAttributes();
            Node rendAttribute = null;
            if (atts != null) {
                rendAttribute = atts.getNamedItem("rend");
            }
            boolean isSpan = false;
            if (rendAttribute != null) {
                String rend = rendAttribute.getNodeValue();
                if (rendAttributeValues.get(rend) != null) {
                    RenderType renderType = rendAttributeValues.get(rend);
                    if (renderType == RenderType.COLOR) {
                        isSpan = true;
                        append(sb, String.format("<span style='color:%s'>", rend));
                    } else if (renderType == RenderType.ITALIC) {
                        isSpan = true;
                        append(sb, String.format("<span style='font-style:italic'>"));
                    } else if (renderType == RenderType.SMALL) {
                        isSpan = true;
                        append(sb, String.format("<span style='font-size:10pt'>"));
                    } else if (renderType == RenderType.BOLD) {
                        isSpan = true;
                        append(sb, "<span style='font-weight:bold'>");
                    } else if (renderType == RenderType.SMALLCAP) {
                        append(sb, "<h4>");
                        closingElement = "</h4>";
                    }
                }
            }


            String lang = getLanguage(node, parentlang);

            // Now process each child node
            NodeList list = node.getChildNodes();
            int len = list.getLength();
            for (int i = 0; i < len; i++) {
                checkDebugging(++debugging);
                node = list.item(i);
                markupInfo = null;
                // Insert a / between choices if necessary
                if (node.getNodeValue() == null && level == choiceLevel) {
                    if (firstChoice) {
                        firstChoice = false;
                    } else {
                        append(sb, ("/"));
                    }
                }

                String _lang = getLanguage(node, lang);
                String nodeName = node.getNodeName();
                if (nodeName.equals("title")) {
                    nodeName = "title";
                }
                boolean isTextNode = node.getNodeType() == Node.TEXT_NODE;

                if (!isTextNode) {      // see if anchor needs to be generated and set subContext to the path for this child node
                    String id = getAttributeValue(node, "xml:id");
                    if (id != null && !ParseLayout.styleOverridden) {
                        append(sb, String.format("<a name='%s'></a>", id));
                    }

                    if (isFragment) {
                        subContext = context;   // have #document so first child name already included in passed in context
                    } else {
                        subContext = String.format("%s.%s", context, node.getNodeName());
                    }

                    // See if markupInfo object for this node exists
                    checkDebugging(debugging);
                    if (ParseLayout.ignoreElement(node, subContext, style)) {
                        continue;
                    }
                    markupInfo = ParseLayout.getMarkup(node, subContext, style);
                } else if (isTextNode) {            // otherwise this is text. Convert it if necessary and append to result
                    String text = node.getNodeValue();
                    if (lang != null) {
                        text = Main.transcode(text, lang);
                    }
                    append(sb, text);
                    markupInfo = null;
                    continue;
                }

                // Reach here if we have a tag
                if (ParseLayout.expectedTags.get(nodeName) == null) {
                    if (errorTags.get(nodeName) == null) {
                        System.out.format("Unexpected tag: %s\n", nodeName);
                        errorTags.put(nodeName, Boolean.TRUE);
                    }
                    markupInfo = null;
                    continue;
                }
                
                if (markupInfo != null) {
                    if (markupInfo.before.size() > 0) {
                        append(sb, markupInfo.format(pi, style, node, markupInfo.before, subContext));
                    }

                    if (markupInfo.format.size() > 0) {
                        append(sb, markupInfo.format(pi, style, node, markupInfo.format, subContext));
//                        emitEnding(sb, isSpan, closingElement);
                        markupInfo = null;
                        continue;
                    }
                }

                /* ************************
                *
                * Reach here if special processing has to be done to node
                *
                ***************************/
                if (nodeName.equals("bibl")) {
                    append(sb, get_bibl(node, _lang, level, subContext));
                } else if (nodeName.equals("ref")) {
                    processRef(node, sb, lang, level, subContext);
                } else if (nodeName.equals("locus")) {
                    LocusInfo li = get_locus(node, "eng"); // ! force language to be english
                    if (li.locus.trim().length() > 0) {
                        append(sb, li.buildLink());
                    }
                } else if (nodeName.equals("seg")) {
                    String seg = getMarkedText(node, lang, level + 1, subContext);
                    append(sb, seg);
                    if ("aAiIuUfFxXeEoOMH".contains(lastseg)) {
                        append(sb, " ");
                    }
                    lastseg = seg.substring(seg.length() - 1);
                } else if (nodeName.equals("pb")) {
                    String type = getAttributeValue(node, "type");
                    if (type == null || type.equals("folio")) {
                        String facs = getAttributeValue(node, "facs");
                        if (facs != null && facs.length() > 1) {
                            facs = facs.substring(1);       // move past #
                            String text = facs.replace("f", "f.");
                            append(sb, _buildLink(facs, text, "green"));
                        }
                    }
                } else if (nodeName.equals("cb")) {
                    int last = inList.size() - 1;
                    if (last >= 0) {
                        int col = inList.get(last);
                        if (col > 0) {
                            append(sb, "</td>");
                        }
                        inList.set(last, col + 1);
                        append(sb, "<td valign='top'>");
                    } else {
                        emitBreak(sb);
                    }
                    columnCount++;
                } else if (nodeName.equals("list")) {
                    inList.add(0);
//                    Style saveStyle = style;
//                    style = Style.MANUSCRIPT_LAYOUT;
                    append(sb, "<table border='1' style='border-collapse:collapse;border:0px;margin:0px;'>\n<tr>\n");
                    append(sb, getMarkedText(node, _lang, level + 1, subContext));
                    append(sb, "</td></tr>\n</table>\n");
                    inList.remove(inList.size() - 1);
//                    style = saveStyle;
                } else if (nodeName.equals("choice")) {
                    choiceLevel = level + 1;
                    firstChoice = true;
                    append(sb, getMarkedText(node, _lang, level + 1, subContext));
                    choiceLevel = -1;
                } else if (nodeName.equals("dimensions")) {
                    append(sb, formatDimensions(node));
                } else {
                    append(sb, getMarkedText(node, _lang, level + 1, subContext));     // just use formatting of children
                }
                if (markupInfo != null) {
                    if (markupInfo.after.size() > 0) {
                        append(sb, markupInfo.format(pi, style, node, markupInfo.after, subContext));
                    }
                    markupInfo = null;
                }
            }

            emitEnding(sb, isSpan, closingElement);
        } catch (Exception ex) {
            displayError(String.format("Error in getMarkedText: %s", ex.getMessage()));
        }
        
        String text = sb.toString();
        if (text.contains("The addition covers")) {
            return text;
        }
        return text;
    }

    private void emitEnding(StringBuilder sb, boolean isSpan, String endElement) {
        if (isSpan) {
            append(sb, "</span>");
        } else if (endElement != null) {
            append(sb, endElement);
        }
    }

    private void savePath (Node node, Hashtable<String, String> h, String context) {
        if (h.get(context) == null) {
            h.put(context, theFileName);
        }
    }
    
    private String formatDimensions (Node node) {
        StringBuilder sb = new StringBuilder();
        String unit = "";
        if (node.getAttributes().getNamedItem("unit") != null) {
            unit = node.getAttributes().getNamedItem("unit").getNodeValue();
        }
        sb.append("Dimensions: ");
        NodeList list = node.getChildNodes();
        int len = list.getLength();
        String delim = "";
        for (int i=0; i<len; i++) {
            Node _node = list.item(i);
            if (_node.getNodeType() != Node.TEXT_NODE) {
                sb.append(delim);
                delim = ", ";
                sb.append(String.format("%s %s%s", _node.getNodeName().trim(), _node.getTextContent(), unit));
            }
        }
        sb.append(".");
        return sb.toString();
    }


    private void processRef(Node node, StringBuilder sb, String parentlang, int level, String context) throws Exception {
        Node tn = node.getAttributes().getNamedItem("target");
        if (tn != null) {
            String tgt = tn.getNodeValue();
            if (tgt.equals("#s7.i")) {
                tgt = "#s7.i";
            }
            if (tgt != null && tgt.length() > 0) {
                if (tgt.startsWith(Main.textSubdirectory)) {
                    reText.reset(tgt);
                    if (reText.find()) {
                        tgt = buildTextLink(reText.getGroup(1), reText.getGroup(2));
                    } else {
                        displayError(String.format("ref target '%s' has unknown text directory. Used as is\n", tgt));
                    }
                } else if (tgt.startsWith(Main.xmlSubdirectory)) {
                    Regex re = new Regex(tgt, Main.xmlSubdirectory + "(.*)");
                    if (re.find()) {
                        tgt = buildHTMLLink(tgt, re);
                    } else {
                        displayError(String.format("ref target '%s' has unknown xml directory. Used as is\n", tgt));
                    }
                } else if (tgt.startsWith("http:")) {
                    //tgt = String.format("javascript:displayAbsoluteURL(\"%s\")", tgt);
                    tgt = String.format(tgt);
                } else if (tgt.charAt(0) != '#') {
                    if (missingPoundSign.get(theFileName) == null) {
                        missingPoundSign.put(theFileName, Boolean.TRUE);
                        displayError(String.format("At least one ref target '%s' is missing #. One added.", tgt));
                    }
                    tgt = "#" + tgt;
                }
            }
            Node lang = node.getAttributes().getNamedItem("xml:lang");
            if (lang != null) {
                parentlang = lang.getNodeValue();
            }
            String linktext = getMarkedText(node, parentlang, level, context);      // in case content is structured.
//            if (lang != null && lang.getNodeValue().equals("sa-Latn-x-SLP1")) {
//                linktext = pi.transcode(linktext, lang.getNodeValue());
//            }

            if (tgt.charAt(0) == '#') {
                sb.append(String.format("<a href='javascript:localReference(\"%s\")'>%s</a>", tgt, linktext));
                //sb.append(String.format("<a href='%s'>%s</a>", tgt, linktext));
            } else if (!tgt.contains("javascript:")){
                sb.append(String.format("<a target='text' href='%s'>%s</a>", tgt, linktext));
            } else {
                sb.append(String.format("<a target='broken' style='color:red' href='%s'>%s</a>", tgt, linktext));
            }
        } else {
            displayError("Missing or empty target attribute of ref element. Search for <ref ...");
            writeXML(node, sb, "red", "white");
        }

    }

    private String buildHTMLLink(String tgt, Regex re) {
        reHTML.reset(tgt);
        if (reHTML.find()) {
            if (reHTML.getGroup(2) != null) {
                tgt = String.format("%s?html=%s%s",
                        Main.htmlServer, reHTML.getGroup(1), reHTML.getGroup(2));
            } else {
                tgt = String.format("%s?html=%s",
                        Main.htmlServer, reHTML.getGroup(1));
            }
        } else if (re != null) {
            tgt = String.format("javascript:displayHTML(\"%s\")", re.getGroup(1).replaceAll("xml", "html"));
        } else {
            tgt = String.format("javascript:displayHTML(\"%s\")", tgt);
        }

        return tgt;

    }

    private boolean isQuotable(Node node)
    {
        NodeList list = node.getChildNodes();
        int len = list.getLength();
        boolean hasTitle= false;
        boolean hasParagraph = false;
        for (int i=0; i<len && (!hasTitle || !hasParagraph); i++)
        {
            Node _node = list.item(i);
            String nodeName = _node.getNodeName();
            if (nodeName.equals("title"))
                hasTitle = true;
            else if (nodeName.equals("p"))
                hasParagraph = true;
        }

        return !(hasTitle && hasParagraph);     // if it doesn't have a title and paragraph within it is quotable
    }

    boolean nodeExists(Node node, String name) throws Exception {
       NodeList _list = XML.getRelativeNodeList(node, name);
       return (_list != null && _list.getLength() > 0 && _list.item(0).getChildNodes().getLength() > 0);
    }


    String getLanguage(Node node, String lang)
    {
        NamedNodeMap atts = node.getAttributes();
        if (atts != null && atts.getNamedItem("xml:lang") != null)
            return atts.getNamedItem("xml:lang").getNodeValue();
        else
            return lang;        
    }

    ArrayList<StringBuilder> get_seal_or_bindingDesc(Node node, String type, String parentlang, int level, String context)
    {
        ArrayList<StringBuilder> a = new ArrayList<StringBuilder>();
        StringBuilder workingElement = new StringBuilder();
        String lang = getLanguage(node, parentlang);
        String nodeName = node.getNodeName();
        workingElement.append(String.format("%c%s ",
                Character.toUpperCase(nodeName.charAt(0)),
                nodeName.substring(1)));
        a.add(workingElement);
        NodeList list = node.getChildNodes();
        int len = list.getLength();
        for (int i=0; i<len; i++)
        {
            Node _node = list.item(i);
            if (_node.getNodeType() == Node.TEXT_NODE)          // shouldn't have anything except whitespace
                workingElement.append(_node.getTextContent());
            else
            {
                String _lang = getLanguage(_node, lang);
                nodeName = _node.getNodeName();
                String subContext = String.format("%s.%s", context, nodeName);
                if (nodeName.equals("p"))
                    workingElement.append(getMarkedText(_node, _lang, level+1, subContext));
                else if (nodeName.equals("condition"))
                {
                    workingElement = new StringBuilder("Condition   ");
                    a.add(workingElement);
                    String condition = getMarkedText(_node, _lang, level+1, subContext); 
                    if (condition.trim().length() > 0)
                    {
                        emitBreak(workingElement);
                        workingElement.append(String.format("%s", condition));
                    }
                }
                else
                    workingElement.append(getMarkedText(_node, _lang, level+1, subContext));
            }
        }
        
        return a;
    }

    ArrayList<StringBuilder> get_foliation(Node node, String lang, int level, String context)
    {
        ArrayList<StringBuilder> a = new ArrayList<StringBuilder>();
        StringBuilder workingElement = new StringBuilder();
        String _lang = getLanguage(node, lang);
        NodeList list = node.getChildNodes();
        int len = list.getLength();
        String id = getAttributeValue(node, "xml:id");
        String nodeName = node.getNodeName();
        workingElement.append(String.format("%c%s ",
                Character.toUpperCase(nodeName.charAt(0)),
                nodeName.substring(1)));
        a.add(workingElement);
        if (id == null)
            id = "";
        for (int i=0; i<len; i++)
        {
            Node _node = list.item(i);
            String subContext = String.format("%s.%s", context, node.getNodeName());
            if (_node.getNodeType() == Node.TEXT_NODE)
            {
                String text = _node.getTextContent().trim();
                if (text.length() > 0) {
                    if (lang != null) {
                        text = Main.transcode(text, _lang);
                    }
                    workingElement.append(" ");
                    workingElement.append(text);
                    workingElement.append(" ");
                }
            }
            else if (_node.getNodeName().equals("formula"))
            {
                workingElement = new StringBuilder("Formula ");
                a.add(workingElement);
                String formula = getMarkedText(_node, _lang, level+1, subContext); 
                if (formula.trim().length() > 0)
                {
                    workingElement.append(String.format("%s%s ", id.length()==0?"":String.format("(%s) ", id), formula));
                }
            }
            else if (_node.getNodeName().equals("signatures")) {
                workingElement = new StringBuilder("Signatures ");
                a.add(workingElement);
                String mt = getMarkedText(_node, _lang, level+1, subContext).trim();  
                if (mt.length() > 0) {
                        workingElement.append(String.format("%s ", mt));
                }
            } else if (_node.getNodeName().equals("locus")) {
                workingElement.append(get_locus(_node, "eng").buildLink());
            }else if (_node.getNodeName().equals("ref")) {
                StringBuilder sb = new StringBuilder();
                try {
                    processRef(_node, sb, lang, level+1, subContext);
                    workingElement.append(sb.toString());
                } catch (Exception ex) {
                    workingElement.append(String.format(" %s %s ",
                        _node.getNodeName(), getMarkedText(_node, _lang, level+1, subContext)));
                }
            } else {
                workingElement.append(String.format(" %s %s ", 
                        _node.getNodeName(), getMarkedText(_node, _lang, level+1, subContext))); 
            }
        }

        return a;
    }


    LocusInfo get_locus(Node node, String parentlang)
    {
        LocusInfo li =  new LocusInfo();
        String lang = getLanguage(node, parentlang);

        li.scheme = getAttributeValue(node, "scheme");
        li.target = getAttributeValue(node, "target");
        li.from = getAttributeValue(node, "from");
        li.to = getAttributeValue(node, "to");
        li.setfacs(getAttributeValue(node, "facs"));
        li.locus = get_xtext(node, lang);

        if (li.from != null) {
            li.from = TextFormatter.normalizeID(li.from);
        }

        if (li.to != null) {
            li.to = TextFormatter.normalizeID(li.to);
        }

        if (li.scheme != null) {
            li.locus = String.format("(%s foliation scheme) %s", li.scheme.substring(1), li.locus);
        }
        
        return li;
    }


    String get_bibl(Node node, String parentlang, int level, String context) {
        StringBuilder sb = new StringBuilder();
        String corresp = null;
        try {
            parentlang = "eng"; // getLanguage(node, parentlang);
            Node correspAttr = node.getAttributes().getNamedItem("corresp");
            if (correspAttr != null) {
                corresp = correspAttr.getNodeValue().trim();
            }
            Node xmllangAttr = node.getAttributes().getNamedItem("xml:lang");
            if (xmllangAttr != null) {
                parentlang = xmllangAttr.getNodeValue();
            }

            NodeList list = node.getChildNodes();
            int len = list.getLength();
            for (int i = 0; i < len; i++) {
                Node _node = list.item(i);
                if (_node.getNodeType() == Node.TEXT_NODE) {            // e.g. MBh
                    String text = _node.getTextContent();       // .trim();
                    if (text.length() > 0) {
                        if (parentlang != null) {
                            text = Main.transcode(text, parentlang);
                        }
                        if (corresp != null) {
                            String tgt = buildHTMLLink(corresp, null);
                            append(sb, String.format("<a target='html' href='%s'>%s</a>", tgt, text));
                        } else {
                            append(sb, text);
                        }
                    }
                } else if (_node.getNodeName().equals("biblScope")) {
                    String from = getAttributeValue(_node, "from");
                    if (from == null) {
                        from = getAttributeValue(_node, "to");
                    }
                    String value = _node.getTextContent().trim();
                    if (from != null) {
                        // from is editions/purana/bhp/iskon/Sb11_0002.tif
                        // value is vol. 11, p. 2 ad 11.1.1, -- p. 800 ad 11.18.38
                        if (from.startsWith("editions")) {// lg or nexus 4 www.lg.com 800-243-0000
                            from = from.substring(0, from.length()-4);
                            append(sb, String.format("<a target='image' href='%s?name=%s.jpg'>%s</a>",
                                    Main.manuscriptServer, from, value));
                        } else {
                            reText.reset(from);
                            if (reText.find()) {
                                String tgt = buildTextLink(reText.getGroup(1), reText.getGroup(2));
                                append(sb, String.format(" <a target='text' href='%s'>%s</a> ",
                                    tgt, value));      // force language to be english
                            } else {
                                reTextNoType.reset(from);
                                if (reTextNoType.find()) {
                                    String tgt = buildTextLink(reTextNoType.getGroup(1), reTextNoType.getGroup(2));
                                    append(sb, String.format(" <a target='text' href='%s'>%s</a> ",
                                            tgt, value));      // force language to be english
                                } else {
                                    append(sb, String.format(" <a target='broken' style='color:red' href='javascript:displayText(\"%s\")'>%s</a> ",
                                        from, value));      // force language to be english
                                }
                            }
                        }
                    } else {
                        append(sb, value);   // just display value, do not put it in a link
                    }
                } else {
                    String subContext = String.format("%s.%s", context, _node.getNodeName());
                    String _lang = getLanguage(_node, parentlang);
                    append(sb, getMarkedText(_node, _lang, level + 1, subContext));
                }

            }
        } catch (Exception ex) {
            displayError(String.format("Error in get_bibl: %s", ex.getMessage()));
        }

        String s = sb.toString().trim();
        if (corresp == null && s.length() > 0) {
            s = String.format("(%s)", s);
        }
        return s;
    }


    String get_dimensions(Node node, String lang) throws Exception
    {
        String _lang = getLanguage(node, lang);
        String unit = getAttributeValue(node, "unit");
        if (unit == null)
            unit = "?";
        Node _node = XML.getRelativeNode(node, "height");
        String height = get_xtext(_node, _lang);        // assume that language is the same
        _node = XML.getRelativeNode(node, "width");
        String width = get_xtext(_node, _lang);

        return String.format("%s x %s %s (h x w)", height, width, unit);
    }

/*
    String get_gap(Node node, String lang, int level, String context) throws Exception {
        StringBuilder sb = new StringBuilder();
        String quantity = getAttributeValue(node, "quantity");

        if (quantity != null) {
            String unit = getAttributeValue(node, "unit");
            String n = getAttributeValue(node, "n");
            String corresp = getAttributeValue(node, "corresp");
            if (quantity.equals("1")) {
                sb.append(String.format("missing 1 %s, ", unit));
            } else {
                if (unit.toLowerCase().equals("leaf")) {
                    sb.append(String.format("missing %s leaves, ", quantity));
                } else {
                    sb.append(String.format("missing %s %ss, ", quantity, unit));
                }
            }
            Node _node = XML.getRelativeNode(node, "desc");
            if (_node != null) {
                String _lang = getLanguage(node, lang);
                sb.append(getMarkedText(_node, _lang, level + 1, context + ".desc"));
            } else {
                sb.append("?");
            }
        }


        return sb.toString();
    }
*/

    String get_xtext(XML xml, String path, String parentlang)
    {
        String value = "";
        try
        {
            Node node = xml.getNode(path);
            value = get_xtext(node, parentlang);
        }
        catch (Exception ex)
        {
            displayError(String.format("Error in get_xtext: %s", ex.getMessage()));
        }

        return value;
    }

    String get_xtext(Node node, String lang)
    {
        boolean unknown = false;
        StringBuilder sb = new StringBuilder();
        try
        {
            if (node != null)
            {
                lang = getLanguage(node, lang);

                NodeList list = node.getChildNodes();
                int len = list.getLength();
                for (int i=0; i<len; i++)
                {
                    node = list.item(i);
                    if (node.getNodeType() == Node.TEXT_NODE)
                    {
                        String text = node.getNodeValue();
                        if (lang != null) {
                            text = Main.transcode(text, lang);
                        }
                        sb.append(text);
                    } else {
                        unknown = true;
                        writeXML(node, sb);         // will presumably be a <g>...</g>
                    }
                }
            }
        }
        catch (Exception ex)
        {
            displayError(String.format("Error in get_xtext: %s", ex.getMessage()));
        }

        if (unknown)
            displayError(String.format("In get_xtext, unexpected element: %s", sb.toString()));
        return sb.toString();
    }


    // Information about second folio and signatures is here
    void witnessDetail(StringBuilder sb, Node node, String label, int level, int levelx, String context) throws Exception
    {
        NodeList _list = XML.getRelativeNodeList(node, "witDetail");        // these are like notes
        if (_list != null)
        {
            int _len = _list.getLength();
            for (int j=0; j<_len; j++)
            {
                Node _node = _list.item(j);
                Node secFol = XML.getRelativeNode(_node, "secFol");
                String val = getMarkedText(secFol, getLanguage(secFol, null), levelx, context + ".secFol"); 
                writeRow(sb, "Second&#160;folio", val, level);

                Node signatures = XML.getRelativeNode(_node, "signatures");
                val = getMarkedText(signatures, getLanguage(signatures, null), levelx, context + ".signatures"); 
                writeRow(sb, "Signatures", val, level);
            }
        }
    }

    void contentSection(StringBuilder sb, Node node, String context, String path, String label, int level) throws Exception {
        Style oldStyle = style;
        String key = path;
        if (key == null) {
            key = node.getNodeName();
        }
        style = ParseLayout.overrideStyle(key, style);        // if style is overridden then no anchors will be generated
        if (path != null) {
            NodeList _list = XML.getRelativeNodeList(node, path);
            if (_list != null) {
                int _len = _list.getLength();
                for (int j = 0; j < _len; j++) {
                    Node _node = _list.item(j);
                    if (path.equals("rubric")) {
                        Node type = _node.getAttributes().getNamedItem("type");
                        if (type != null) {
                            label = String.format("%s&#160;rubric", capitalizeFirstLetter(type.getNodeValue()));
                        }
                    }
                    _contentSection(sb, _node, label, level, String.format("%s.%s", context, _node.getNodeName()));
                }
            }
        } else {
            _contentSection(sb, node, label, level, String.format("%s.%s", context, node.getNodeName()));
        }

        ParseLayout.styleOverridden = false;        // in case style was overridden.
        style = oldStyle;       // in case it was changed
    }

    void _contentSection(StringBuilder sb, Node _node, String label, int level, String context) throws Exception {
        String val = getMarkedText(_node, null, 0, context); 
        String name = _node.getNodeName();
        String refatt = References.hasInplaceAttribute.get(name);
        if (refatt != null) {
            Node tn = _node.getAttributes().getNamedItem(refatt);
            if (tn != null) {
                String value = tn.getNodeValue();
                String[] attvalues = value.split("\\s+");
                if (attvalues.length > 1) {
                    checkDebugging(++debugging);
                }
                for (String attvalue : attvalues) {
                    XMLChunk chunk = References.xmlidContent.get(String.format("%s%s", theFileName, attvalue));
                    if (chunk == null) {
                        displayError(String.format("%s is not accessible", attvalue));
                    } else {
                        StringReader sr = new StringReader(chunk.xml);
                        XML xml = new XML(sr);
                        Node refnode = xml.getNode("/");        // .getFirstChild();
                        val += getMarkedText(refnode, chunk.language, 0, chunk.path); 
                    }
                }
            }
        }

        writeRow(sb, label, val, level);
    }


    public void appendBuffer(String s)
    {
        currentBuffer.append(s);
    }


    public String getCharacters()
    {
        int top = languageStack.size() - 1;
        if (top >= 0)
        {
            String language = languageStack.get(languageStack.size()-1);
            String s = currentBuffer.toString();
            currentBuffer.setLength(0);
            return Main.transcode(s, language);
        }
        else
            return "";          // probably an attempt to get characters before the root element (there aren't any).
    }

        // I. Manuscript Identifiers section
    class Handler_msIdentifierStruct implements ElementHandler
    {

        public void process(String path)
        {
            String val =  null;
            StringBuilder sb = new StringBuilder();

            try
            {
            // Collection
            Node identifierNode = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier");
            if (identifierNode == null)
                return;

            Node _node = XML.getRelativeNode(identifierNode, "collection");
            if (_node != null)
            {
                val = get_xtext(_node, null);
                writeRow(sb, "Collection", val);
            }

            // Item
            _node = XML.getRelativeNode(identifierNode, "idno");
            if (_node != null)
            {
                val = get_xtext(_node, null);
                writeRow(sb, "Item", val);
            }
            // Repository
            _node = XML.getRelativeNode(identifierNode, "repository");
            if (_node != null)
            {
                val = get_xtext(_node, null);
                writeRow(sb, "Repository", val);
            }

            // Institution
            _node = XML.getRelativeNode(identifierNode, "institution");
            if (_node != null)
            {
                val = get_xtext(_node, null);
                writeRow(sb, "Institution", val);
            }

            // Location
            String settlement = "?";
            String region = "?";
            String country = "?";
            _node = XML.getRelativeNode(identifierNode, "settlement");
            if (_node != null)
                settlement = get_xtext(_node, null);
            _node = XML.getRelativeNode(identifierNode, "region");
            if (_node != null)
                region = get_xtext(_node, null);
            _node = XML.getRelativeNode(identifierNode, "country");
            if (_node != null)
                country = get_xtext(_node, null);

            if (!settlement.equals("?") || !region.equals("?") || !country.equals("?"))
            {
                val = String.format("%s, %s, %s", settlement, region, country);
                writeRow(sb, "Location", val);
            }

            // Catalog and Item
            NodeList list = XML.getRelativeNodeList(identifierNode, "altIdentifier[@type='catalog']");
            if (list != null)
            {
                int len = list.getLength();
                for (int i=0; i<len; i++)
                {
                    _node = list.item(i);
                    Node nodeCatalog = XML.getRelativeNode(_node, "collection");
                    if (nodeCatalog != null)
                    {
                        val = get_xtext(nodeCatalog, null);
                        writeRow(sb, "Catalog", val);
                    }

                    Node nodeIdno = XML.getRelativeNode(_node, "idno");
                    if (nodeIdno != null)
                    {
                        val = get_xtext(nodeIdno, null);
                        writeRow(sb, "Item", val);
                    }
                }
            }

            if (sb.length() > 0)
            {
                manuscriptIdentifiers.append("<h3>Manuscript Identifiers</h3>\n");
                manuscriptIdentifiers.append("<table>\n");
                manuscriptIdentifiers.append(sb.toString());
                manuscriptIdentifiers.append("</table>\n");
            }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_msIdentifierStruct: %s", ex.getMessage()));
            }
        }
    }



    // Contents section
    class Handler_msContents implements ElementHandler
    {
        public void process(String path)
        {
            int workNumber = 1;
            StringBuilder sb = new StringBuilder();
            try
            {
                String context = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents";
                Node contentNode = pi.getXML(context);
                NodeList list = contentNode.getChildNodes();
                int len = list.getLength();
                for (int i=0; i<len; i++) {
                    Node node = list.item(i);
                    if (node.getNodeName().equals("msItemStruct")) {
                        processItemStruct(node, sb, 0, "", workNumber++, context + ".msItemStruct");
                    } else if (node.getNodeName().equals("msItem")) {
                        processItem(node, sb, 0, "", workNumber++, context + ".msItemStruct");
                    }
                }
                if (sb.length() > 0)
                {
                    contents.append("<h3>Contents</h3>\n");
                    contents.append(sb.toString());
                }


            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_msContents.process", ex.getMessage()));
            }
        }


        String getType(Node node) {
            NamedNodeMap attributes = node.getAttributes();
            if (attributes != null) {
                Node typeAttribute = attributes.getNamedItem("type");
                if (typeAttribute != null) {
                    return typeAttribute.getNodeValue();
                }
            }
            return null;
        }

        private void processItemStruct(Node node, StringBuilder sb, int levelno, String prefix, 
                                        int workNumber, String context)
        {
            try
            {
                if (node != null)
                {
                    if (levelNoLastEndTable > levelno) {
                        sb.append("<br/>");
                    }
                    sb.append("<table>\n");
                    String val = getAttributeValue(node, "defective");
                    boolean complete = false;
                    if (val != null) {
                        complete = val.toLowerCase().equals("false");
                    }

                    String status;
                    if (complete) {
                        status = " (complete)";
                    } else {
                        status = " (incomplete)";
                    }

                    sb.append(String.format(String.format("<h4>%sWork&#160;%s%s</h4>\n", getHTMLIndentation(levelno), String.format("%s%d", prefix, workNumber), status)));
                    // Work number

                    // Locus
                    Node _node = XML.getRelativeNode(node, "locus");
                    if (_node != null)
                    {
                        LocusInfo li = get_locus(_node, "en");
                        String link = li.buildLink();
                        writeRow(sb, "Locus", link, levelno);
                    }

                    // Author
                    _node = XML.getRelativeNode(node, "author");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, 0, context + ".author"); 
                        if (val.trim().length() == 0)
                        {
                            _node = XML.getRelativeNode(node, "author/persName");
                            if (_node != null)
                                val = get_xtext(_node, null);
                        }
                        writeRow(sb, "Author", val, levelno);
                    }

                    // Title
                    _node = XML.getRelativeNode(node, "title[@type='main']");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, 0, context + ".title");
                        writeRow(sb, "Title", val, levelno);
                    }

                    // Part
                    NodeList list = XML.getRelativeNodeList(node, "title[@type='part']");
                    if (list != null) {
                        int len = list.getLength();
                        for (int i=0; i<len; i++) {
                            _node = list.item(i);
                            val = getMarkedText(_node, null, 0, context + ".title");
                            writeRow(sb, "Part", val, levelno);
                        }
                    }

                    // Cover Rubric
                    contentSection(sb, node, context, "rubric", "Rubric", levelno);

                    // Incipit
                    contentSection(sb, node, context, "incipit", "Incipit", levelno);

                    // Explicit
                    contentSection(sb, node, context, "explicit", "Explicit", levelno);

                    // Final rubric
                    contentSection(sb, node, context, "finalRubric", "Final&#160;rubric", levelno);

                    // Colophon
                    if (nodeExists(node, "colophon")) {
                        contentSection(sb, node, context, "colophon", "Colophon", levelno);
                    }

                    // Note
                    contentSection(sb, node, context, "note", "Note", levelno);

                    // Filiation
                    contentSection(sb, node, context, "filiation", "Filiation", levelno);

                    // Second folio
                    witnessDetail(sb, node, "Second folio", levelno, 0, context);

                    // Language
                    _node = XML.getRelativeNode(node, "textLang");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, 0, context + ".textLang");
                        writeRow(sb, "Language", val, levelno);
                    }
                    sb.append("</table>\n");
                    levelNoLastEndTable = levelno;

                    // What to do with witDetail and filiation
                    list = XML.getRelativeNodeList(node, "msItemStruct");      // Note: cannot have an msItem child
                    int len = list.getLength();
                    for (int i=0; i<len; i++)
                    {
                        _node = list.item(i);
                        processItemStruct(_node, sb, levelno+1, String.format("%s%d.", prefix, workNumber), i+1,
                                context + ".msItemStruct");
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Handler_msContents failed with: %s", ex.getMessage()));
            }
        }

        // Similar to processItemStruct except that children can appear in any order.
        private void processItem (Node node, StringBuilder sb, int levelno, String prefix, int workNumber,
                                    String context)
        {
            try
            {
                if (node != null)
                {
                    int subWorkNumber = 1;
                    boolean sectionEnded = false;       // set true when </table> emitted.
                    if (levelNoLastEndTable > levelno) {
                        sb.append("<br/>");
                    }
                    sb.append("<table>\n");
                    String val = getAttributeValue(node, "defective");
                    boolean complete = false;
                    if (val != null) {
                        complete = val.toLowerCase().equals("false");
                    }

                    String status;
                    if (complete) {
                        status = " (complete)";
                    } else {
                        status = " (incomplete)";
                    }

                    sb.append(String.format(String.format("<h4>%sWork&#160;%s%s</h4>\n", getHTMLIndentation(levelno), String.format("%s%d", prefix, workNumber), status)));

                    // Work number
                    NodeList list = node.getChildNodes();
                    int len = list.getLength();

                    for (int i=0; i<len; i++) {
                        Node _node = list.item(i);
                        String nodeName = _node.getNodeName();
                        if (nodeName.equals("#text")) {
                            continue;
                        }
                        String subContext = String.format("%s.%s", context, nodeName);

                        // sectionEnded==true means that we have processed an msItemStruct or msItem
                        // we need to generated another <table> because they generated a </table>
                        if (sectionEnded && !nodeName.equals("msItemStruct") && !nodeName.equals("msItem")) {
                            if (levelNoLastEndTable > levelno) {
                                sb.append("<br/>");
                            }
                            sb.append("<table>\n");
                            sectionEnded = false;
                        }
                        
                        if (nodeName.equals("locus")) {
                            LocusInfo li = get_locus(_node, "en");
                            String link = li.buildLink();
                            writeRow(sb, "Locus", link, levelno);
                        } else if (nodeName.equals("author")) {
                            val = getMarkedText(_node, null, 0, subContext);
                            if (val.trim().length() == 0)
                            {
                                _node = XML.getRelativeNode(_node, "persName");
                                if (_node != null)
                                    val = get_xtext(_node, null);
                            }
                            writeRow(sb, "Author", val, levelno);
                        } else if (nodeName.equals("title")) {
                            String type = getType(_node);
                            String label = null;
                            if (type.equals("main")) {
                                label = "Title";
                             } else if (type.equals("part")) {
                                 label = "Part";
                             }
                             if (label != null) {
                                val = getMarkedText(_node, null, 0, subContext);
                                writeRow(sb, label, val, levelno);
                             }
                        } else if (nodeName.equals("rubric")) {
                            String type = getType(_node);
                            String label = null;

                            if (type != null) {
                                label = String.format("%s&160;rubric", capitalizeFirstLetter(type));
                            } else {
                                label = "Rubric";
                            }
                            contentSection(sb, _node, context, null, "Rubric", levelno);
                        } else if (nodeName.equals("incipit")) {
                            contentSection(sb, _node, context, null, "Incipit", levelno);
                        } else if (nodeName.equals("explicit")) {
                            contentSection(sb, _node, context, null, "Explicit", levelno);
                        } else if (nodeName.equals("finalRubric")) {
                            contentSection(sb, _node, context, null, "Final&#160;rubric", levelno);
                        } else if (nodeName.equals("colophon")) {
                            contentSection(sb, _node, context, null, "Colophon", levelno);
                        } else if (nodeName.equals("note")) {
                            contentSection(sb, _node, context, null, "Note", levelno);
                        } else if (nodeName.equals("filliation")) {
                            contentSection(sb, _node, context, null, "Filiation", levelno);
                        } else if (nodeName.equals("textLang")) {
                            val = getMarkedText(_node, null, 0, subContext);
                            writeRow(sb, "Language", val, levelno);
                        } else if (nodeName.equals("witDetail")) {
                            witnessDetail(sb, node, "Second folio", levelno, 0, subContext);
                        } else if (nodeName.equals("msItemStruct")) {
                            if (!sectionEnded) {
                                sb.append("</table>\n");
                                levelNoLastEndTable = levelno;
                                sectionEnded = true;
                            }
                            processItemStruct(_node, sb, levelno+1,
                                    String.format("%s%d.", prefix, workNumber),
                                    subWorkNumber, subContext);
                            subWorkNumber++;
                        } else if (nodeName.equals("msItem")) {
                            if (!sectionEnded) {
                                sb.append("</table>\n");
                                levelNoLastEndTable = levelno;
                                sectionEnded = true;
                            }
                            processItem(_node, sb, levelno+1, String.format("%s%d.", prefix, workNumber), 
                                    subWorkNumber, subContext);
                            subWorkNumber++;
                        }
                    }
                    if (!sectionEnded) {
                        sb.append("</table>\n");
                        levelNoLastEndTable = levelno;
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Handler_msContents failed with: %s", ex.getMessage()));
            }
        }
    }

    private String capitalizeFirstLetter (String s) {
        return String.format("%c%s",
                Character.toUpperCase(s.charAt(0)),
                s.substring(1));
    }


    // Physical description section
    class Handler_physDesc implements ElementHandler
    {
        public void process(String path)
        {
            try
            {
                StringBuilder sb = new StringBuilder();
                String context = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc";
                Node objectDescNode = pi.getXML(context);

                if (objectDescNode != null)
                {
                    // Form
                    String val = getAttributeValue(objectDescNode, "form");
                    if (val != null && val.length() > 1) {
                        writeRow(sb, "Form", String.format("%c%s", 
                                Character.toUpperCase(val.charAt(0)), 
                                val.substring(1)));
                    }

                    // Material
                    String subContext = "supportDesc/support/material";
                    Node _node = XML.getRelativeNode(objectDescNode, subContext);
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, 0, String.format("%s.%s", context, subContext));
                        writeRow(sb, "Material", val);
                    }

                    // Watermark
                    subContext = "supportDesc/support/watermark";
                    _node = XML.getRelativeNode(objectDescNode, subContext);
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, 0, String.format("%s.%s", context, subContext));
                        writeRow(sb, "Watermark", val);
                    }

                    // Extent
                    subContext = "supportDesc/extent/measure";
                    _node = XML.getRelativeNode(objectDescNode, subContext);
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, 0, String.format("%s.%s", context, subContext)); // getAttributeValue(_node, "quantity");
                        if (val.length() > 0) {
                            writeRow(sb, "Extent", val);
                        }
                    }

                    // Dimensions
                    _node = XML.getRelativeNode(objectDescNode, "supportDesc/extent/dimensions");
                    if (_node != null)
                    {
                        val = get_dimensions(_node, null);
                        writeRow(sb, "Dimension", val);
                    }

                    // Foliation
                    subContext = "supportDesc/foliation";
                    NodeList list = XML.getRelativeNodeList(objectDescNode, subContext);
                    if (list != null)
                    {
                        int len = list.getLength();
                        for (int i=0; i<len; i++)
                        {
                            _node = list.item(i);
                            ArrayList<StringBuilder> a = get_foliation(_node, null, 0, subContext);
                            boolean first = true;
                            String label = null;
                            for (StringBuilder line : a) {
                                String[] fields = line.toString().split(" ", 2);
                                if (first) {
                                    label = fields[0];
                                    first = false;
                                } else {
                                    label = String.format("&nbsp;&nbsp;&nbsp;%s",fields[0]);
                                }
                                writeRow(sb, label, fields[1]);
                            }
                        }
                    }

                    // Collation
                    subContext = "supportDesc/collation";
                    _node = XML.getRelativeNode(objectDescNode, subContext);
                    if (_node != null)
                    {
                        ArrayList<StringBuilder> a = get_foliation(_node, null, 0, subContext);
                        for (StringBuilder line : a) {
                            String[] fields = line.toString().split(" ", 2);
                            writeRow(sb, fields[0], fields[1]);
                        }
                    }

                    // Condition
                    subContext = "supportDesc/condition";
                    _node = XML.getRelativeNode(objectDescNode, subContext);
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, 0, String.format("%s.%s", context, subContext));           
                        writeRow(sb, "Condition", val);
                    }

                    // Binding
                    subContext = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.bindingDesc";
                    Node bindingDescNode = pi.getXML(subContext);
                    if (bindingDescNode != null) {
                        list = XML.getRelativeNodeList(bindingDescNode, "binding");
                        if (list != null) {
                            int len = list.getLength();
                            for (int i = 0; i < len; i++) {
                                _node = list.item(i);
                                formatSealOrBindingDesc(sb, _node, "Binding", null, 0, String.format("%s.binding.%s", subContext, _node.getNodeName()));
//                                writeRow(sb, "Binding", val);
//                            }
                            }
                        }
                    }
                    // Seal
                    subContext = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.sealDesc";
                    Node sealDescNode = pi.getXML(subContext);
                    if (sealDescNode != null)
                    {
                        list = XML.getRelativeNodeList(sealDescNode, "seal");
                        if (list != null)
                        {
                            int len = list.getLength();
                            for (int i=0; i<len; i++)
                            {
                                _node = list.item(i);
                                formatSealOrBindingDesc(sb, _node, "Seal", null, 0,
                                        String.format("%s.seal.%s", subContext, _node.getNodeName()));
                            }
                        }
                    }


                    // Accompanying matter
                    subContext = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.accMat";
                    Node accMatNode = pi.getXML(subContext);
                    if (accMatNode != null)
                    {
                        val = getMarkedText(accMatNode, getLanguage(accMatNode, null), 0, subContext);
                        writeRow(sb, "Accompanying material", val);
                    }

                    // Layout
                    subContext = "layoutDesc/layout";
                    list = XML.getRelativeNodeList(objectDescNode, subContext);
                    if (list != null)
                    {
                        int len = list.getLength();
                        for (int i=0; i<len; i++)
                        {
                            _node = list.item(i);
                            val = getMarkedText(_node, null, 0,
                                    String.format("%s.%s.%s", context, subContext, _node.getNodeName()));           
                            writeRow(sb, "Layout", val);
                        }
                    }
                }
                if (sb.length() > 0)
                {
                    physicalDescription.append("<h3>Physical description</h3>\n");
                    physicalDescription.append("<table>\n");
                    physicalDescription.append(sb.toString());
                    physicalDescription.append("</table>\n");
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_physDesc: %s", ex.getMessage()));
            }
        }
        
        
        private void formatSealOrBindingDesc (StringBuilder sb, Node node, String type, String parentlang, int level, String context) {
            ArrayList<StringBuilder> a =
                    get_seal_or_bindingDesc(node, "Binding", null, 0, String.format("%s.binding.%s", context, node.getNodeName()));
            boolean first = true;
            String label = null;
            for (StringBuilder line : a) {
                String[] fields = line.toString().split(" ", 2);
                if (first) {
                    label = fields[0];
                    first = false;
                } else {
                    label = String.format("&nbsp;&nbsp;&nbsp;%s", fields[0]);
                }
                writeRow(sb, label, fields[1]);
            }
        }
    }



    // Hands section
    class Handler_handDesc implements ElementHandler
    {

        public void process(String path)
        {
            try
            {
                String val = null;
                StringBuilder sb = new StringBuilder();
                String context = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc";
                Node node = pi.getXML(context);
                if (node != null)
                {
                    Node _node = XML.getRelativeNode(node, "summary");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, 0, context + ".summary");  
                        writeRow(sb, "Summary", val);
                    }
                    NodeList list = XML.getRelativeNodeList(node, "handNote");
                    if (list != null)
                    {
                        int len = list.getLength();
                        for (int i=0; i<len; i++)
                        {
                            _node = list.item(i);
                            String _lang = getLanguage(_node, null);
                            val = getMarkedText(_node, _lang, 0, String.format("%s.handNote.%s", context, _node.getNodeName()));   
                            writeRow(sb, String.format("Hand %d", i+1), val);
                        }
                    }

                    context = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions";
                    node = pi.getXML(context);
                    if (node != null)
                    {
                        val = getMarkedText(node, null, 0, context); 
                        writeRow(sb, "Additions", val);
                    }

                    if (sb.length() > 0)
                    {
                        hands.append("<h3>Hands</h3>\n");
                        hands.append("<table>\n");
                        hands.append(sb.toString());
                        hands.append("</table>");
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_handDesc: %s", ex.getMessage()));
            }
        }
    }


    // Decoration section
    class Handler_decoDesc implements ElementHandler
    {
        public void process(String path)
        {
            try
            {
                String val = null;
                StringBuilder sb = new StringBuilder();

                String context = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='color']";
                Node node = pi.getXML(context);
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), 0, context); 
                    writeRow(sb, "Color", val);
                }

                context = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='border']";
                node = pi.getXML(context);
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), 0, context); 
                    writeRow(sb, "Border", val);
                }

                context = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='illustration']";
                node = pi.getXML(context);
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), 0, context); 
                    writeRow(sb, "Illustration", val);
                }

                if (sb.length() > 0)
                {
                    decoration.append("<h3>Decoration</h3>\n");

                    decoration.append("<table>\n");

                    decoration.append(sb.toString());

                    decoration.append("</table>");
                }
                
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_decoDesc", ex.getMessage()));
            }
        }
    }


    // History section
    class Handler_history implements ElementHandler
    {
        public void process(String path)
        {
            try
            {
                String val = null;

                StringBuilder sb = new StringBuilder();

                String context = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.origin";
                Node node = pi.getXML(context);
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), 0, context);
                    writeRow(sb, "Origin", val);
                }

                context = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.provenance";
                node = pi.getXML(context);
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), 0, context);
                    writeRow(sb, "Provenance", val);
                }

                context = "TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.acquisition";
                node = pi.getXML(context);
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), 0, context);
                    writeRow(sb, "Acquisition", val);
                }

                if (sb.length() > 0)
                {
                    history.append("<h3>History</h3>\n");

                    history.append("<table>\n");
                    history.append(sb.toString());
                    history.append("</table>");
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_history: %s", ex.getMessage()));
            }
        }
    }


    // Subject headings section
    class Handler_profileDesc implements ElementHandler
    {
        public void process(String path)
        {
            try
            {
                Node _node = null;
                String val = null;
                StringBuilder sb = new StringBuilder();
                String context = "TEI.teiHeader.profileDesc";
                Node node = pi.getXML(context);
                if (node != null)
                {
                    String subContext = "textClass/keywords[@scheme='#LCSH']/list/item";
                    NodeList list = XML.getRelativeNodeList(node, subContext);
                    if (list != null)
                    {
                        val = null;
                        _node = null;
                        int len = list.getLength();
                        for (int i=0; i<len; i++)
                        {
                            _node = list.item(i);
                            val = getMarkedText(_node, getLanguage(_node, null), 0,
                                    String.format("%s.%s.%s", context, subContext, _node.getNodeName())); 
                            writeRow(sb, "SubjectLC", val);
                        }
                    }

                    _node = XML.getRelativeNode(node, "textClass/catRef");
                    if (_node != null)
                    {
                        String target = getAttributeValue(_node, "target");
                        if (target != null)
                        {
                            Regex re = new Regex(target, "#(.*)$");
                            if (re.find())
                            {
                                target = re.getGroup(1);
                                target = classifications.get(target);
                            }
                            else
                                target = null;
                        }
                        if (target == null)
                            target = "";

                        String scheme = getAttributeValue(_node, "scheme");
                        if (scheme != null)
                            scheme = String.format("(%s)", scheme);
                        else
                            scheme = "";

                        if (target.length() > 0 || scheme.length() > 0)
                        {
                            val = String.format("%s %s", target, scheme);
                            writeRow(sb, "SubjectSL", val);
                        }
                    }
                }

                if (sb.length() > 0)
                {
                    subjectHeadings.append("<h3>Subject headings</h3>\n");
                    subjectHeadings.append("<table>\n");
                    subjectHeadings.append(sb.toString());
                    subjectHeadings.append("</table>");
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_profileDesc: %s", ex.getMessage()));
            }
        }
    }


    // Facsimile section
    class Handler_facsimile implements ElementHandler
    {
        public void process(String path)
        {
            try
            {
                if (checkDebugging(debugging)) {
                    //boolean rc = pi.getXMLDebug("TEI.facsimile");
                    //if (!rc) {
                    //    return;
                    //}
                }

                StringBuilder sbZone = new StringBuilder();
                StringBuilder sb = new StringBuilder();
                Node node = pi.getXML("TEI.facsimile");
                String xmlid = getAttributeValue(node, "xml:id");
                if (xmlid != null) {
                    sb.append(String.format("<a name='%s'/>", xmlid));
                }
                if (node != null)
                {
                    String id = "";
                    if (node.getAttributes().getNamedItem("xml:id") != null) {
                        id = node.getAttributes().getNamedItem("xml:id").getNodeValue();
                        id = String.format("&nbsp;(%s)", id.replace("images", "work "));
                    }
                    sb.append(String.format("<div id='tableScroller%d'>\n", currentFacsimile));
                    sb.append(String.format("<table id='facsTable%d' border='0'>", currentFacsimile));
                    currentFacsimile++;
                    sb.append("<tr><th>Whole image</th>");
                    NodeList list = XML.getRelativeNodeList(node, "surface");
                    int columns = countColumnsNeeded(list);
                    if (columns == 2) {
                        sb.append("<th>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</th>");
                        sb.append("<th>Individual pages</th>");
                    }
                    sb.append("</tr>\n");

                    int len = list.getLength();
                    for (int i=0; i<len; i++)
                    {
                        sbZone.setLength(0);
                        Node surface = list.item(i);
                        sb.append("<tr>\n");
                        NodeList surfaceChildren = surface.getChildNodes();
                        int numSurfaceChildren = surfaceChildren.getLength();
                        int numZones = 0;
                        for (int j=0; j<numSurfaceChildren; j++) {
                            Node surfaceChild = surfaceChildren.item(j);
                            String childName = surfaceChild.getNodeName();
                            if (childName.equals("graphic")) {
                                sb.append("<td>");
                                buildLink(surfaceChild, sb);
                                sb.append("</td>");
                            } else if (columns == 2 && childName.equals("zone")) {
                                if (numZones == 0) {
                                    sbZone.append("<td>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td><td>");
                                }
                                numZones++;
                                buildLink(XML.getRelativeNode(surfaceChild, "graphic"), sbZone);
                                sbZone.append("&nbsp;&nbsp;");
                            }
                        }
                        if (columns == 2) {         // if only one page on image don't show second column.
                            sb.append(sbZone.toString());
                        }
                        sb.append("</tr>\n");
                    }
                    sb.append("</table>\n");
                    sb.append("</div>\n");
                    if (sb.toString().trim().length() > 0)
                    {
                        facsimile.append(String.format("<h3>Facsimile %s</h3>\n", id));
                        facsimile.append(sb.toString());
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_facsimile: %s", ex.getMessage()));
            }
        }
    }

    private int countColumnsNeeded(NodeList list) {
        int len = list.getLength();
        for (int i = 0; i < len; i++) {
            Node surface = list.item(i);
            NodeList surfaceChildren = surface.getChildNodes();
            int numSurfaceChildren = surfaceChildren.getLength();
            int numZones = 0;
            for (int j = 0; j < numSurfaceChildren; j++) {
                Node surfaceChild = surfaceChildren.item(j);
                String childName = surfaceChild.getNodeName();
                if (childName.equals("zone")) {
                    numZones++;
                }
            }
            if (numZones > 1) {         // if only one page on image don't show second column.
                return 2;
            }
        }

        return 1;
    }

    private void buildLink(Node graphicNode, StringBuilder sb) throws Exception {
        String id = getAttributeValue(graphicNode, "xml:id");
        if (id != null) {
            String desc = "";
            Node descNode = XML.getRelativeNode(graphicNode, "desc");
            if (descNode != null) {
                desc = descNode.getTextContent();
            }
            sb.append(buildLinkURL(id, desc));
        }
    }


    // This build link in list of facsimiles
    private String buildLinkURL (String id, String desc) {
        id = id.trim();
        String url = zoneMap.get(id);
        if (url != null) {
            return _buildLink(id, desc, "black");
        } else {
         // server will locate. Link is javascript:displayImage(...)
            return String.format("<a style='color:red' href=\"javascript:displayImage('%s')\">%s</a>", id, desc);
        }
    }


    // Transcription section
    class Handler_transcription implements ElementHandler
    {
        public void process(String path)
        {
            transcription.append("<br/><button id='btnToggle' onclick='toggleTranscription()'>Show manuscript layout</button><br/>");
            _process(Style.TEXT_STRUCTURE, "text structure");
            _process(Style.MANUSCRIPT_LAYOUT, "manuscript layout");
        }


        private void _process (Style s, String desc) {
            style = s;
            try
            {
                StringBuilder sb = new StringBuilder();
                StringBuilder sb1 = new StringBuilder();
                String context = "TEI.text";
                Node node = pi.getXML(context);
                if (node != null)
                {
                    String lang = "sa-Latn-x-KH";
                    if (node.hasAttributes())       // override KH if xml:lang specified
                    {
                        Node attlang = node.getAttributes().getNamedItem("xml:lang");
                        if (attlang != null)
                            lang = attlang.getNodeValue();
                    }
                    append(sb, getMarkedText(node, lang, 0, context));
                    if (sb.toString().trim().length() > 0)
                    {
                        if (s == Style.MANUSCRIPT_LAYOUT) {
                            append(sb1, String.format("<div id='%s' style='display:none'>\n", s.toString()));
                            transcription.append(sb1.toString());
                        } else {
                            append(sb1, String.format("<div id='%s' style='display:block'>\n", s.toString()));
                            transcription.append(sb1);
                        }
                        sb1.setLength(0);
                        append(sb1, String.format(String.format("<h3>Transcription (%s)</h3>\n", desc)));
                        append(sb1, sb.toString().trim());
                        transcription.append(sb1.toString());
                        transcription.append("</div>\n");
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_transcription: %s", ex.getMessage()));
            }
            
            style = Style.HEADER;
        }
    }


    // Last revision
    class Handler_recordHist implements ElementHandler
    {
        public void process(String path)
        {
            try
            {
                Node node = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist");
                if (node != null)
                {
                    NodeList list = XML.getRelativeNodeList(node, "change");
                    if (list != null)
                    {
                        int len = list.getLength();
                        if (len > 0)
                        {
                            Node _node = list.item(len - 1);
                            _node = XML.getRelativeNode(_node, "date");
                            if (_node != null)
                            {
                                changes.append("<h3>Record history</h3>");
                                writeRow(changes, "Revised", String.format(" %s", _node.getTextContent()));
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_recordHist: %s", ex.getMessage()));
            }
        }

    }

    private String buildTextLink(String text, String anchor) {

        int idx = text.lastIndexOf(".");
        if (idx != -1) {
            text = text.substring(0, idx);
        }
        
        if (anchor.length() == 0) {
            return String.format("../LoadText?text=%s", text);
        } else {
            return String.format("../LoadText?text=%s&anchor=%s#%s",
                    text, anchor, anchor);
        }
    }

    private String _buildLink(String facs, String locus, String color) {
        String url = zoneMap.get(facs);
        if (url == null) {
            return String.format("[[facs=%s, locus=%s]]", facs, locus);
        } else {
            reNoExtension.reset(url);
            if (reNoExtension.find()) {
                url = reNoExtension.getGroup(1);
            }
            reNoExtension.reset(theFileName);
            String fn = theFileName;
            if (reNoExtension.find()) {
                fn = reNoExtension.getGroup(1);
            }
            return String.format("<a style='color:%s' target='image' href='%s?type=cropped&file=%s&image=%s'>%s</a>",
                    color, Main.manuscriptServer, fn, url, locus);
        }
    }

    class LocusInfo {

        String scheme;
        String locus;
        String from;
        String to;
        String target;
        String facs;
        String[] facList;

        void setfacs(String facs) {
            if (facs != null) {
                facs = TextFormatter.normalizeID(facs);
                this.facs = facs.replace("#", "");
                facList = this.facs.split("\\s+");
            }
        }

        public String buildLink() {
            if (target != null) {
                return String.format("<a href='%s'>%s</a>", target, locus);
            } else if (facs != null) {
                if (facList.length == 1) {
                    return _buildLink(facs, locus, "black");
                } else if (facList.length == 2) {
                    String link1 = _buildLink(facList[0], formatFolioID(facList[0]), "black");
                    String link2 = _buildLink(facList[1], formatFolioID(facList[1]), "black");
                    return String.format("%s&ndash;%s", link1, link2);
                } else {
                    return facs;
                }
            } else if (from != null) {
                return String.format("<a style='color:red' href=\"javascript:displayImage('f%s')\">%s</a>", from, locus);
            } else {
                return "?";
            }
        }

    }




    public String formatFolioID(String id) {
        if (id.length() > 2 && id.charAt(1) != '.') {
            return "f. " + id.substring(1);
        } else {
            return id;
        }
    }
}
