package siva;

import java.io.BufferedWriter;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.Config;
import org.sanskritlibrary.ResourceInputStream;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class ParseTEIExtract extends ParseTEI {
    public ParseTEIExtract (String fileName, ResourceInputStream ris, MarcRecord mr, Config config) throws Exception {
        super(fileName, null, null, ris, mr, config);
        
        reserveEntries();
        setKeys();
    }

    public ParseTEIExtract (String fileName, String extraction, BufferedWriter bw, ResourceInputStream ris, Config config) throws Exception {
        super(fileName, extraction, bw, ris, null, config);
        
        reserveEntries();
        setKeys();
    }

    protected void prepareParse(SAXParser saxParser) {
    }
    
    @Override
    public void endDocument() throws SAXException {
        
    }    

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        try {
            debug++;
            pushPath(qName);
            String thepath = getPath();
            pi.recordStatistics(qName, thepath, atts, theFileName);

            if (thepath.startsWith(extractionPath)) {
                int len = extractionPath.length();
                if (thepath.length() == len || thepath.charAt(len) == '.') {
                    pi.rememberFile(theFileName);
                    bw.write("<");
                    bw.write(qName);
                    int numatts = atts.getLength();
                    for (int i = 0; i < numatts; i++) {
                        String name = atts.getQName(i);
                        String value = atts.getValue(i);
                        bw.write(String.format(" %s=\"%s\"", name, value));
                    }
                    bw.write(">");
                }
            }
        } catch (Exception ex) {
            displayError(ex.getMessage());
            throw new SAXException(ex.getMessage());
        }
    }
    
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        try {
            String thepath = getPath();
            if (thepath.startsWith(extractionPath)) {
                int len = extractionPath.length();
                if (thepath.length() == len || thepath.charAt(len - 1) == '.') {
                    bw.write(String.format("</%s>", qName));
                }
            }
            if (thepath.equals(extractionPath)) {
                bw.write("\n");
            }
            popPath();
        } catch (Exception ex) {
            displayError(ex.getMessage());
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
        try {
            String thepath = getPath();
            if (thepath.startsWith(extractionPath)) {
                int len = extractionPath.length();
                if (thepath.length() == len || thepath.charAt(len - 1) == '.') {
                    bw.write(new String(ch, start, length));
                }
            }

        } catch (Exception ex) {
            displayError(ex.getMessage());
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void comment(char[] ch, int start, int length) throws SAXException {
        if (getPath().equals(extractionPath)) {
            StringBuilder tsb = new StringBuilder();
            if (newElement != null) {
                newElement = null;
                tsb.append(">");
            }
            tsb.append("<!--");
            tsb.append(new String(ch, start, length));
            tsb.append("-->");
            if (endCommentWithLineFeed) {
                tsb.append("\n");
            }
            try {
                bw.write(tsb.toString());
            } catch (Exception ex) {
                throw new SAXException(ex);
            }
        }
    }

    @Override
    public void processingInstruction(String target, String data) throws SAXException {
        if (extractionPath != null && !getPath().equals(extractionPath)) {
            return;
        }
        try
        {
            bw.write("<?");
            bw.write(target);
            bw.write(" ");
            bw.write(data);
            bw.write("?>\n");
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }
}
