package siva;

import java.io.BufferedWriter;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.Config;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.ResourceInputStream;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class ParseTEIFormat extends ParseTEI {
    public ParseTEIFormat (String fileName, ResourceInputStream ris, Config config) throws Exception {
        super(fileName, null, null, ris, null, config);
        
        reserveEntries();
        setKeys();
    }

    protected void prepareParse(SAXParser saxParser) throws Exception {
        shortName = shortName.replace(".xml", ".html");
        this.bw = IOAssistance.openWriter(Main.htmlpath + shortName);
        this.bw.write(String.format("<html><head><meta http-equiv='Content-Type' content='text/html; charset=utf-8'/><title>%s</title><style type='text/css'>br{margin-bottom:8px}</style><script type='text/javascript'>%s</script></head><body onload='init()'>\n", shortName, script));
    }
    
    @Override
    public void endDocument() throws SAXException {
        try {
            bw.write(fixup(manuscriptIdentifiers.toString()));
            bw.write(fixup(contents.toString()));
            bw.write(fixup(physicalDescription.toString()));
            bw.write(fixup(hands.toString()));
            bw.write(fixup(decoration.toString()));
            bw.write(fixup(history.toString()));
            bw.write(fixup(subjectHeadings.toString()));
            bw.write(fixup(changes.toString()));
            bw.write(fixup(facsimile.toString()));
            bw.write(transactionFixup(fixup(transcription.toString())));
            bw.write(doMappings());
            bw.write("</body></html>");
            bw.close();
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        try {
            debug++;
            if (qName.equals("locus") && atts.getValue("facs") == null && atts.getValue("target") == null) {
                displayError("locus element has neither a facs nor a target attribute");
            }
            if (qName.equals("graphic")) {
                qName = "graphic";
            }
            pi.pushElement(qName, atts);
        } catch (Exception ex) {
            displayError(ex.getMessage());
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        try {
            if (locator.getLineNumber() == 168) {
                localName = null;
            }
            pi.popElement(qName);
        } catch (Exception ex) {
            displayError(ex.getMessage());
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
        try {
            String s = new String(ch, start, length);
            pi.append(new String(s));

        } catch (Exception ex) {
            displayError(ex.getMessage());
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void comment(char[] ch, int start, int length) throws SAXException {
        boolean displayComment = false;
        
    }

    @Override
    public void processingInstruction(String target, String data) throws SAXException {
    }
}
