package siva;

import java.io.BufferedWriter;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.Config;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.ResourceInputStream;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import static siva.ParseTEI.embeddedTags;

public class ParseTEIInsert extends ParseTEI {
    public ParseTEIInsert (String fileName, ResourceInputStream ris, MarcRecord mr, Config config) throws Exception {
        super(fileName, null, null, ris, mr, config);
        
        reserveEntries();
        setKeys();
    }

    protected void prepareParse(SAXParser saxParser) throws Exception {
        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description,
        saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        this.bw = IOAssistance.openWriter("./XML-fixedup/" + shortName);
        this.bw.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        insertMARC = mr != null;
    }
    
    @Override
    public void endDocument() throws SAXException {
        try {
            bw.write(insertBuffer.toString());
            bw.close();
            insertBuffer.setLength(0);
        } catch (Exception ex) {
            displayError(ex.getMessage());
            throw new SAXException(ex);
        }
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        try {
            debug++;
        if (newElement != null && newElement.equals(pi.currentPath.toString()))     // this means that charactersInsert was not called
        {
            insertBuffer.append(">");
            newElement = null;
        }

        if (embeddedTags.get(qName) == null)
        {
            int i = insertBuffer.length() - 1;
            while (i >= 0)
            {
                if (insertBuffer.charAt(i) <= 0x20)
                {
                    insertBuffer.deleteCharAt(i);
                    i--;
                }
                else
                    break;
            }
            insertBuffer.append("\n");
            insertBuffer.append(getXMLIndentation(tagLevel));
        }
        
        pi.pushElement(qName, atts);
        tagLevel++;
        newElement = pi.currentPath.toString();
        if (qName.equals("TEI"))
            endCommentWithLineFeed = false;      // comments before root element need line feed, those after root element do not.

        insertBuffer.append("<");
        insertBuffer.append(qName);
        int numatts = atts.getLength();
        for (int i=0; i<numatts; i++)
        {
            String name = atts.getQName(i);
            String value = atts.getValue(i);
            insertBuffer.append(String.format(" %s=\"%s\"", name, value));
        }
        if (newElement.equals("TEI.teiHeader.fileDesc.titleStmt.title"))
        {
            insertBuffer.append(">");
            Regex re = new Regex(theFileName, "(.+?)(\\d+)");
            if (re.find())
            {
                String x = re.getGroup(1).toLowerCase();
                String y = re.getGroup(2);
                if (x.equals("penn"))
                    x = "UPenn";
                else if (x.equals("br"))
                    x = "Brown";
                String title = String.format("SLMsCat %s %s", x, y);
                insertBuffer.append(title);
                newElement = null;
            }
        }
        else if (newElement.endsWith(".title"))
        {
            if (atts.getValue("rend") == null)
            {
                insertBuffer.append(" rend='smallcap'");
            }
        }
        else if (newElement.equals("TEI.teiHeader"))
        {
            if (atts.getValue("xml:lang") == null)
            {
                insertBuffer.append(" xml:lang='en'");
            }
        }
        else if (newElement.equals("TEI.text"))
        {
            if (atts.getValue("xml:lang") == null)
            {
                if (theFileName.startsWith("penn"))
                    insertBuffer.append(" xml:lang='sa-Latn-x-KH'");
                else
                    insertBuffer.append(" xml:lang='sa-Latn-x-SLP1'");
            }
        }
        else if (newElement.endsWith(".pb"))
        {
            if (atts.getValue("type") == null)
                insertBuffer.append(" type='folio'");
        }
        else if (newElement.endsWith(".measure"))
        {
            if (atts.getValue("quantity") != null)
                measureQuantity = atts.getValue("quantity");
            else
                measureQuantity = null;
        }
            
        } catch (Exception ex) {
            displayError(ex.getMessage());
            throw new SAXException(ex.getMessage());            
        }
    }
    
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        try {
        if (newElement != null && newElement.equals(pi.currentPath.toString()))     // have <elem att=.../> instead of <elem att=...>...</elem>
        {
            newElement = null;
            if (measureQuantity != null)
            {
                String unit = "folios";
                if (measureQuantity.equals("1"))
                    unit = "folio";
                insertBuffer.append(String.format(">%s %s</measure>", measureQuantity, unit));
                measureQuantity = null;
            }
            else if (onetag.get(qName) != null)
            {
                insertBuffer.append("/>");
                insertBuffer.append(String.format("\n%s", getXMLIndentation(tagLevel - 1)));
            }
            else
                insertBuffer.append(String.format("></%s>", qName));
            
            pi.rememberEmptyElement(qName);
        }
        else
        {
            String s = "";
            if (embeddedTags.get(qName) == null && containsEOL(qName))
            {
                trimRight();
                s = String.format("\n%s", getXMLIndentation(tagLevel - 1));
            }
            insertBuffer.append(String.format("%s</%s>", s, qName));
        }

        if (insertMARC && pi.currentKey.toString().equals("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.altIdentifier"))
        {
            insertMARC = false;
            writeMARCIdentifier("marc:leader", mr.leader);
            writeMARCIdentifier("marc:oclc", mr.control001);
            writeMARCIdentifier("marc:data", mr.control008);
            writeMARCIdentifier("marc:institution", mr.datafield035);
            writeMARCIdentifier("marc:master", mr.datafield079);
        }
        pi.popElement(qName);
        tagLevel--;
        } catch (Exception ex) {
            displayError(ex.getMessage());
            throw new SAXException(ex.getMessage());            
        }
    }
    
    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
        try {
        if (newElement != null)
        {
            newElement = null;
            insertBuffer.append(">");
            while(length > 0)
            {
                if (ch[start] > 0x20)
                    break;
                else
                {
                    start++;
                    length--;
                }
            }
        }
        String s = new String(ch, start, length);
        s = s.replace("&", "&amp;");
        String ind = String.format("\n%s", getXMLIndentation(tagLevel));
        s = s.replaceAll("\\s*\n\\s*", ind);
        insertBuffer.append(s);
            
        } catch (Exception ex) {
            displayError(ex.getMessage());
            throw new SAXException(ex.getMessage());            
        }
    }

    @Override
    public void comment(char[] ch, int start, int length) throws SAXException {
        StringBuilder tsb = new StringBuilder();
        if (newElement != null) {
            newElement = null;
            tsb.append(">");
        }
        tsb.append("<!--");
        tsb.append(new String(ch, start, length));
        tsb.append("-->");
        if (endCommentWithLineFeed) {
            tsb.append("\n");
        }
        insertBuffer.append(tsb.toString());
    }

    @Override
    public void processingInstruction(String target, String data) throws SAXException {
        if (extractionPath != null && !getPath().equals(extractionPath)) {
            return;
        }
        try
        {
            bw.write("<?");
            bw.write(target);
            bw.write(" ");
            bw.write(data);
            bw.write("?>\n");
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }
}
