package siva;

import java.io.BufferedWriter;
import java.io.File;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.ElementInfo;
import org.sanskritlibrary.ElementStack;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.SAXParser;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

// Pass 1 check accuracy of reference attributes (if value contains a # it is a reference attribute and a reference attribute must contain a #
//        collect all internal references in internalReferences
// Pass 2 Collect content of all elements with an xml:id that is in internalReferences
//        Put all xml:id values in targetReferences.
// Pass 3 Verify that all references are legal
public class References implements SAXParserHandler
{
    static public int LINK = 1;
    static public int INPLACE = 2;
    static public Hashtable<String, Integer> referenceAttributes = new Hashtable<String, Integer> ();
    static {
        referenceAttributes.put("bibl.corresp", LINK);
        referenceAttributes.put("biblScope.corresp", LINK);
        referenceAttributes.put("biblScope.from", LINK);
        referenceAttributes.put("biblScope.to", LINK);
        referenceAttributes.put("foreign.corresp", LINK);
        referenceAttributes.put("quote.corresp", LINK);
        referenceAttributes.put("note.corresp", LINK);
        referenceAttributes.put("altIdentifier.corresp", LINK);
        referenceAttributes.put("surrogates.corresp", LINK);
        referenceAttributes.put("ref.target", LINK);
        referenceAttributes.put("catRef.target", LINK);
        referenceAttributes.put("link.target", LINK);
        referenceAttributes.put("locus.target", LINK);
        referenceAttributes.put("rubric.corresp", INPLACE);
        referenceAttributes.put("incipit.corresp", INPLACE);
        referenceAttributes.put("explicit.corresp", INPLACE);
        referenceAttributes.put("finalRubric.corresp", INPLACE);
        referenceAttributes.put("colophon.corresp", INPLACE);
        referenceAttributes.put("locus.facs", LINK);
        referenceAttributes.put("pb.facs", LINK);
        referenceAttributes.put("title.ref", LINK);
        referenceAttributes.put("author.ref", LINK);
        referenceAttributes.put("persName.ref", LINK);
        referenceAttributes.put("locus.scheme", LINK);
        referenceAttributes.put("keywords.scheme", LINK);
        referenceAttributes.put("p.synch", LINK);
        referenceAttributes.put("seg.synch", LINK);
        referenceAttributes.put("trailer.synch", LINK);
        referenceAttributes.put("lg.corresp", LINK);
        referenceAttributes.put("gap.corresp", LINK);
        referenceAttributes.put("ab.corresp", LINK);
        referenceAttributes.put("note.sameAs", LINK);
        referenceAttributes.put("speaker.corresp", LINK);
    }

    static public Hashtable<String, String> hasInplaceAttribute = new Hashtable<String, String>();

    static {
        Enumeration<String> en = referenceAttributes.keys();
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            Integer status = referenceAttributes.get(key);
            if (status != null && status == INPLACE) {
                String fields[] = key.split("\\.");
                hasInplaceAttribute.put(fields[0], fields[1]);
            }
        }
    }

    static boolean wellFormednessError = false;
    static boolean checkURL = false;
    /*
     * inplace
     *  Created in pass 1 from referenceAttributes (e.g. rubric.corresp). The shortname of the file is prepended to the value
     * of the referenceAttribute along with a #. Note the reference attribute may contain multiple references. If so each
     * reference is add to inplace. (e.g., #c.s1, #c.s2, #c.s3, #c.s4, #c.s5 has 5 references)
     *  Used in pass 2 to determine which elements get copied to a buffer and saved in xmlidContent. For each element with an
     * xml:id attribute see if it's value is in inplace. If it is, copy it to the buffer.
     */
    static public Hashtable<String, Boolean> inplace = new Hashtable<String, Boolean>();

    /*
     * texts
     * contains SLP texts of texts that are referenced in the documents.
     */
    static public Hashtable<String, String> texts = new Hashtable<String, String>();

    /*
     * external references
     * Keeps track of every element that has an xml:id. The file name is stored with the xml:id so these can
     * be can be used to resolve external references.
     */
    static public Hashtable<String, Boolean> externalReferences = new Hashtable<String, Boolean>();

    /*
     * internalReferences
     * Used to detect nested xml:ids that are both referenced somewhere in the document.
     */
    static Hashtable<String, Boolean> internalReferences = new Hashtable<String, Boolean>();

    /*
     * noWellFormed
     * Used to keep track of documents that are not well-formed so that they can be skipped on passes 2 and 3
     */
    static public Hashtable<String, Boolean> notWellFormed = new Hashtable<String, Boolean>();
    
    /*
     * xmlidContent
     * Contains content for INPLACE elements
     */
    static public Hashtable<String, XMLChunk> xmlidContent = new Hashtable<String, XMLChunk>();

    static BufferedWriter errorFile;
    private static Regex re = new Regex("", "#([^\\s]+)");
    private static Regex refacs = new Regex("", "^#f(?:[IVXLCDM]+[a-z]?:)?(?:\\d{1,3}|[ivxlcdm]+)[a-z]?[rv]$");
    private static Regex retext = new Regex("", "^texts/([^#]+)#(.*)");
    private static Regex rehttp = new Regex("", "^https?://(.*)");
    public static int numErrors = 0;

    private ElementStack stack = new ElementStack();
    private boolean lineSkipDone = false;
    private String filename;
    private String shortname;
    private int currentLevel;
    private StringBuilder buffer = new StringBuilder();
    private Locator locator;
    private int passNumber;
    private boolean collectFormula = false;
    private StringBuilder formulaBuffer = new StringBuilder();
    SAXParser saxParser;
    private ArrayList<CopyBuffer> copyBufferList = new ArrayList<CopyBuffer>();

    static Hashtable<String, Boolean> iskon = new Hashtable<String, Boolean>();
    static String[] iskonData = new String[] {
"editions/purana/bhp/iskon/Sb03_0001.tif",
"editions/purana/bhp/iskon/Sb05_002.tif",
"editions/purana/bhp/iskon/Sb05_577.tif",
"editions/purana/bhp/iskon/Sb06_001.tif",
"editions/purana/bhp/iskon/Sb01_001.tif",
"editions/purana/bhp/iskon/Sb01_257.tif",
"editions/purana/bhp/iskon/Sb01_809.tif",
"editions/purana/bhp/iskon/Sb02_001.tif",
"editions/purana/bhp/iskon/Sb02_507.tif",
"editions/purana/bhp/iskon/Sb03_1240.tif",
"editions/purana/bhp/iskon/Sb04_001.tif",
"editions/purana/bhp/iskon/Sb04_805.tif",
"editions/purana/bhp/iskon/Sb06_605.tif",
"editions/purana/bhp/iskon/Sb07_001.tif",
"editions/purana/bhp/iskon/Sb07_584.tif",
"editions/purana/bhp/iskon/Sb08_002.tif",
"editions/purana/bhp/iskon/Sb08_025.tif",
"editions/purana/bhp/iskon/Sb08_140.tif",
"editions/purana/bhp/iskon/Sb08_560.tif",
"editions/purana/bhp/iskon/Sb09_001.tif",
"editions/purana/bhp/iskon/Sb09_536.tif",
"editions/purana/bhp/iskon/Sb10-1_0001.tif",
"editions/purana/bhp/iskon/Sb10-2_0094.tif",
"editions/purana/bhp/iskon/Sb10-2_0125.tif",
"editions/purana/bhp/iskon/Sb10-2_0133.tif",
"editions/purana/bhp/iskon/Sb10-2_0165.tif",
"editions/purana/bhp/iskon/Sb10-3_0868.tif",
"editions/purana/bhp/iskon/Sb10-4_989.tif",
"editions/purana/bhp/iskon/Sb10-5_0001.tif",
"editions/purana/bhp/iskon/Sb10-6_0583.tif",
"editions/purana/bhp/iskon/Sb10-6_0587.tif",
"editions/purana/bhp/iskon/Sb10-6_0608.tif",
"editions/purana/bhp/iskon/Sb10-6_0614.tif",
"editions/purana/bhp/iskon/Sb10-6_0818.tif",
"editions/purana/bhp/iskon/Sb10-6_0823.tif",
"editions/purana/bhp/iskon/Sb10-6_0841.tif",
"editions/purana/bhp/iskon/Sb10-6_0924.tif",
"editions/purana/bhp/iskon/Sb10-6_0932.tif",
"editions/purana/bhp/iskon/Sb10-6_1447.tif",
"editions/purana/bhp/iskon/Sb10-6_1680.tif",
"editions/purana/bhp/iskon/Sb11_0002.tif",
"editions/purana/bhp/iskon/Sb11_0800.tif",
"editions/purana/bhp/iskon/Sb11_0806.tif",
"editions/purana/bhp/iskon/Sb11_0814.tif",
"editions/purana/bhp/iskon/Sb11_0890.tif",
"editions/purana/bhp/iskon/Sb11_0895.tif",
"editions/purana/bhp/iskon/Sb11_1350.tif",
"editions/purana/bhp/iskon/Sb11_1354.tif",
"editions/purana/bhp/iskon/Sb12_001.tif",
"editions/purana/bhp/iskon/Sb12_002.tif",
"editions/purana/bhp/iskon/Sb12_047.tif",
"editions/purana/bhp/iskon/Sb12_075.tif",
"editions/purana/bhp/iskon/Sb12_343.tif",
};
    static {
        for (String s : iskonData) {
            iskon.put(s, true);
        }
    }

    static void setFile (BufferedWriter errorFile) {
        References.errorFile = errorFile;
    }

    public References (int passNumber) {
        this.passNumber = passNumber;
        saxParser = new SAXParser(this);
    }

    private void writeError(String error) {
        numErrors++;
        skipLine();         // maybe
        try {            
            errorFile.write(error);
            errorFile.write("\n");
        } catch (Exception ex) {
            System.out.println(error);
        }
    }

    public void parse(String filename) throws Exception {
        lineSkipDone = false;
        this.filename = filename;
        this.shortname = new File(filename).getName();
        if (notWellFormed.get(shortname) != null) {         // this was discovered in pass 1, don't keep on rediscovering it.
            return;
        }
        try {
            saxParser.parse(filename);
        } catch (Exception ex) {
            System.out.println(ex.getMessage());
            // error should have been reported elsewhere
        }
    }

    public void warning(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void error(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void fatalError(SAXParseException exception) throws SAXException {
        String msg = String.format("%s:%d.%d %s\n", shortname, locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        writeError(msg);
        wellFormednessError = true;
        notWellFormed.put(shortname, true);
        throw new SAXException("Boom!");
    }

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    public void startDocument() throws SAXException {
//        System.out.println("\n");
    }

    private void skipLine() {
        if (!lineSkipDone) {
            try {
                errorFile.write("\n");
            } catch (Exception ex) {
                System.out.println();
            }
        }
        lineSkipDone = true;
    }

    public void endDocument() throws SAXException {
        if (passNumber == 1) {
            endDocumentPass1();
        } else if (passNumber == 2) {
            endDocumentPass2();
        } else if (passNumber == 3) {
            endDocumentPass3();
        }
    }


    private void endDocumentPass1 () {
    }

    private void endDocumentPass2 () {
    }

    private void endDocumentPass3 () {
    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    public void endPrefixMapping(String prefix) throws SAXException {
    }

    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        if (passNumber == 1) {
            startElementPass1(uri, localName, qName, atts);
        } else if (passNumber == 2) {
            startElementPass2(uri, localName, qName, atts);
        } else if (passNumber == 3) {
            startElementPass3(uri, localName, qName, atts);
        }
    }

    /*
     * The purpose of this pass is to
     * 1. initialize externalReferences with all xml:id values
     * 2. initialize internalReferences with all references found in reference attributes
     * 3. report non-reference attributes that contain a #
     * 4. report reference attributes that are not xml/ or http:// that do not contain a #
     */
    private void startElementPass1(String uri, String localName, String qName, Attributes atts) throws SAXException {
        stack.pushElement(qName, atts);
        int numatts = atts.getLength();

        for (int i = 0; i < numatts; i++) {
            String name = atts.getQName(i);
            String value = atts.getValue(i);
            String key = String.format("%s.%s", qName, name);
            Integer status = referenceAttributes.get(key);
            if (name.equals("xml:id")) {                    // add to external references in case some other file references this one
                String refkey = String.format("%s#%s", shortname, value);
                externalReferences.put(refkey, true);
            } else if (value.contains("#")) {                  // it contains a reference
                if (status == null) {                   // but it is not a reference attribute
                    writeError(String.format("%s:%d.%d %s.%s has a reference (%s) but is not supposed to.",
                            shortname, locator.getLineNumber(), locator.getColumnNumber(), qName, name, value));
                } else {       // it's a reference attribute
                    String[] values = value.split(" ");
                    for (String _value : values) {
                        if (_value.charAt(0) == '#') {                    // it is a local reference
                            key = String.format("%s%s", shortname, _value);
                            internalReferences.put(key, Boolean.TRUE);
                            if (status != null && status == INPLACE) {
                                inplace.put(key, Boolean.TRUE);
                            }
                        }
                    }
                }
            } else {            // does not contain a reference
                if (status != null) {
                    if (!value.startsWith("xml/") && !value.startsWith("http") && !value.equals("unknown")) {
                        if (iskon.get(value) == null) {
                            writeError(String.format("%s:%d.%d %s.%s='%s'",
                                shortname, locator.getLineNumber(), locator.getColumnNumber(), qName, name, value));
                        }
                    }
                }
            }
        }

        if (qName.equals("formula")) {
            if (stack.getElement(stack.topIndex()-1).qName.equals("foliation")) {
                collectFormula = true;
            }
        }
    }


    /*
     * The purpose of this pass is to
     * 1. Collect the elements that will be inserted inline
     * 2. Report inline xml:ids that are nested within other inline xml:ids
     * 3. Report illegal references in reference attributes
     */
    private void startElementPass2(String uri, String localName, String qName, Attributes atts) throws SAXException {
        stack.pushElement(qName, atts);
        String value = atts.getValue("xml:id");
        if (value != null) {        // this might be referenced from someplace else.
            String key = String.format("%s#%s", shortname, value);

            if (inplace.get(key) != null && internalReferences.get(key) != null) {
                // This is an inplace reference that is used someplace in the document
                CopyBuffer cb = new CopyBuffer(key, stack.getPath());
                copyBufferList.add(cb);
            }
        }


        if (copyBufferList.size() > 0) {
            for (CopyBuffer cb : copyBufferList) {
                buffer = cb.buffer;
                buffer.append(String.format("<%s", qName));
                int numatts = atts.getLength();
                for (int i = 0; i < numatts; i++) {
                    buffer.append(String.format(" %s='%s'", atts.getQName(i), atts.getValue(i)));
                }
                buffer.append(">");
            }
        }

        // check for legal references now since we have collected all targets in pass 1
        int numatts = atts.getLength();
        for (int i = 0; i < numatts; i++) {
            String attname = atts.getQName(i);
            String key = String.format("%s.%s", qName, attname);
            if (referenceAttributes.get(key) != null) {
                String[] fields = atts.getValue(i).split(" ");
                for (String field : fields) {
                    if (field.startsWith("texts/")) {          // if it is a texts reference
                        legalTextReference(field, true);       // error message will be printed in legalTextReference
                    } else if (field.startsWith("xml/")) {            // it references another xml file
                        field = field.substring(4);
                        if (externalReferences.get(field) == null) {
                            int idx = field.indexOf("#");
                            String wfkey = null;
                            if (idx != -1) {
                                wfkey = field.substring(0, idx);
                            }
                            if (idx == -1) {
                                String xmlfilename = Main.xmlpath + field;
                                File f = new File(xmlfilename);
                                if (f.exists() == false) {
                                    String msg = String.format("%s:%d.%d file does not exist: %s='xml/%s'",
                                            shortname, locator.getLineNumber(), locator.getColumnNumber(),
                                            key, field);
                                    writeError(msg);
                                }
                            } else if (notWellFormed.get(wfkey) == null) {          // should never be true
                                String msg = String.format("%s:%d.%d illegal xml reference: %s='xml/%s'",
                                        shortname, locator.getLineNumber(), locator.getColumnNumber(),
                                        key, field);
                                writeError(msg);
                            }
                        }
                    } else if (field.startsWith("http")) {
                        if (checkURL) {
                            if (!exists(field)) {
                                writeError(String.format("%s:%d.%d value of %s.%s (%s) does not exist",
                                        shortname, locator.getLineNumber(), locator.getColumnNumber(),
                                        qName, attname, field));
                            }
                        }
                    } else if (field.charAt(0) == '#') {        // an internal reference
                        if (!attname.equals("facs")) {
                            String refkey = String.format("%s#%s", shortname, field.substring(1));
                            if (externalReferences.get(refkey) == null) {
                                String msg = String.format("%s:%d.%d illegal reference: %s='%s'",
                                        shortname, locator.getLineNumber(), locator.getColumnNumber(), key, field);
                                writeError(msg);
                            }
                        }
                    } else if (!legalTextReference(Main.textSubdirectory + field, false)) {      // if it is not a text reference without the texts/
                        String[] sp = field.split("#");
                        String msg = null;
                        if (sp.length > 0 && texts.get(sp[0]) != null && texts.get(sp[0]).equals("?")) {
                            msg = String.format("%s:%d.%d illegal text reference - %s='%s'",
                                    shortname, locator.getLineNumber(), locator.getColumnNumber(), key, field);
                        } else {
                            msg = String.format("%s:%d.%d illegal reference -> %s='%s'",
                                    shortname, locator.getLineNumber(), locator.getColumnNumber(), key, field);
                        }
                        if (!field.equals("unknown") && iskon.get(field) == null) {
                            writeError(msg);
                        }
                    }

                }
            }
        }
    }

    private boolean legalTextReference(String field, boolean showError) {
        retext.reset(field);
        if (retext.find()) {
            if (!checkTextReference(retext.getGroup(1), retext.getGroup(2), showError)) {
                if (showError) {
                    String msg = String.format("%s:%d.%d %s not found in texts",
                            shortname, locator.getLineNumber(), locator.getColumnNumber(),
                            field);
                    writeError(msg);
                }
                return false;
            }
        } else {
            if (showError) {
                writeError(String.format("%s:%d.%d %s does not match retext", shortname,
                        locator.getLineNumber(), locator.getColumnNumber(), field));
            }
            return false;
        }

        return true;
    }


    private void startElementPass3(String uri, String localName, String qName, Attributes atts) throws SAXException {
    }

    private boolean checkTextReference(String textname, String verse, boolean showError) {

        if (texts.get(textname) == null) {
            try {
                String fn = String.format("%s%s", Main.textpath, textname);
                if (!fn.endsWith("slp")) {
                    fn += ".slp";
                }
                String slp = IOAssistance.readEntireFile(fn);
                texts.put(textname, slp);
            } catch (Exception ex) {
                texts.put(textname, "?");
                if (showError) {
                    writeError(String.format("%s not found", textname));
                }
            }
        }

        return texts.get(textname).contains(verse);
    }
    

    private boolean exists(String URLName) {
        try {
            /*
            rehttp.reset(URLName);
            if (rehttp.find()) {
                URLName = rehttp.getGroup(1);
            }
             */
            
            HttpURLConnection.setFollowRedirects(false);
            // note : you may also need
            //        HttpURLConnection.setInstanceFollowRedirects(false)
            HttpURLConnection con =
                    (HttpURLConnection) new URL(URLName).openConnection();
            con.setRequestMethod("HEAD");
            int responseCode = con.getResponseCode();
            boolean rc = (responseCode == HttpURLConnection.HTTP_OK ||
                    responseCode == HttpURLConnection.HTTP_MOVED_TEMP ||
                    responseCode == HttpURLConnection.HTTP_MOVED_PERM ||
                    responseCode == HttpURLConnection.HTTP_MULT_CHOICE ||
                    responseCode == HttpURLConnection.HTTP_SEE_OTHER);
            if (rc == false) {
                rc = false;
            }
            return rc;
        } catch (Exception e) {
            return false;
        }
    }

    private void displayIllegalReference (String element, String attribute, String value) {
        System.out.format("attribute %s of element %s references a non-existent key\n");
    }

    public void endElement(String uri, String localName, String qName) throws SAXException {
        if (passNumber == 1) {
            endElementPass1(uri, localName, qName);
        } else if (passNumber == 2) {
            endElementPass2(uri, localName, qName);
        } else if (passNumber == 3) {
            endElementPass3(uri, localName, qName);
        }
    }

    private void endElementPass1(String uri, String localName, String qName) throws SAXException {
        ElementInfo ei = stack.popElement();
        if (qName.equals("locus")) {
            String to = ei.h.get("to");
            String from = ei.h.get("from");
            String facs = ei.h.get("facs");
            if (from == null && to != null) {
                writeError(String.format("%s:%d.%d locus element has to attribute but no from attribute",
                        shortname, locator.getLineNumber(), locator.getColumnNumber()));
            } else if (facs != null) {
                if (facsOK(facs)) {
                    if (from != null) {                 // make sure that there is not just a facs attribute
                        String s = null;
                        if (to == null || to.equals(from)) {
                            //s = String.format("f. %s", from);
                            s = String.format("#f%s", from);
                        } else {
                            //s = String.format("ff. %s--%s", from, to);
                            s = String.format("#f%s #f%s", from, to);
                        }

                        if (!facs.equals(s)) {
                            String msg = String.format("%s:%d.%d locus from='%s' to='%s' but facs='%s'", shortname, locator.getLineNumber(), locator.getColumnNumber(),
                                    from, to == null ? "" : String.format("'%s'", to), facs);
                            writeError(msg);
                        }
                    }
                } else {
                    String msg = String.format("%s:%d.%d Illegal facs value: %s",
                            shortname, locator.getLineNumber(), locator.getColumnNumber(), facs);
                    writeError(msg);
                }
            }

            buffer.setLength(0);
        } else if (qName.equals("formula") && collectFormula) {
                //System.out.format("%s: %s\n", shortname, formulaBuffer.toString().trim());
                formulaBuffer.setLength(0);
                collectFormula = false;
        }
    }

    private boolean facsOK(String facs) {
        boolean rc = true;

        if (facs.equals("#cover") || facs.startsWith("#add") || facs.startsWith("#acc")) {
            rc = true;
        } else {
            String[] fields = facs.split(" ");
            for (String field : fields) {
                refacs.reset(field);
                if (!refacs.find()) {
                    rc = false;
                    break;
                }
            }
        }
        return rc;
    }

    private void endElementPass2 (String uri, String localName, String qName) throws SAXException {
        if (copyBufferList.size() > 0) {
            for (CopyBuffer cb : copyBufferList) {
                cb.buffer.append(String.format("</%s>", qName));
            }
            String path = stack.getPath();
            int last = copyBufferList.size() - 1;
            CopyBuffer cb = copyBufferList.get(last);
            if (cb.path.equals(path)) {
                xmlidContent.put(cb.key, new XMLChunk(fixupAmpersand(cb.buffer.toString()), stack.getLanguage(), path));
                copyBufferList.remove(last);
            }
        }
        stack.popElement();
    }

    private String fixupAmpersand(String s) {
        return s.replaceAll("&", "&amp;");
    }

    private void endElementPass3 (String uri, String localName, String qName) throws SAXException {
    }

    public void characters(char[] ch, int start, int length) throws SAXException {
        switch (passNumber) {
            case 1:
                ElementInfo ei = stack.peekElement();
                if (ei.qName.equals("locus")) {
                    buffer.append(new String(ch, start, length));
                }
                if (collectFormula) {
                    formulaBuffer.append(new String(ch, start, length));
                }
                break;
            
            case 2: 
                if (copyBufferList.size() > 0) {
                    for (CopyBuffer cb : copyBufferList) {
                        cb.buffer.append(new String(ch, start, length));
                    }
                }
                break;
            
            case 3: break;
        }
    }

    class Location {
        int row;
        int col;
        public Location(Locator locator) {
            this.row = locator.getLineNumber();
            this.col = locator.getColumnNumber();
        }

        public int getLineNumber () {
            return row;
        }

        public int getColumnNumber () {
            return col;
        }

    }
}
