package siva;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.FSM;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.ResourceInputStream;


public class Main implements ResourceInputStream
{
    static public String htmlpath = "C:/Users/ralph/Documents/PeterScharf/SanskritLibrary/siva/HTMLnew/";
    static public String textpath = "C:/Users/ralph/Documents/PeterScharf/SITA1/SITA/texts/";
    static public String textDirectory = "texts/";          // last directory of textpath
    static public String xmlpath = "C:/Users/ralph/Documents/PeterScharf/SanskritLibrary/siva/XMLnew/";
    static public String facspath = "";

    // f=name (in XML directory)
    // p=path
    public static void main(String[] args) throws Exception
    {
        new Main(args);
    }

    public Main(String[] args) throws Exception
    {
        if (args.length != 1)
        {
            printUsage();
            return;
        }

        findComments(new File("C:/Users/ralph/Documents/LatestGrantFiles/annotations/"), new CollectComments(), 0);
        if (args != null) return;
        boolean pathsRead = true;
        File f = new File("paths.txt");
        if (f.exists()) {
            BufferedReader br = IOAssistance.openReader("paths.txt");
            String line = null;
            while ((line=br.readLine()) != null) {
                String[] fields = line.split("\\s+=\\s*", 2);
                if (fields.length == 2) {
                    String dir = fields[1];
                    char ch = dir.charAt(dir.length()-1);
                    if (ch != '\\' && ch != '/') {
                        dir += "/";
                    }
                    f = new File(dir);
                    if (!f.exists()) {
                        System.out.format("%s does not exist\n", line);
                        pathsRead = false;
                    } else {
                        String type = fields[0];
                        if (type.equals("xmlpath")) {
                            xmlpath = dir;
                        } else if (type.equals("htmlpath")) {
                            htmlpath = dir;
                        } else if (type.equals("textpath")) {
                            textpath = dir;
                        } else if (type.equals("facspath")) {
                            facspath = dir;
                        } else {
                            System.out.format("%s is not a legal type. Should be one of xmlpath, htmlpath, textpath or facspath\n", type);
                            pathsRead = false;
                        }
                    }
                } else {
                    System.out.format("%s is not correctly formatted. Should begin with xmlpath, htmlpath, textpath or facspath followed by a blank followed by a directory name\n", line);
                    pathsRead = false;
                }
            }
        }

        if (!pathsRead) {
            System.out.println("Please fix errors and run again");
            return;
        }

        Regex re = new Regex(textpath, "([^\\\\/]+)[\\\\/]$");
        if (re.find()) {
            textDirectory = re.getGroup(1) + "/";
        } else {
            System.out.format("Illegal textpath: %s\n", textpath);
            return;
        }

        //if (compareVersions()) return;

        re = new Regex(args[0], "^([pfFhi]).*=.*?([^\\s]+)$");
        if (re.find())
        {
            String command = re.getGroup(1);
            String arg = re.getGroup(2);
            if (command.equals("F"))
            {
                xmlpath = "XML-fixedup/";
                command = "f";
            }
            if (command.equals("f"))
            {
                if (arg.equals("*"))
                    formatPennCatalog();
                else
                {
                    arg = arg.toLowerCase();
                     if (!arg.endsWith(".xml"))
                        arg += ".xml";
                    f = new File(xmlpath + arg);
                    formatPennCatalog(f);
                    arg = arg.replace(".xml", ".html");
                    copyfile(new File(htmlpath + arg), htmlpath + "c.html");
                }
            }
            else if (command.equals("p"))
                extractFromPennCatalog(arg);
            else if (command.equals("i"))
                preprocessPennCatalog();
            else if (command.equals("h"))
                displayHTML();
        }
        else
            printUsage();
    }

    Hashtable<String, String> directoryPaths = new Hashtable<String, String>();
    private void findComments(File directory, CollectComments cc, int level) throws Exception {
        File[] files = directory.listFiles();
        for (File file : files) {
            if (file.isDirectory() && !file.getAbsolutePath().endsWith("Supp")) {
                directoryPaths.put(file.getName(), file.getAbsolutePath());
                findComments(file, cc, level + 1);
            } else {
                String filename = file.getName();
                if (filename.endsWith(".xml") && filename.charAt(0) != '.') {
                    char ch = filename.charAt(0);
                    if (ch == '_' || (ch >='0' && ch<='9')) {
                        cc.parse(file.getAbsolutePath());
                    }
                }
            }
        }

        String lastmss = "";

        int expectedFolioNumber = 1;
        char expectedFolioChar = 'r';

        int beginFolioNumber = 1;
        char beginFolioChar = 'r';
        int lastFolioNumber = 0;
        char lastFolioChar = ' ';

        Regex re = new Regex("", "^([^/]+)/([^/]+).*::(.*)$");
        Regex refolio = new Regex("", "^f.\\s*(\\d+)([rv])");
        if (level == 0) {
            String last = "";
//            StringBuilder sb = new StringBuilder();
            int currentFolioNumber = 0;
            char currentFolioChar = ' ';
            int count = 0;
            BufferedWriter bw = IOAssistance.openWriter("comments.txt");
            ArrayList<String> result = cc.getComments();
            Collections.sort(result);
            for (String s : result) {
                if (!s.equals(last)) {
                    last = s;
                    re.reset(s);
                    if (re.find()) {
                        String mss = re.getGroup(1);
                        String page = re.getGroup(2);
                        String folio = re.getGroup(3);
                        refolio.reset(folio);

                        if (refolio.find()) {
                            currentFolioNumber = Integer.parseInt(refolio.getGroup(1));
                            currentFolioChar = refolio.getGroup(2).charAt(0);
                        } else {
                            System.out.format("???%s\n", folio);            // should never reach here
                        }
                        boolean computeNext = true;
                        if (!mss.equals(lastmss)) {             // if at start of a mss
                            if (count > 0) {
//                                sb.append(String.format("f. %d%c -- f.%d%c", beginFolioNumber, beginFolioChar, lastFolioNumber, lastFolioChar));
//                                bw.write("****\n");
//                                bw.write(String.format("%s\n\n", sb.toString()));
//                                sb.setLength(0);
                            } else {
                                count = 0;
                            }
                            bw.write("----------------------------------------\n");
                            bw.write(mss);
                            bw.write("\n");
                            lastmss = mss;
                            beginFolioNumber = currentFolioNumber;
                            beginFolioChar = currentFolioChar;
                            if (currentFolioNumber != 1 || currentFolioChar != 'r') {
//                                sb.append(String.format("First folio is f. %d%c\n", currentFolioNumber, currentFolioChar));
                                String path = directoryPaths.get(mss);
                                bw.write(String.format("***First folio is f. %d%c (in %s)\n",
                                        currentFolioNumber, currentFolioChar, path));
                                if (currentFolioChar == 'v') {
                                    expectedFolioNumber = currentFolioChar + 1;
                                    expectedFolioChar = 'r';
                                } else {
                                    expectedFolioNumber = currentFolioChar;
                                    expectedFolioChar = 'v';
                                }
                            } else {
                                expectedFolioNumber = 1;
                                expectedFolioChar = 'v';
                            }
                            computeNext = false;
                        } else if (currentFolioNumber != expectedFolioNumber || currentFolioChar != expectedFolioChar) {
                            //sb.append(String.format("f. %d%c -- f.%d%c\n", beginFolioNumber, beginFolioChar, lastFolioNumber, lastFolioChar));
                            //sb.append(String.format("Expected: f. %d%c, Seen: f. %d%c\n",
                            //        expectedFolioNumber, expectedFolioChar, currentFolioNumber, currentFolioChar));
                            String path = directoryPaths.get(mss);
                            bw.write(String.format("***Expected: f. %d%c, Seen: f. %d%c (in %s)\n",
                                    expectedFolioNumber, expectedFolioChar, currentFolioNumber, currentFolioChar, path));
                            if (currentFolioChar == 'v') {
                                expectedFolioNumber = currentFolioNumber + 1;
                                expectedFolioChar = 'r';
                            } else {
                                expectedFolioNumber = currentFolioNumber;
                                expectedFolioChar = 'v';
                            }
                            beginFolioNumber = currentFolioNumber;
                            beginFolioChar = currentFolioChar;

                            computeNext = false;
                        }


                        if (computeNext) {
                            if (currentFolioChar == 'v') {
                                expectedFolioNumber++;
                                expectedFolioChar = 'r';
                            } else {
                                expectedFolioChar = 'v';
                            }
                        }
                        lastFolioNumber = currentFolioNumber;
                        lastFolioChar = currentFolioChar;

                        bw.write(String.format("  %s %s\n", page, folio));
                        count++;
                    } else {
                        System.out.format("????", s);
                    }
                }
            }
//            sb.append(String.format("f. %d%c -- f.%d%c", beginFolioNumber, beginFolioChar, lastFolioNumber, lastFolioChar));
//            bw.write("****\n");
//            bw.write(String.format("%s\n\n", sb.toString()));
//            sb.setLength(0);

            bw.close();
        }
    }

    private boolean compareVersions () throws Exception {
        boolean tabDelimited = false;
        Hashtable<String, LocationInfo> v1 = new PathCollection("XML").getInfo();
        Hashtable<String, LocationInfo> v2 = new PathCollection("XMLnew").getInfo();
/*
        Enumeration<String> en = v2.keys();
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            if (v1.get(key) == null) {
                LocationInfo li = v2.get(key);
                System.out.format("[%s] %s:%d.%d %d\n", key, new File(li.filename).getName(), li.row, li.col, li.count);
            }
        }
*/
        BufferedWriter bw = IOAssistance.openWriter("pathcomp.html");
        bw.write("<html><head></head><body><table border='1'><tr><th>Row</th><th>Old</th><th>New</th></tr>\n");
        ArrayList<String> a1 = sortPaths(v1);
        ArrayList<String> a2 = sortPaths(v2);

        int len1 = a1.size();
        int len2 = a2.size();
        int len = len1;
        if (len < len2) {
            len = len2;
        }

        int rownum = -1;
        int idx1 = 0;
        int idx2 = 0;
        while (idx1 < len1 || idx2 < len2) {
            rownum++;
            String p1 = idx1 < len1 ? a1.get(idx1) : "";
            String p2 = idx2 < len2 ? a2.get(idx2) : "";
            if (p1.equals(p2)) {
                if (tabDelimited) {
                    bw.write(String.format("%s\t\t%s\t\n", p1, p2));
                } else {
                    bw.write(String.format("<tr><td>%d</td><td>%s</td><td>%s</td></tr>\n", rownum, p1, p2));
                }
                idx1++;
                idx2++;
            } else if (p1.length() > 0 && p1.compareTo(p2) < 0) {
                String loc = String.format("%s", v1.get(p1).toString());
                if (tabDelimited) {
                    bw.write(String.format("%s\t%s\t\t\n", p1, loc));
                } else {
                    bw.write(String.format("<tr><td>%d</td><td>%s<br/><span style='color:red'>%s</span></td><td>&nbsp;</td></tr>\n",
                        rownum, p1, loc));
                }
                idx1++;
            } else if (p2.length() > 0){
                String loc = String.format("%s", v2.get(p2).toString());
                if (tabDelimited) {
                    bw.write(String.format("\t\t%s\t%s\n", p2, loc));
                } else {
                    bw.write(String.format("<tr><td>%d</td><td>&nbsp;</td><td>%s<br/><span style='color:red'>%s</span></td></tr>\n",
                        rownum, p2, loc));
                }
                idx2++;
            }
        }
        bw.write("</tr></body></html>");
        bw.close();
        return true;
    }

    private ArrayList<String> sortPaths(Hashtable<String, LocationInfo> h) {
        ArrayList<String> a = new ArrayList<String>();
        Enumeration<String> en = h.keys();
        while (en.hasMoreElements()) {
            a.add(en.nextElement());
        }

        Collections.sort(a, new LengthSort());

        return a;
    }

    private void copyfile(File f1, String dtFile){
    try
    {
      File f2 = new File(dtFile);

      InputStream in = new FileInputStream(f1);
      OutputStream out = new FileOutputStream(f2);  // (f2, true) to append in to out

      byte[] buf = new byte[1024];
      int len;
      while ((len = in.read(buf)) > 0)
      {
        out.write(buf, 0, len);
      }
      in.close();
      out.close();
    }
    catch(FileNotFoundException ex)
    {
      System.out.println(ex.getMessage() + " in the specified directory.");
    }
    catch(IOException e){
      System.out.println(e.getMessage());
    }
  }

    void printUsage()
    {
        System.out.println("Usage:");
        System.out.println("Assumes that all .xml file names in the XML directory are lower case!");
        System.out.println("f=<name of XML file in XML directory>, e.g. f=penn0349 converts XML/penn0394.xml to HTML/penn0394.html. (Note: the .xml is optional)");
        System.out.println("f=* will convert all files in the XML directory to html files in the HTML directory.");
        System.out.println("p=<path>, e.g. p=TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier will print out all msIdentifier elements found in the files in the XML directory.");
        System.out.println("h=*, create siva/kitchenSink.html from all .html files in siva/HTML");
    }



    void formatPennCatalog() throws Exception {
        
        String teiDir = xmlpath;
        validate();
//        generateHTMLComparison("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents", false);

//        if (teiDir != null) return;

//        String teiDir = "C:/Users/ralph/Documents/PeterScharf/SanskritLibrary/siva/XML/";       // review how old was done

        File dir = new File(teiDir);
        File[] files = dir.listFiles();

        for (File file : files) {
            String fn = file.getName();
            if (!fn.equals("index.html") && fn.endsWith(".xml")) {
                formatPennCatalog(file);
                break;
            }
        }
    }

    void validate() throws Exception {
        References.checkURL = false;
        String teiDir = xmlpath;
        File dir = new File(teiDir);
        File[] files = dir.listFiles();
        BufferedWriter bw = IOAssistance.openWriter("validationReport.txt");
        References.setFile(bw);

        // first pass, collected all xml:ids and verify that reference ids appear to be references
        int numfiles = 0;
        for (File file : files) {
            String fn = file.getName();
            {
                if (!fn.equals("index.html") && fn.endsWith(".xml")) {
                    numfiles++;
                    validateFile(file, 1);
                }
            }
        }

        if (References.wellFormednessError) {
            bw.write(String.format("Note: %d/%d files had well-formedness errors. These files will be ignored in the remaining passes. Their absence may cause other errors.\n",
                    References.notWellFormed.size(), numfiles));
        }

        // second pass, collect inplace xml and validate references
        for (File file : files) {
            String fn = file.getName();
            {
                if (!fn.equals("index.html") && fn.endsWith(".xml")) {
                    validateFile(file, 2);
                }
            }
        }

        if (References.numErrors != 0) {
            System.out.format("%d error found during validation. See file validationReport.txt\n", References.numErrors);
        }

        bw.write(String.format("*****%d errors found.*****\n", References.numErrors));

        bw.close();
    }

    void validateFile (File file, int passNumber) throws Exception {
        References r = new References(passNumber);
        r.parse(file.getAbsolutePath());
    }

    void generateHTMLComparison(String elementName, boolean isName) throws Exception {
        String teiDirOld = "C:\\Users\\ralph\\Documents\\PeterScharf\\SanskritLibrary\\siva\\XML\\";
        String teiDirNew = "C:\\Users\\ralph\\Documents\\PeterScharf\\SanskritLibrary\\siva\\XMLnew\\";
        File dir = new File(teiDirOld);
        File[] files = dir.listFiles();
        BufferedWriter bw = IOAssistance.openWriter(String.format("compare_%s.html", elementName));
        bw.write(String.format("<html><head></head><body>"));
        CompareVersions.setFile(bw);
        int count = 0;
        for (File file : files) {
            String fn = file.getName();
            if (!fn.equals("index.html") && fn.endsWith(".xml")) {
                String newfn = file.getAbsolutePath();
                newfn = newfn.replace(teiDirOld, teiDirNew);
                File newf = new File(newfn);
                if (newf.exists()) {
                    bw.write(String.format("<h2>%s</h2>\n", fn));
                    bw.write("<table border='1px' style='border-collapse:collapse;width:1200px'><tr>\n");
                    if (isName) {
                        generateHTMLFromName(file, elementName);
                        generateHTMLFromName(newf, elementName);
                    } else {
                        generateHTMLFromPath(file, elementName);
                        generateHTMLFromPath(newf, elementName);
                    }
                    bw.write("</tr></table>\n");
                }
                count++;
            }
        }

        bw.write("</body></html>\n");
        bw.close();
    }

    void generateHTMLFromName (File file, String elementName) throws Exception {
        CompareVersions cv = new CompareVersions();
        cv.setElementName(elementName);
        cv.parse(file.getAbsolutePath());
    }

    void generateHTMLFromPath (File file, String elementPath) throws Exception {
        CompareVersions cv = new CompareVersions();
        cv.setElementPath(elementPath);
        cv.parse(file.getAbsolutePath());
    }

    void displayHTML() throws Exception
    {
        SectionLabel[] sls = new SectionLabel[]
        {
            new SectionLabel("Manuscript Identifiers", "Collection"),
            new SectionLabel("Manuscript Identifiers", "Repository"),
            new SectionLabel("Manuscript Identifiers", "Institution"),
            new SectionLabel("Manuscript Identifiers", "Location"),
            new SectionLabel("Manuscript Identifiers", "Catalog"),
            new SectionLabel("Manuscript Identifiers", "Item"),
            new SectionLabel("Contents", "Work"),
            new SectionLabel("Contents", "Locus"),
            new SectionLabel("Contents", "Author"),
            new SectionLabel("Contents", "Title"),
            new SectionLabel("Contents", "Part"),
            new SectionLabel("Contents", "Cover rubric"),
            new SectionLabel("Contents", "Incipit"),
            new SectionLabel("Contents", "Explicit"),
            new SectionLabel("Contents", "Final rubric"),
            new SectionLabel("Contents", "Colophon"),
            new SectionLabel("Contents", "Filiation"),
            new SectionLabel("Contents", "Second folio"),
            new SectionLabel("Contents", "Language"),
            new SectionLabel("Physical description", "Form"),
            new SectionLabel("Physical description", "Material"),
            new SectionLabel("Physical description", "Watermark"),
            new SectionLabel("Physical description", "Extent"),
            new SectionLabel("Physical description", "Dimension"),
            new SectionLabel("Physical description", "Foliation"),
            new SectionLabel("Physical description", "Collation"),
            new SectionLabel("Physical description", "Condition"),
            new SectionLabel("Physical description", "Binding"),
            new SectionLabel("Physical description", "Seal"),
            new SectionLabel("Physical description", "Accompanying matter"),
            new SectionLabel("Physical description", "Layout"),
            new SectionLabel("Hands", "Summary"),
            new SectionLabel("Hands", "Hand"),
            new SectionLabel("Hands", "Additions"),
            new SectionLabel("Decoration", "Cover"),
            new SectionLabel("Decoration", "Border"),
            new SectionLabel("Decoration", "Illustration"),
            new SectionLabel("History", "Origin"),
            new SectionLabel("History", "Provenance"),
            new SectionLabel("History", "Acquisition"),
            new SectionLabel("Subject headings", "SubjectLC"),
            new SectionLabel("Subject headings", "SubjectSL"),
            new SectionLabel("Facsimile", "*"),
            new SectionLabel("Transcription", "*"),
        };
        BufferedWriter bw = IOAssistance.openWriter("kitchenSink.html");
        bw.write("<html><head><title>Kitchen Sink</title><meta http-equiv='Content-Type' content='text/html; charset=utf-8'/><body>");
        File dir = new File("HTML");
        File[] files = dir.listFiles();
        String header = "";
        boolean firstTime = true;
        for (SectionLabel sl : sls)
        {
            int n = 0;
            for (File file : files)
            {
                String fn = file.getName().toLowerCase();
                if (fn.startsWith("br") || fn.startsWith("penn"))
                {
                    String s = String.format("<h3>%s[%s]</h3>", sl.section, sl.label);
                    if (!header.equals(s))
                    {
                        if (!firstTime)
                            bw.write("</table>");
                        else
                            firstTime = false;
                        bw.write(s);
                        bw.write("<table>");
                        header = s;
                    }
                    new ParseHTML(file.getAbsolutePath(), sl.section, sl.label, bw);
                }
            }
        }

        bw.write("</table></body></html>");
        bw.close();
    }

    void preprocessPennCatalog() throws Exception
    {
        String originalTEIDir = "XML-matthias/";
        String tempXMLDir = "XML-renamed/";
        String marcDir = "MARC3/";
        String[] marcfiles = new String[] {"mbh_bhp.xml", "penn_skt_2records.xml", "up2618.xml"};
        Hashtable<String, MarcRecord> marcRecords = new Hashtable<String, MarcRecord>();

        renameFiles(originalTEIDir, tempXMLDir);

        Hashtable<String, String> h = mapItemNumbersToFiles(tempXMLDir);

        for (String marcfile : marcfiles)
        {
            ParseMARC pm = new ParseMARC(marcDir + marcfile, h, tempXMLDir, this, marcRecords);
        }

        File dir = new File(tempXMLDir);
        File[] xmlfiles = dir.listFiles();
        for (File xmlfile: xmlfiles)
        {
            String fn = xmlfile.getName();
            String key = String.format("%s%s", tempXMLDir, fn);
            new ParseTEI(key, ParseTEI.ACTION_INSERT, this, marcRecords.get(key));
        }

        BufferedWriter bw = IOAssistance.openWriter("empty.txt");
        PathInfo.writeEmptyElements(bw);
        bw.close();
    }

    Hashtable<String, String> renameFiles(String originalTEIDir, String tempXMLDir) throws Exception
    {
        Hashtable h = new Hashtable<String, String>();
        String[] directories = new String[]
        {
            "MSS-BhG",
            "MSS-BhP",
            "MSS-Br-Skt-Mbh",
            "Mss-Br-Te-BhP",
            "Mss-Br-Te-Mbh",
            "MSS-MBh",
            "MSS-misc",
            "MSS-stotra-as",
            "MSS-stotra-gm",
            "MSS-stotra-misc",
            "MSS-stotra-sr",
            "MSS-stotra-vsn",
        };

        for (String directory : directories)
        {
            renameXMLFiles(originalTEIDir + directory, tempXMLDir);
        }

        return h;
    }

    void renameXMLFiles(String directory, String tempXMLDir) throws Exception
    {
        File dir = new File(directory);
        File[] files = dir.listFiles();
        Regex re = new Regex("", "(\\d+)");
        for (File file : files)
        {
            String name = file.getName();
            if (name.endsWith(".xml"))
            {
                re.reset(name);
                if (re.find())
                {
                    String prefix = "?";
                    String itemNumber = re.getGroup(1);
                    if (name.startsWith("MsDescPenn"))
                        prefix = "penn";
                    else if (name.startsWith("BrMs"))
                        prefix = "br";

                    String target = String.format("%s/%s%s.xml", tempXMLDir, prefix, itemNumber);
                    copyfile(file, target);
                }
            }
        }
    }

    static Hashtable<String, String> mapItemNumbersToFiles(String teiDir) throws Exception
    {
        Hashtable h = new Hashtable<String, String>();

        getXMLFiles(teiDir, h);

        return h;
    }


    static void getXMLFiles(String directory, Hashtable<String, String> h) throws Exception
    {
        File dir = new File(directory);
        String[] files = dir.list();
        Regex re = new Regex("", "(\\d+)");
        for (String file : files)
        {
            if (file.endsWith(".xml"))
            {
                re.reset(file);
                if (re.find())
                {
                    String itemNumber = re.getGroup(1);
                    h.put(itemNumber, String.format("%s%s", directory, new File(file).getName()));
                }
            }
        }
    }



    void formatPennCatalog(File f) throws Exception
    {
        if (f.exists())
            new ParseTEI(f.getAbsolutePath(), ParseTEI.ACTION_FORMAT, this);
        else
            System.out.format("System cannot find file: %s\n", f.getAbsolutePath());
    }


    void extractFromPennCatalog(String extraction) throws Exception
    {
        String teiDir = "XML";
        BufferedWriter bw = null;
        bw = IOAssistance.openWriter("extract.xml");
        bw.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        bw.write(String.format("<extractions path='%s'>\n", extraction));
        File dir = new File(teiDir);
        File[] files = dir.listFiles();
        for (File file : files)
        {
            String fn = file.getName();
            if (!fn.equals("index.html") && fn.endsWith(".xml"))
            {
                bw.write(String.format("\n\n<file name='%s'/>\n\n", fn));
                new ParseTEI(file.getAbsolutePath(), ParseTEI.ACTION_EXTRACT, extraction, bw, this);
            }
        }
        bw.write("<allpaths>");
        ParseTEI.writeAllPaths(bw);
        bw.write("</allpaths>");
        bw.write("\n</extractions>\n");

        bw.close();
    }

    public InputStream[] getInputStreams(String sourceEncoding, String targetEncoding)
    {
        String[] fileNames = FSM.createFileNames("", sourceEncoding, targetEncoding);
        InputStream[] iss = new InputStream[fileNames.length];
        int i = 0;

        for (String name : fileNames)
        {
            String fname = String.format(".%s", name);
            try
            {
                iss[i++] = new FileInputStream(fname);
            } catch (Exception ex)
            {
            }
        }

        return iss;
    }

    class LengthSort implements Comparator<String> {

        public int compare(String o1, String o2) {
            int len1 = o1.length();
            int len2 = o1.length();
            if (len1 != len2) {
                return o1.length() - o2.length();
            } else {
                return o1.compareTo(o2);
            }
        }

    }
}
