/* Creates kitchensink.html*/
package siva;

import java.io.BufferedWriter;
import java.util.ArrayList;
import java.io.File;
import java.io.IOException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;

public class ParseHTML extends DefaultHandler //implements ErrorHandler, ContentHandler, DTDHandler, EntityResolver
{
    String currentDirectory;
    Locator locator;
    ArrayList<String> path = new ArrayList<String>();
    String section;
    String label;
    BufferedWriter bw;
    boolean inSection = false;
    boolean inRow = false;
    String filename;
    StringBuilder buffer = new StringBuilder();

    public ParseHTML(String fn, String section, String label, BufferedWriter bw)
    {
        this.bw = bw;
        this.section = section;
        this.label = label;
        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            File f = new File(fn);
            filename = f.getName();
            filename = filename.substring(0, filename.length()-5);
            saxParser.parse(f, this);

        }
        catch (Throwable t)
        {
            t.printStackTrace ();
        }
    }


    private String getPath()
    {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (String name : path)
        {
            sb.append(String.format("%s%s", delim, name));
            delim = ".";
        }

        return sb.toString();
    }


    private void pushPath(String qname, Attributes atts)
    {
        String valign = "";
        if (atts.getValue("valign") != null)
            valign = String.format("[valign='%s']", atts.getValue("valign"));
        path.add(qname+valign);
    }


    // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        System.out.format(String.format("Error parsing %s.htm: %s\n", filename, exception.getMessage()));
        throw exception;
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
        try
        {
        }
        catch(Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void endDocument() throws SAXException
    {
        try
        {
            if (inRow)
                bw.write(String.format("<tr><td valign='top'><b>%s</b></td><td>%s</td></tr>\n", filename, buffer.toString().trim()));
        }
        catch(Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        try
        {
            pushPath(qName, atts);
            if (inRow)
            {
                if (!qName.equals("br") && !qName.equals("h3"))
                    buffer.append(getTag(qName, atts));
                else if (qName.equals("h3"))
                {
                    bw.write(String.format("<tr><td valign='top'><b>%s</b></td><td>%s</td></tr>\n", filename, buffer.toString().trim()));
                    inRow = false;
                }
            }
            if (getPath().equals("html.body.h3"))
                inSection = false;
        }
        catch(Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }


    private String getTag(String qName, Attributes atts)
    {
        StringBuilder sb = new StringBuilder();

        sb.append(String.format("<%s", qName));
        int len = atts.getLength();
        for (int i=0; i<len; i++)
        {
            String name = atts.getQName(i);
            String value = atts.getValue(i);
            sb.append(String.format(" %s='%s'", name, value));
        }
        sb.append(">");
        return sb.toString();
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        try
        {
            if (getPath().equals("html.body.h3"))
            {
                if (buffer.toString().trim().equals(section))
                {
                    if (section.equals("Facsimile"))
                        section = "Facsimile";
                    inSection = true;
                    inRow = label.equals("*");
                }
                buffer.setLength(0);
            }
            else if (inSection && !inRow && getPath().equals("html.body.table.tr.td[valign='top']"))
            {
                String s = buffer.toString().trim();
                if (s.charAt(0) == 0xA0)
                    s = s.replaceAll("\u00A0", "");

                if (s.startsWith(label) || label.equals("*"))
                {
                    bw.write(String.format("<tr style='color:black'><td valign='top'><b>%s</b></td>", filename));
                    inRow = true;
                }
                buffer.setLength(0);
            }
            else if (inRow)
            {
                if (getPath().equals("html.body.table.tr.td"))
                {
                    inRow = false;
                    bw.write(String.format("%s</tr>\n", buffer.toString().trim()));
                    buffer.setLength(0);
                }
            }
            else
                buffer.setLength(0);

            if (inRow && !qName.equals("h3"))
                buffer.append(String.format("</%s>", qName));
            path.remove(path.size() - 1);
        }
        catch(Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        buffer.append(new String(ch, start, length));
    }

    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void processingInstruction(String target, String data) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void skippedEntity(String name) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    // For DTDHandler
    @Override
    public void notationDecl(String name, String publicId, String systemId) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void unparsedEntityDecl(String name, String publicId, String systemId, String notationName) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    // For EntityHandler
    @Override
    public InputSource resolveEntity(String publicId, String systemId) throws SAXException, IOException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }
}
