/*
http://www.tei-c.org/release/doc/tei-p5-doc/en/html/index-toc.html
*/
/*
 * Currently only penn0452, penn0488, penn0489, penn0490 have zones
 */

package siva;

import java.io.BufferedWriter;
import java.util.ArrayList;
import java.io.File;
import java.io.StringReader;
import java.util.Hashtable;
import org.sanskritlibrary.XML;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.ext.DefaultHandler2;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.FSM;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.ResourceInputStream;
import org.sanskritlibrary.SubjectClassifications;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


public class ParseTEI extends DefaultHandler2 //implements ErrorHandler, ContentHandler, DTDHandler, EntityResolver
{
    public enum RenderType {
        COLOR, SMALLCAPS;  //; is optional
}
//    static public final String textDirectory = "texts/";
    static public final String textURL = "http://sanskrit1d.ccv.brown.edu/tomcat/Announce/";
    static public final String indent = "&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;&#160;";
    static public final String xmlindent = "                                                                                                              ";
    static public final String rowTemplate = "<tr><td valign='top'>%s%s:</td><td>%s</td></tr>\n";
    static public final String rowTemplateNoLabel = "<tr><td valign='top'></td><td>%s</td></tr>\n";
    static public final String rowTemplateHeader = "<tr><th valign='top'>%s:</th><th>%s</th></tr>\n";
    static public final String rowTemplateNoLabelHeader = "<tr><th valign='top'></th><th>%s</th></tr>\n";
    static public final int ACTION_INSERT = 1;
    static public final int ACTION_FORMAT = 2;
    static public final int ACTION_EXTRACT = 3;
    static public String target = "roman";
    static private SubjectClassifications sc = null;
    static private Hashtable<String, String> classifications;
    static Hashtable<String, FSM[]> fsmh = new Hashtable<String, FSM[]>();
    static Hashtable<String, Boolean>onetag = new Hashtable<String, Boolean>();
    static Hashtable<String, Boolean> embeddedTags = new Hashtable<String, Boolean>();
    static Hashtable<String, RenderType> rendAttributeValues = new Hashtable<String, RenderType>();
    static
    {
        rendAttributeValues.put("smallcap", RenderType.SMALLCAPS);
        rendAttributeValues.put("red", RenderType.COLOR);
        onetag.put("lb", Boolean.TRUE);
        onetag.put("pb", Boolean.TRUE);
        onetag.put("catRef", Boolean.TRUE);
        onetag.put("graphic", Boolean.TRUE);
        onetag.put("cb", Boolean.TRUE);
        onetag.put("milestone", Boolean.TRUE);
        onetag.put("refState", Boolean.TRUE);
        onetag.put("caesura", Boolean.TRUE);
        onetag.put("link", Boolean.TRUE);
        onetag.put("anchor", Boolean.TRUE);

        embeddedTags.put("add", Boolean.TRUE);
        embeddedTags.put("biblScope", Boolean.TRUE);
        embeddedTags.put("catRef", Boolean.TRUE);
        embeddedTags.put("corr", Boolean.TRUE);
        embeddedTags.put("persName", Boolean.TRUE);
        embeddedTags.put("date", Boolean.TRUE);
        embeddedTags.put("desc", Boolean.TRUE);
        embeddedTags.put("del", Boolean.TRUE);
        embeddedTags.put("gap", Boolean.TRUE);
        embeddedTags.put("graphic", Boolean.TRUE);
        embeddedTags.put("sic", Boolean.TRUE);
        embeddedTags.put("unclear", Boolean.TRUE);
        embeddedTags.put("foreign", Boolean.TRUE);
    }

    private int debugging = 0;
    private int tagLevel = 0;
    private boolean biblBreak = true;
    String pendingText = null;
    private PathInfo pi;
    ResourceInputStream ris;
    String currentDirectory;
    Locator locator;
    ArrayList<String> path = new ArrayList<String>();
    int actionType;
    int debug = 0;
    int ignoreUntil = -1;
    BufferedWriter bw;
    boolean endCommentWithLineFeed = true;      // for comments before root element
    String extractionPath;
    StringBuilder contents = new StringBuilder();
    StringBuilder manuscriptIdentifiers = new StringBuilder();
    StringBuilder physicalDescription = new StringBuilder();
    private Hashtable<String, Boolean> references = null;
    private Hashtable<String, String> targets = null;

    StringBuilder hands = new StringBuilder();

    StringBuilder decoration = new StringBuilder();
    StringBuilder history = new StringBuilder();

    StringBuilder subjectHeadings = new StringBuilder();

    StringBuilder facsimile = new StringBuilder();

    StringBuilder transcription = new StringBuilder();
    StringBuilder changes = new StringBuilder();

    ArrayList<ChangeInfo>changeItems = new ArrayList<ChangeInfo>();

    String theFileName;

    ArrayList<String> languageStack = new ArrayList<String>();
    StringBuilder currentBuffer = new StringBuilder();
    boolean biblLink = false;
    String newElement = null;
    MarcRecord mr = null;
    boolean insertMARC = true;;
    String measureQuantity = null;

    StringBuilder insertBuffer = new StringBuilder();

    public ParseTEI(String fileName, int actionType, ResourceInputStream ris, MarcRecord mr) throws Exception
    {
        this(fileName, actionType, null, null, ris, mr);
    }

    public ParseTEI(String fileName, int actionType, ResourceInputStream ris) throws Exception
    {
        this(fileName, actionType, null, null, ris, null);
    }

    public ParseTEI(String fileName, int actionType, String extractionPath, BufferedWriter bw, ResourceInputStream ris) throws Exception
    {
        this(fileName, actionType, extractionPath, bw, ris, null);
    }

    public ParseTEI(String fileName, int actionType, String extractionPath, BufferedWriter bw, ResourceInputStream ris, MarcRecord mr) throws Exception
    {
        this.ris = ris;
        this.mr = mr;
        File fn = new File(fileName);
        String name = fn.getName();
        theFileName = name;             // for debugging
        this.actionType = actionType;
        this.extractionPath = extractionPath;
        this.bw = bw;

        if (sc == null)
        {
            sc = new SubjectClassifications(new File("IndicSubjectClassification.xml"), fsmh, ris);
            classifications = sc.getClassifications();
        }

        pi = new PathInfo();

        if (actionType == ACTION_FORMAT || actionType == ACTION_EXTRACT)
        {
            pi.setResourceInputStream(ris);
            pi.setFSMH(fsmh);
            reserveEntries();

            // Set keys
            pi.key("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote", "type");

            // Set handlers (will be called when end tag of path is reached)
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc", new Handler_physDesc());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier", new Handler_msIdentifierStruct());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents", new Handler_msContents());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions", new Handler_handDesc());  // wait until additions to process handDesc
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc", new Handler_decoDesc());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history", new Handler_history());
            pi.handler("TEI.teiHeader.profileDesc", new Handler_profileDesc());
            pi.handler("TEI.facsimile", new Handler_facsimile());
            pi.handler("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist", new Handler_recordHist());
            pi.handler("TEI.text", new Handler_transcription());
        }

        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try
        {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            if (actionType == ACTION_INSERT)
            {
                // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description,
                saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
                this.bw = IOAssistance.openWriter("./XML-fixedup/" + name);
                this.bw.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
                insertMARC = mr != null;
            }
            else if (actionType == ACTION_FORMAT)
            {
                name = name.replace(".xml", ".html");
                this.bw = IOAssistance.openWriter(Main.htmlpath + name);
                this.bw.write(String.format("<html><meta http-equiv='Content-Type' content='text/html; charset=utf-8'/><head><title>%s</title><style type='text/css'>br{margin-bottom:8px}</style></head><body>\n", name));
            }
            else if (actionType == ACTION_EXTRACT)
            {
                // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description,
                endCommentWithLineFeed = true;
                saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
            }

            saxParser.parse(fn, this);
        }
        catch (Throwable t)
        {
            t.printStackTrace ();
        }
    }

    private void displayError(String message)
    {
        System.out.println(String.format("%s (%d): %s\n", theFileName, locator.getLineNumber(), message));
    }

    public static void writeAllPaths(BufferedWriter bw) throws Exception
    {
        PathInfo.writeStatistics(bw);
    }


    private String getPath()
    {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (String name : path)
        {
            sb.append(String.format("%s%s", delim, name));
            delim = ".";
        }

        return sb.toString();
    }


    private void pushPath(String qname)
    {
        path.add(qname);
    }

    private String popPath()
    {
        int top = path.size()-1;
        String s = path.get(top);
        path.remove(top);
        return s;
    }

    // For ErrorHandler
    @Override
    public void warning(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void error(SAXParseException exception) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException
    {
        String message = String.format("%d:%d %s\n", locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        System.out.println(message);
        throw exception;
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
    }

    @Override
    public void endDocument() throws SAXException
    {
        if (actionType == ACTION_FORMAT || actionType == ACTION_INSERT)
        {
            try
            {
                if (actionType == ACTION_FORMAT)
                {
                    bw.write(fixup(manuscriptIdentifiers.toString()));
                    bw.write(fixup(contents.toString()));
                    bw.write(fixup(physicalDescription.toString()));
                    bw.write(fixup(hands.toString()));
                    bw.write(fixup(decoration.toString()));
                    bw.write(fixup(history.toString()));
                    bw.write(fixup(subjectHeadings.toString()));
                    bw.write(fixup(facsimile.toString()));
                    bw.write(fixup(transcription.toString()));
                    bw.write(fixup(changes.toString()));
                    bw.write("</body></html>");
                }
                else if (actionType == ACTION_INSERT)
                {
                    bw.write(insertBuffer.toString());
                    insertBuffer.setLength(0);
                }
                bw.close();
            }
            catch (Exception ex)
            {
                throw new SAXException(ex.getMessage());
            }
        }
    }

    private String fixup(String s)
    {
        return s.replaceAll("& ", "&amp; ");
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException
    {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        debug++;
        try
        {
            if (actionType == ACTION_INSERT)
                startElementInsert(uri, localName, qName, atts);
            else if (actionType == ACTION_FORMAT)
                startElementFormat(uri, localName, qName, atts);
            else if (actionType == ACTION_EXTRACT)
                startElementExtract(uri, localName, qName, atts);
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        try
        {
            if (actionType == ACTION_INSERT)
                endElementInsert(uri, localName, qName);
            else if (actionType == ACTION_FORMAT)
                endElementFormat(uri, localName, qName);
            else if (actionType == ACTION_EXTRACT)
                endElementExtract(uri, localName, qName);
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }


    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        try
        {
            if (actionType == ACTION_INSERT)
                charactersInsert(ch, start, length);
            else if (actionType == ACTION_FORMAT)
                charactersFormat(ch, start, length);
            else if (actionType == ACTION_EXTRACT)
                charactersExtract(ch, start, length);
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    public void startElementInsert(String uri, String localName, String qName, Attributes atts) throws Exception
    {
        if (qName.equals("lb"))
            qName = "lb";
        if (newElement != null && newElement.equals(pi.currentPath.toString()))     // this means that charactersInsert was not called
        {
            insertBuffer.append(">");
            newElement = null;
        }

        if (embeddedTags.get(qName) == null)
        {
            int i = insertBuffer.length() - 1;
            while (i >= 0)
            {
                if (insertBuffer.charAt(i) <= 0x20)
                {
                    insertBuffer.deleteCharAt(i);
                    i--;
                }
                else
                    break;
            }
            insertBuffer.append("\n");
            insertBuffer.append(getXMLIndentation(tagLevel));
        }
        
        pi.pushElement(qName, atts);
        tagLevel++;
        newElement = pi.currentPath.toString();
        if (qName.equals("TEI"))
            endCommentWithLineFeed = false;      // comments before root element need line feed, those after root element do not.

        insertBuffer.append("<");
        insertBuffer.append(qName);
        int numatts = atts.getLength();
        for (int i=0; i<numatts; i++)
        {
            String name = atts.getQName(i);
            String value = atts.getValue(i);
            insertBuffer.append(String.format(" %s=\"%s\"", name, value));
        }
        if (newElement.equals("TEI.teiHeader.fileDesc.titleStmt.title"))
        {
            insertBuffer.append(">");
            Regex re = new Regex(theFileName, "(.+?)(\\d+)");
            if (re.find())
            {
                String x = re.getGroup(1).toLowerCase();
                String y = re.getGroup(2);
                if (x.equals("penn"))
                    x = "UPenn";
                else if (x.equals("br"))
                    x = "Brown";
                String title = String.format("SLMsCat %s %s", x, y);
                insertBuffer.append(title);
                newElement = null;
            }
        }
        else if (newElement.endsWith(".title"))
        {
            if (atts.getValue("rend") == null)
            {
                insertBuffer.append(" rend='smallcap'");
            }
        }
        else if (newElement.equals("TEI.teiHeader"))
        {
            if (atts.getValue("xml:lang") == null)
            {
                insertBuffer.append(" xml:lang='en'");
            }
        }
        else if (newElement.equals("TEI.text"))
        {
            if (atts.getValue("xml:lang") == null)
            {
                if (theFileName.startsWith("penn"))
                    insertBuffer.append(" xml:lang='sa-Latn-x-KH'");
                else
                    insertBuffer.append(" xml:lang='sa-Latn-x-SLP1'");
            }
        }
        else if (newElement.endsWith(".pb"))
        {
            if (atts.getValue("type") == null)
                insertBuffer.append(" type='folio'");
        }
        else if (newElement.endsWith(".measure"))
        {
            if (atts.getValue("quantity") != null)
                measureQuantity = atts.getValue("quantity");
            else
                measureQuantity = null;
        }
    }

    public void endElementInsert(String uri, String localName, String qName) throws Exception
    {
        if (newElement != null && newElement.equals(pi.currentPath.toString()))     // have <elem att=.../> instead of <elem att=...>...</elem>
        {
            newElement = null;
            if (measureQuantity != null)
            {
                String unit = "folios";
                if (measureQuantity.equals("1"))
                    unit = "folio";
                insertBuffer.append(String.format(">%s %s</measure>", measureQuantity, unit));
                measureQuantity = null;
            }
            else if (onetag.get(qName) != null)
            {
                insertBuffer.append("/>");
                insertBuffer.append(String.format("\n%s", getXMLIndentation(tagLevel - 1)));
            }
            else
                insertBuffer.append(String.format("></%s>", qName));
            
            pi.rememberEmptyElement(qName);
        }
        else
        {
            String s = "";
            if (embeddedTags.get(qName) == null && containsEOL(qName))
            {
                trimRight();
                s = String.format("\n%s", getXMLIndentation(tagLevel - 1));
            }
            insertBuffer.append(String.format("%s</%s>", s, qName));
        }

        if (insertMARC && pi.currentKey.toString().equals("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier.altIdentifier"))
        {
            insertMARC = false;
            writeMARCIdentifier("marc:leader", mr.leader);
            writeMARCIdentifier("marc:oclc", mr.control001);
            writeMARCIdentifier("marc:data", mr.control008);
            writeMARCIdentifier("marc:institution", mr.datafield035);
            writeMARCIdentifier("marc:master", mr.datafield079);
        }
        pi.popElement(qName);
        tagLevel--;
    }

    private void writeMARCIdentifier(String type, String idno) throws Exception
    {
        String margin = getXMLIndentation(tagLevel-1);
        String margin2 = margin + "  ";
        insertBuffer.append(String.format("\n%s<altIdentifier type='%s'>\n%s<idno>%s</idno>\n%s</altIdentifier>",
                            margin, type, margin2, idno, margin));
    }

    public void charactersInsert(char[] ch, int start, int length) throws Exception
    {
        if (newElement != null)
        {
            newElement = null;
            insertBuffer.append(">");
            while(length > 0)
            {
                if (ch[start] > 0x20)
                    break;
                else
                {
                    start++;
                    length--;
                }
            }
        }
        String s = new String(ch, start, length);
        s = s.replace("&", "&amp;");
        String ind = String.format("\n%s", getXMLIndentation(tagLevel));
        s = s.replaceAll("\\s*\n\\s*", ind);
        insertBuffer.append(s);
    }


    private void trimRight()
    {
        int n = insertBuffer.length() - 1;
        while (n >= 0)
        {
            if (insertBuffer.charAt(n) <= 0x20)
                insertBuffer.deleteCharAt(n--);
            else
                break;
        }
    }

    private boolean containsEOL(String qName)
    {
        int namelen = qName.length();
        int n = insertBuffer.length() - 1;
        while (n > 0)
        {
            char ch = insertBuffer.charAt(n);
            if (ch == '\n')
                return true;
            else if (ch == '<')         // check if we are at the start tag <qName>
            {
                if (n > namelen)        // just to be safe
                {
                    int i = n+1;          // start of tag name
                    boolean found = true;
                    for (int j=0; j<namelen; j++, i++)
                    {
                        if (insertBuffer.charAt(i) != qName.charAt(j))      // we haven't found the start tag
                        {
                            found = false;
                            break;
                        }
                    }
                    if (found)              
                    {
                        ch = insertBuffer.charAt(i);
                        if (ch == ' ' || ch == '>')
                            return false;    // if we have found the start tag, this means that we did not encounter any \n chars
                        // otherwise, keep going.
                    }
                }
                else
                    return false;           // shouldn't ever reach here.
            }
            n--;
        }
        
        return false;       // to make the compiler happy
    }

    private String getXMLIndentation(int level)
    {
        return xmlindent.substring(0, level*2);
    }
    

    public void startElementFormat(String uri, String localName, String qName, Attributes atts) throws Exception
    {
        pi.pushElement(qName, atts);
    }

    public void endElementFormat(String uri, String localName, String qName) throws Exception
    {
        if (qName.equals("rubric")) {
            qName = "rubric";
        }
        pi.popElement(qName);
    }

    public void charactersFormat(char[] ch, int start, int length) throws Exception
    {
        pi.append(new String(ch, start, length));
    }

    public void startElementExtract(String uri, String localName, String qName, Attributes atts) throws Exception
    {
        pushPath(qName);
        String thepath = getPath();
        pi.recordStatistics(qName, thepath, atts, theFileName);

        if (thepath.startsWith(extractionPath))
        {
            int len = extractionPath.length();
            if (thepath.length() == len || thepath.charAt(len) == '.')
            {
                pi.rememberFile(theFileName);
                bw.write("<");
                bw.write(qName);
                int numatts = atts.getLength();
                for (int i=0; i<numatts; i++)
                {
                    String name = atts.getQName(i);
                    String value = atts.getValue(i);
                    bw.write(String.format(" %s=\"%s\"", name, value));
                }
                bw.write(">");
            }
        }
    }

    public void endElementExtract(String uri, String localName, String qName) throws Exception
    {
        String thepath = getPath();
        if (thepath.startsWith(extractionPath))
        {
            int len = extractionPath.length();
            if (thepath.length() == len || thepath.charAt(len-1) == '.')
                bw.write(String.format("</%s>", qName));
        }
        if (thepath.equals(extractionPath))
            bw.write("\n");
        popPath();
    }

    public void charactersExtract(char[] ch, int start, int length) throws Exception
    {
        String thepath = getPath();
        if (thepath.startsWith(extractionPath))
        {
            int len = extractionPath.length();
            if (thepath.length() == len || thepath.charAt(len-1) == '.')
                bw.write(new String(ch, start, length));
        }
    }

    @Override
    public void comment(char[] ch, int start, int length) throws SAXException
    {
        try
        {
            boolean displayComment = false;
            if (actionType == ACTION_INSERT)
                displayComment = true;
            else if (actionType == ACTION_EXTRACT && getPath().equals(extractionPath))
                displayComment = true;
            StringBuilder tsb = new StringBuilder();
            if (displayComment)
            {
                if (newElement != null)
                {
                    newElement =  null;
                    tsb.append(">");
                }
                tsb.append("<!--");
                tsb.append(new String(ch, start, length));
                tsb.append("-->");
                if (endCommentWithLineFeed)
                    tsb.append("\n");

                if (actionType == ACTION_INSERT)
                    insertBuffer.append(tsb.toString());
                else
                    bw.write(tsb.toString());
            }
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void processingInstruction(String target, String data) throws SAXException
    {
        if (bw == null || actionType == ACTION_FORMAT)
            return;

        if (extractionPath != null && !getPath().equals(extractionPath))
            return;

        try
        {
            bw.write("<?");
            bw.write(target);
            bw.write(" ");
            bw.write(data);
            bw.write("?>\n");
        }
        catch (Exception ex)
        {
            throw new SAXException(ex.getMessage());
        }
    }

    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
    }



    private void reserveEntries()
    {
        // Identifiers (handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier is reached)
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier");         // mixed bag

        // Contents (handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents is reached)
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents");

				// handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions is reached.
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions");      // specialPara

				// handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc is reached
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='color']");    // specialPara
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='border']");   // specialPara
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='illustration']"); // specialPara

        // Physical description (handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc is reached)
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc");        // mixed bag
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.bindingDesc");        // bunch of conditions
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.sealDesc");        // bunch of conditions
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.accMat");            // specialPara
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc");       // bunch


        // History (handler called when end element TEI.teiHeader.fileDesc.sourceDesc.msDesc.history is reached)
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.origin");          // specialPara
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.provenance");      // specialPara
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.acquisition");     // specialPara

        // Subject headings (handler called when end element TEI.teiHeader.profileDesc is reached)
        pi.put("TEI.teiHeader.profileDesc");

        // Facsimile (handler called when end element TEI.facsimile is reached)
        pi.put("TEI.facsimile");

        // Transcription (handler called when end element TEI.text is reached)
        pi.put("TEI.text");         // kitchen sink

        // Change (Handler called when end element of TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist is reached)
        pi.put("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist");
    }


    public void writeRow(StringBuilder sb, String value)
    {
        sb.append(value);
    }

    public void writeRow(StringBuilder sb, String type, String value, int level)
    {
        if (type.length() == 0)
            sb.append(String.format(rowTemplateNoLabel, value));
        else if (value.trim().length() > 0)
            sb.append(String.format(rowTemplate, getHTMLIndentation(level), type, value));
    }

    private String getHTMLIndentation(int level)
    {
        return indent.substring(0, 6*3*level);
    }

    public void writeRow(StringBuilder sb, String type, String value)
    {
        writeRow(sb, type, value, 0);
    }

    public void writeRowHeader(StringBuilder sb, String type, String value)
    {
        if (type.length() == 0)
            sb.append(String.format(rowTemplateNoLabelHeader, value));
        else if (value.trim().length() > 0)
            sb.append(String.format(rowTemplateHeader, type, value));
    }

    public void writeXML(Node node, StringBuilder sb) throws Exception
    {
        String tagName = writeStartTag(node, sb);
        NodeList list = node.getChildNodes();
        int len = list.getLength();
        if (len > 0)
            sb.append("&gt;");
        for (int i=0; i<len; i++)
        {
            writeXML(list.item(i), sb);
        }
        if (tagName != null)
        {
            if (len == 0)
                sb.append("/&gt;");
            else
                sb.append(String.format("&lt;/%s&gt;", tagName));
        }
    }

    public void writeXML (Node node, StringBuilder sb, String backgroundColor, String textColor) throws Exception {
        sb.append(String.format("<span style='background-color:%s; color:%s'>", backgroundColor, textColor));
        writeXML(node, sb);
        sb.append("</span>");
    }


    public String writeStartTag(Node node, StringBuilder sb)
    {
        if (node.getNodeType() == Node.TEXT_NODE)
        {
            sb.append(node.getTextContent());
            return null;
        }
        else
        {
            sb.append(String.format("&lt;%s", node.getNodeName()));
            NamedNodeMap atts = node.getAttributes();
            int len = atts.getLength();
            for (int i=0; i<len; i++)
            {
                Node att = atts.item(i);
                sb.append(String.format(" %s='%s'",
                        att.getNodeName(), att.getNodeValue()));
            }
            return node.getNodeName();
        }
    }


    String getAttributeValue(XML xml, String path, String attname)
    {
        String value = null;
        try
        {
            Node node = xml.getNode(path);
            value = getAttributeValue(node, attname);
        }
        catch (Exception ex)
        {
            
        }
        
        return value;
    }

    String getAttributeValue(Node node, String attname)
    {
        String value = null;
        try
        {
            if (node != null)
            {
                NamedNodeMap atts = node.getAttributes();
                Node att = atts.getNamedItem(attname);
                if (att != null)
                    value = att.getNodeValue();
            }
        }
        catch (Exception ex)
        {
            displayError(String.format("Error in getAttributeValue: %s", ex.getMessage()));
        }

        return value;
    }

    String getMarkedText(Node node, String parentlang, boolean manuscriptView)
    {
        String lastseg = "";
        boolean unknown = false;
        StringBuilder sb = new StringBuilder();
        try
        {
            debugging++;
            if (node != null)
            {
                NamedNodeMap atts = node.getAttributes();
                Node rendAttribute = atts.getNamedItem("rend");
                boolean isSpan = false;
                if (rendAttribute != null) {
                    String rend = rendAttribute.getNodeValue();
                    if (rendAttributeValues.get(rend) != null && rendAttributeValues.get(rend) == RenderType.COLOR) {
                        isSpan = true;
                        sb.append(String.format("<span style='color:%s'>", rend));
                    }
                }
                String lang = getLanguage(node, parentlang);
                NodeList list = node.getChildNodes();
                int len = list.getLength();
                for (int i=0; i<len; i++)
                {
                    node = list.item(i);
                    String _lang = getLanguage(node, lang);
                    String nodeName = node.getNodeName();
                    boolean isTextNode = node.getNodeType() == Node.TEXT_NODE;
                    if (pendingText != null)        // probably a conditional break that will be ignored if punctuation is next.
                    {
                        if (!isTextNode)
                            sb.append(pendingText);
                        else
                        {
                            String text = node.getNodeValue();
                            if (text.length() > 0)
                            {
                                char ch = text.charAt(0);
                                if (ch != '.' && ch != ',')
                                    sb.append(pendingText);
                            }
                        }
                        pendingText = null;
                    }

                    if (isTextNode)
                    {
                        String text = node.getNodeValue();
                        if (lang != null)
                            text = pi.transcode(text, lang);
                        sb.append(text);
                    }
                    else if (nodeName.equals("foreign"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("persName"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("origPlace"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("origDate"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("quote"))
                    {
                        String quote = getMarkedText(node, _lang, manuscriptView).trim();
                        if (quote.length() > 0)
                        {
                            if (isQuotable(node))
                            {
                                boolean needStartQuote = true;
                                boolean needEndQuote = true;

                                if (sb.length() > 0)
                                    sb.append(" ");         // if not needed, will be collapsed by HTML

                                if (quote.startsWith("<p>"))    // check if need to push start quote inside of <p>
                                {
                                    quote = String.format("<p>&#8220;%s", quote.substring(3));    // put start quote inside paragraph
                                    needStartQuote = false;
                                }

                                if (quote.endsWith("</p>"))     // check if need to push end quote inside of </p>
                                {
                                    quote = String.format("%s&#8221;</p>", quote.substring(0, quote.length()-4));
                                    needEndQuote = false;
                                }

                                if (needStartQuote)
                                    sb.append("&#8220;");
                                sb.append(quote);
                                if (needEndQuote)
                                    sb.append("&#8221; ");
                            }
                            else
                                sb.append(quote);
                        }
                    }
                    else if (nodeName.equals("p"))
                    {
                        String p = getMarkedText(node, _lang, manuscriptView);

                        if (p.trim().length() > 0)
                        {
                            boolean italics = false;
                            sb.append("<p>");
                            String rend = getAttributeValue(node, "rend");
                            if (rend != null)
                                italics = rend.equals("italic");
                            if (rend != null && !italics)
                                System.out.format("rend attribute of p is %s\n", rend);
                            if (italics)
                                sb.append("<i>");
                            sb.append(p);
                            if (italics)
                                sb.append("</i>");
                            sb.append("</p>");
                        }
                    }
                    else if (nodeName.equals("material"))
                    {
                        sb.append("<a href='todo.html'>");
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                        sb.append("</a>");
                    }
                    else if (nodeName.equals("note"))
                    {
                        newLine(sb);
                        String resp = getAttributeValue(node, "resp");
                        if (resp != null)
                            resp = String.format("(%s) ", resp);
                        else
                            resp = "";
                        sb.append(String.format("Note: %s", resp));
                        sb.append(getMarkedText(node, "eng", manuscriptView));          // assume it is English
                        sb.append("<br/>");
                    }
                    else if (nodeName.equals("locus"))
                    {
                        String locus = get_locus(node, "eng"); // ! force language to be english
                        if (locus.trim().length() > 0)
                        {
                            sb.append(locus);          
                            pendingText = ":<br/>";
                        }
                    }
                    else if (nodeName.equals("lb"))
                    {
                        newLine(sb);
                        String n = getAttributeValue(node, "n");
                        if (n != null)
                            sb.append(String.format("Line %s:<br/>", n));
                    }
                    else if (nodeName.equals("seg"))
                    {
                        String seg = getMarkedText(node, lang, manuscriptView);
                        sb.append(seg);
                        if ("aAiIuUfFxXeEoOMH".contains(lastseg))
                            sb.append(" ");
                        lastseg = seg.substring(seg.length()-1);
                    }
                    else if (nodeName.equals("bibl"))
                    {
                        if (biblBreak)          // always true except for acquisitions
                            newLine(sb);
                        sb.append(get_bibl(node, _lang, manuscriptView));
                    }
                    else if (nodeName.equals("title"))
                    {
                        String title = getMarkedText(node, _lang, manuscriptView);

                        if (title.trim().length() > 0)
                        {
                            boolean smallcap = false;
                            String rend = getAttributeValue(node, "rend");
                            if (rend != null)
                                smallcap = rend.equals("smallcap");
                            else
                                smallcap = true;
                            if (rend != null && !smallcap)
                                System.out.format("rend attribute of title is %s\n", rend);
                            if (smallcap)
                                sb.append("<h4>");
                            else
                                sb.append("<p>");
                            sb.append(title);
                            if (smallcap)
                                sb.append("</h4>");
                            else
                                sb.append("</p>");
                        }
                    }
                    else if (nodeName.equals("date"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("placeName"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("add"))
                    {
                        sb.append("(");
                    }
                    else if (nodeName.equals("corr"))
                    {
                        sb.append("{");
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                        sb.append("}");
                    }
                    else if (nodeName.equals("del"))
                    {
                        sb.append("[");
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                        sb.append("]");
                    }
                    else if (nodeName.equals("sic"))
                    {
                        sb.append("[");
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                        sb.append("]");
                    }
                    else if (nodeName.equals("unclear"))
                    {
                        sb.append("&#8224;");
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                        sb.append("&#8224;");
                    }
                    else if (nodeName.equals("desc"))
                    {
                        sb.append("(");
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                        sb.append(")");
                    }
                    else if (nodeName.equals("lg"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("gap"))
                        sb.append(get_gap(node, _lang, manuscriptView));
                    else if (nodeName.equals("l"))
                    {
                        newLine(sb);
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    }
                    else if (nodeName.equals("ab"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("pb"))
                    {
                        String type = getAttributeValue(node, "type");
                        if (type == null || type.equals("folio"))
                        {
                            String n = getAttributeValue(node, "n");
                            if (n == null)
                                n = "?";
                            newLine(sb);
                            sb.append(String.format("f. %s:<br/>", n));
                        }
                        else
                            newLine(sb);
                        // sb.append(getMarkedText(node, _lang));  // not needed because pb has no content.
                    }
                    else if (nodeName.equals("div"))
                    {
                        String type = getAttributeValue(node, "type");
                        if (type != null)
                        {
                            if (type.equals("mUla"))
                                sb.append(String.format("%s:<br/>", "m\u016b"));
                            else if (type.equals("commentary"))
                                sb.append("Commentary:<br/>");
                            else if (type.equals("subcommentary"))
                                sb.append("Subcommentary:<br/>");
                            else
                            {
                                String n = getAttributeValue(node, "n");
                                if (n == null)
                                    n = "";
                                sb.append(String.format("%s %s:<br/>", type, n));
                            }
                        }
                        else
                            newLine(sb);
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    }
                    else if (nodeName.equals("body"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("sp"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("speaker"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("orgName"))
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    else if (nodeName.equals("ref")) {
                        processRef(node, sb);
                    } else if (nodeName.equals("head")) {
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    }else if (nodeName.equals("trailer")) {
                        sb.append("<div style='color:red'>");
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                        sb.append("</div>");
                    } else if (nodeName.equals("s")) {
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    } else if (nodeName.equals("item")) {
                        sb.append(getMarkedText(node, _lang, manuscriptView));
                    } else if (nodeName.equals("cb")) {
                        sb.append("<br/>");
                    }
                    else {
                        unknown = true;
                        displayError(String.format("Unexpected element: %s. Copied directly to HTML (search for <%s>)",
                                nodeName, nodeName));
                        writeXML(node, sb, "red", "white");
                    }
                }

                if (isSpan) {
                    sb.append("</span>");
                }
            }
        }
        catch (Exception ex)
        {
            displayError(String.format("Error in getMarkedText: %s", ex.getMessage()));
        }

        if (pendingText != null)
        {
            sb.append(pendingText);
            pendingText = null;
        }

        return sb.toString();
    }


    private void processRef(Node node, StringBuilder sb) throws Exception {
        Node tn = node.getAttributes().getNamedItem("target");
        if (tn != null) {
            String tgt = tn.getNodeValue();
            if (tgt != null && tgt.length() > 0) {
                if (tgt.startsWith(Main.textDirectory)) {
                    Regex re = new Regex(tgt, Main.textDirectory + "(.*)");
                    if (re.find()) {
                        tgt = String.format("%s?doc=%s", textURL, re.getGroup(1).replace("#", "*"));
                    } else {
                        System.out.format("ref target '%s' has unknown text directory. Used as is\n", tgt);
                    }
                } else if (tgt.charAt(0) != '#') {
                    System.out.format("ref target '%s' is missing #. One added.\n", tgt);
                    tgt = "#" + tgt;
                }
            }
            sb.append(String.format("<a target='x' href='%s'>%s</a>", tgt, node.getTextContent()));
        } else {
            displayError("Missing or empty target attribute of ref element. Search for <ref ...");
            writeXML(node, sb, "red", "white");
        }

    }

    private boolean isQuotable(Node node)
    {
        NodeList list = node.getChildNodes();
        int len = list.getLength();
        boolean hasTitle= false;
        boolean hasParagraph = false;
        for (int i=0; i<len && (!hasTitle || !hasParagraph); i++)
        {
            Node _node = list.item(i);
            String nodeName = _node.getNodeName();
            if (nodeName.equals("title"))
                hasTitle = true;
            else if (nodeName.equals("p"))
                hasParagraph = true;
        }

        return !(hasTitle && hasParagraph);     // if it doesn't have a title and paragraph within it is quotable
    }

    boolean nodeExists(Node node, String name) throws Exception {
       NodeList _list = XML.getRelativeNodeList(node, name);
       return (_list != null && _list.getLength() > 0 && _list.item(0).getChildNodes().getLength() > 0);
    }

    private void newLine(StringBuilder sb)
    {
        String s = sb.toString().trim();
        if (s.length() != 0 && !s.endsWith("<br/>"))
            sb.append("<br/>");
    }

    String getLanguage(Node node, String lang)
    {
        NamedNodeMap atts = node.getAttributes();
        if (atts != null && atts.getNamedItem("xml:lang") != null)
            return atts.getNamedItem("xml:lang").getNodeValue();
        else
            return lang;        
    }

    String get_seal_or_bindingDesc(Node node, String parentlang, boolean manuscriptView)
    {
        StringBuilder sb = new StringBuilder();
        String lang = getLanguage(node, parentlang);
        NodeList list = node.getChildNodes();
        int len = list.getLength();
        for (int i=0; i<len; i++)
        {
            Node _node = list.item(i);
            if (_node.getNodeType() == Node.TEXT_NODE)          // shouldn't have anything except whitespace
                sb.append(_node.getTextContent());
            else
            {
                String _lang = getLanguage(_node, lang);
                String nodeName = _node.getNodeName();
                if (nodeName.equals("p"))
                    sb.append(getMarkedText(_node, _lang, manuscriptView));
                else if (nodeName.equals("condition"))
                {
                    String condition = getMarkedText(_node, _lang, manuscriptView);
                    if (condition.trim().length() > 0)
                    {
                        newLine(sb);
                        sb.append(String.format("Condition: %s.", condition));
                    }
                }
                else
                    sb.append(getMarkedText(_node, _lang, manuscriptView));
            }
        }
        
        return sb.toString();
    }

    String get_foliation(Node node, String lang, boolean manuscriptView)
    {
        StringBuilder sb = new StringBuilder();
        String _lang = getLanguage(node, lang);
        NodeList list = node.getChildNodes();
        int len = list.getLength();
        String id = getAttributeValue(node, "xml:id");
        if (id == null)
            id = "";
        for (int i=0; i<len; i++)
        {
            Node _node = list.item(i);
            if (_node.getNodeType() == Node.TEXT_NODE)
            {
                String text = _node.getTextContent();
                if (lang != null)
                    text = pi.transcode(text, _lang);
                sb.append(text);
            }
            else if (_node.getNodeName().equals("formula"))
            {
                String formula = getMarkedText(_node, _lang, manuscriptView);
                if (formula.trim().length() > 0)
                {
                    newLine(sb);
                    sb.append(String.format("%s formula: %s", id, formula));
                }
            }
            else
                sb.append(getMarkedText(_node, _lang, manuscriptView));
        }

        return sb.toString();
    }


    String get_locus(Node node, String parentlang)
    {
        String lang = getLanguage(node, parentlang);

        String scheme = getAttributeValue(node, "scheme");

        String locus = get_xtext(node, lang);

        if (scheme != null)
            return String.format("(%s foliation scheme) %s", scheme.substring(1), locus);
        else
            return locus;
    }


    String get_lg(Node node, String parentlang, boolean manuscriptView)
    {
        String result = "";
        StringBuilder sb = new StringBuilder();
        try
        {
            String lang = getLanguage(node, parentlang);

            NodeList list = XML.getRelativeNodeList(node, "l");
        
            int len = list.getLength();
            for (int i=0; i<len; i++)
            {
                Node _node = list.item(i);
                newLine(sb);
                sb.append(getMarkedText(_node, lang, manuscriptView));
            }
        }
        catch (Exception ex)
        {
            displayError(String.format("Error in get_lg: %s", ex.getMessage()));
        }

        return result;
    }


    String get_bibl(Node node, String parentlang, boolean manuscriptView)
    {
        StringBuilder sb = new StringBuilder();
        try
        {
            parentlang = "eng"; // getLanguage(node, parentlang);

            NodeList list = node.getChildNodes();
            int len = list.getLength();
            for (int i=0; i<len; i++)
            {
                Node _node = list.item(i);
                if (_node.getNodeType() == Node.TEXT_NODE)
                    sb.append(_node.getTextContent());
                else
                {
                    String _lang = getLanguage(_node, parentlang);
                    if (node.getNodeName().equals("biblScope"))
                        sb.append(String.format("(%s)", getMarkedText(_node, _lang, manuscriptView)));      // force language to be english
                    else
                        sb.append(getMarkedText(_node, _lang, manuscriptView));
                }
            }
        }
        catch (Exception ex)
        {
            displayError(String.format("Error in get_bibl: %s", ex.getMessage()));
        }

        if (sb.toString().trim().length() > 0)
            return String.format("(%s)", sb);
        else
            return "";
    }


    String get_dimensions(Node node, String lang) throws Exception
    {
        String _lang = getLanguage(node, lang);
        String unit = getAttributeValue(node, "unit");
        if (unit == null)
            unit = "?";
        Node _node = XML.getRelativeNode(node, "height");
        String height = get_xtext(_node, _lang);        // assume that language is the same
        _node = XML.getRelativeNode(node, "width");
        String width = get_xtext(_node, _lang);

        return String.format("%s x %s %s", height, width, unit);
    }


    String get_gap(Node node, String lang, boolean manuscriptView) throws Exception
    {
        StringBuilder sb = new StringBuilder("[");
        String quantity = getAttributeValue(node, "quantity");
        if (quantity != null)
        {
            String unit = getAttributeValue(node, "unit");
            String n = getAttributeValue(node, "n");
            String corresp = getAttributeValue(node, "corresp");
            if (quantity.equals("1"))
                sb.append(String.format("missing 1 %s, ", unit));
            else
            {
                if (unit.toLowerCase().equals("leaf"))
                    sb.append(String.format("missing %s leaves, ", quantity));
                else
                    sb.append(String.format("missing %s %ss, ", quantity, unit));
            }
        }
        else
            sb.append("missing: ");

        Node _node = XML.getRelativeNode(node, "desc");
        if (node != null)
        {
            String _lang = getLanguage(node, lang);
            sb.append(getMarkedText(_node, _lang, manuscriptView));
        }
        else
            sb.append("?");

        sb.append("]");

        return sb.toString();
    }

    String get_xtext(XML xml, String path, String parentlang)
    {
        String value = "";
        try
        {
            Node node = xml.getNode(path);
            value = get_xtext(node, parentlang);
        }
        catch (Exception ex)
        {
            displayError(String.format("Error in get_xtext: %s", ex.getMessage()));
        }

        return value;
    }

    String get_xtext(Node node, String lang)
    {
        boolean unknown = false;
        StringBuilder sb = new StringBuilder();
        try
        {
            if (node != null)
            {
                lang = getLanguage(node, lang);

                NodeList list = node.getChildNodes();
                int len = list.getLength();
                for (int i=0; i<len; i++)
                {
                    node = list.item(i);
                    if (node.getNodeType() == Node.TEXT_NODE)
                    {
                        String text = node.getNodeValue();
                        if (lang != null)
                            text = pi.transcode(text, lang);
                        sb.append(text);
                    }
                    else if (node.getNodeName().equals("ref"))
                    {
                        processRef(node, sb);
                    } else {
                        unknown = true;
                        writeXML(node, sb);         // will presumably be a <g>...</g>
                    }
                }
            }
        }
        catch (Exception ex)
        {
            displayError(String.format("Error in get_xtext: %s", ex.getMessage()));
        }

        if (unknown)
            displayError(String.format("In get_xtext, unexpected element: %s", sb.toString()));
        return sb.toString();
    }


    // Information about second folio and signatures is here
    void witnessDetail(StringBuilder sb, Node node, String label, int level, boolean manuscriptView) throws Exception
    {
        NodeList _list = XML.getRelativeNodeList(node, "witDetail");        // these are like notes
        if (_list != null)
        {
            int _len = _list.getLength();
            for (int j=0; j<_len; j++)
            {
                Node _node = _list.item(j);
                Node secFol = XML.getRelativeNode(_node, "secFol");
                String val = getMarkedText(secFol, getLanguage(secFol, null), manuscriptView);
                writeRow(sb, "Second&#160;folio", val, level);

                Node signatures = XML.getRelativeNode(_node, "signatures");
                val = getMarkedText(signatures, getLanguage(signatures, null), manuscriptView);
                writeRow(sb, "Signatures", val, level);
            }
        }
    }

    void contentSection(StringBuilder sb, Node node, String path, String label, int level, boolean manuscriptView) throws Exception
    {
        NodeList _list = XML.getRelativeNodeList(node, path);
        if (_list != null)
        {
            int _len = _list.getLength();
            for (int j=0; j<_len; j++)
            {
                Node _node = _list.item(j);
                String val = getMarkedText(_node, null, manuscriptView);
                String name = _node.getNodeName();
                String refatt = References.hasInplaceAttribute.get(name);
                if (refatt != null) {
                    Node tn = _node.getAttributes().getNamedItem(refatt);
                    if (tn != null) {
                        String attvalue = tn.getNodeValue();
                        XMLChunk chunk = References.xmlidContent.get(String.format("%s%s", theFileName, attvalue));
                        StringReader sr = new StringReader(chunk.xml);
                        XML xml = new XML(sr);
                        Node refnode = xml.getNode("/").getFirstChild();

                        val += getMarkedText(refnode, chunk.language, manuscriptView);
                    }
                }

                writeRow(sb, label, val, level);
            }
        }
    }


    public void appendBuffer(String s)
    {
        currentBuffer.append(s);
    }


    public String getCharacters()
    {
        int top = languageStack.size() - 1;
        if (top >= 0)
        {
            String language = languageStack.get(languageStack.size()-1);
            String s = currentBuffer.toString();
            currentBuffer.setLength(0);
            return pi.transcode(s, language);
        }
        else
            return "";          // probably an attempt to get characters before the root element (there aren't any).
    }

        // I. Manuscript Identifiers section
    class Handler_msIdentifierStruct implements ElementHandler
    {

        public void process(String path, boolean manuscriptView)
        {
            String val =  null;
            StringBuilder sb = new StringBuilder();

            try
            {
            // Collection
            Node identifierNode = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msIdentifier");
            if (identifierNode == null)
                return;

            Node _node = XML.getRelativeNode(identifierNode, "collection");
            if (_node != null)
            {
                val = get_xtext(_node, null);
                writeRow(sb, "Collection", val);
            }

            // Item
            _node = XML.getRelativeNode(identifierNode, "idno");
            if (_node != null)
            {
                val = get_xtext(_node, null);
                writeRow(sb, "Item", val);
            }
            // Repository
            _node = XML.getRelativeNode(identifierNode, "repository");
            if (_node != null)
            {
                val = get_xtext(_node, null);
                writeRow(sb, "Repository", val);
            }

            // Institution
            _node = XML.getRelativeNode(identifierNode, "institution");
            if (_node != null)
            {
                val = get_xtext(_node, null);
                writeRow(sb, "Institution", val);
            }

            // Location
            String settlement = "?";
            String region = "?";
            String country = "?";
            _node = XML.getRelativeNode(identifierNode, "settlement");
            if (_node != null)
                settlement = get_xtext(_node, null);
            _node = XML.getRelativeNode(identifierNode, "region");
            if (_node != null)
                region = get_xtext(_node, null);
            _node = XML.getRelativeNode(identifierNode, "country");
            if (_node != null)
                country = get_xtext(_node, null);

            if (!settlement.equals("?") || !region.equals("?") || !country.equals("?"))
            {
                val = String.format("%s, %s, %s", settlement, region, country);
                writeRow(sb, "Location", val);
            }

            // Catalog and Item
            NodeList list = XML.getRelativeNodeList(identifierNode, "altIdentifier[@type='catalog']");
            if (list != null)
            {
                int len = list.getLength();
                for (int i=0; i<len; i++)
                {
                    _node = list.item(i);
                    Node nodeCatalog = XML.getRelativeNode(_node, "collection");
                    if (nodeCatalog != null)
                    {
                        val = get_xtext(nodeCatalog, null);
                        writeRow(sb, "Catalog", val);
                    }

                    Node nodeIdno = XML.getRelativeNode(_node, "idno");
                    if (nodeIdno != null)
                    {
                        val = get_xtext(nodeIdno, null);
                        writeRow(sb, "Item", val);
                    }
                }
            }

            if (sb.length() > 0)
            {
                manuscriptIdentifiers.append("<h3>Manuscript Identifiers</h3>\n");
                manuscriptIdentifiers.append("<table>\n");
                manuscriptIdentifiers.append(sb.toString());
                manuscriptIdentifiers.append("</table>\n");
            }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_msIdentifierStruct: %s", ex.getMessage()));
            }
        }
    }



    // Contents section
    class Handler_msContents implements ElementHandler
    {
        public void process(String path, boolean manuscriptView)
        {
            int workNumber = 1;
            StringBuilder sb = new StringBuilder();
            try
            {
                Node contentNode = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.msContents");
                NodeList list = XML.getRelativeNodeList(contentNode, "msItemStruct");
                if (list != null)
                {
                    int len = list.getLength();
                    for (int i=0; i<len; i++)
                    {
                        Node node = list.item(i);
                        processItemStruct(node, sb, 0, "", workNumber++, manuscriptView);
                    }
                }

                if (sb.length() > 0)
                {
                    contents.append("<h3>Contents</h3>\n");
                    contents.append(sb.toString());
                }


            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_msContents.process", ex.getMessage()));
            }
        }

        private void processItemStruct(Node node, StringBuilder sb, int levelno, String prefix, int workNumber, boolean manuscriptView)
        {
            try
            {
                if (node != null)
                {
                    sb.append("<table>\n");
                    String val = getAttributeValue(node, "defective");
                    boolean complete = false;
                    if (val != null)
                        complete = val.toLowerCase().equals("false");

                    // Work number
                    sb.append(String.format(String.format("<h4>%sWork&#160;%s</h4>\n", getHTMLIndentation(levelno), String.format("%s%d", prefix, workNumber))));

                    // Locus
                    Node _node = XML.getRelativeNode(node, "locus");
                    if (_node != null)
                    {
                        val = get_locus(_node, "en");
                        writeRow(sb, "Locus", String.format("%s (%s)",val, complete ? "complete" : "incomplete"), levelno);
                    }

                    // Author
                    _node = XML.getRelativeNode(node, "author");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, manuscriptView);
                        if (val.trim().length() == 0)
                        {
                            _node = XML.getRelativeNode(node, "author/persName");
                            if (_node != null)
                                val = get_xtext(_node, null);
                        }
                        writeRow(sb, "Author", val, levelno);
                    }

                    // Title
                    _node = XML.getRelativeNode(node, "title[@type='main']");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, manuscriptView);
                        writeRow(sb, "Title", val, levelno);
                    }

                    // Part
                    _node = XML.getRelativeNode(node, "title[@type='part']");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, manuscriptView);
                        writeRow(sb, "Part", val, levelno);
                    }

                    // Cover Rubric
                    contentSection(sb, node, "rubric[@type='cover']", "Cover&#160;rubric", levelno, manuscriptView);

                    // Incipit
                    contentSection(sb, node, "incipit", "Incipit", levelno, manuscriptView);

                    // Explicit
                    contentSection(sb, node, "explicit", "Explicit", levelno, manuscriptView);

                    // Final rubric
                    contentSection(sb, node, "finalRubric", "Final&#160;rubric", levelno, manuscriptView);

                    // Colophon
                    if (!nodeExists(node, "colophon")) {
                        writeRow(sb, "Colophon", "None", levelno);
                    } else {
                        contentSection(sb, node, "colophon", "Colophon", levelno, manuscriptView);
                    }

                    // Note
                    contentSection(sb, node, "note", "Note", levelno, manuscriptView);

                    // Filiation
                    contentSection(sb, node, "filiation", "Filiation", levelno, manuscriptView);

                    // Second folio
                    witnessDetail(sb, node, "Second folio", levelno, manuscriptView);

                    // Language
                    _node = XML.getRelativeNode(node, "textLang");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, manuscriptView);
                        writeRow(sb, "Language", val, levelno);
                    }
                    sb.append("</table>\n");

                    // What to do with witDetail and filiation
                    NodeList list = XML.getRelativeNodeList(node, "msItemStruct");
                    int len = list.getLength();
                    for (int i=0; i<len; i++)
                    {
                        _node = list.item(i);
                        processItemStruct(_node, sb, levelno+1, String.format("%s%d.", prefix, workNumber), i+1, manuscriptView);
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Handler_msContents failed with: %s", ex.getMessage()));
            }
        }
    }



    // Physical description section
    class Handler_physDesc implements ElementHandler
    {
        public void process(String path, boolean manuscriptView)
        {
            try
            {
                StringBuilder sb = new StringBuilder();

                Node objectDescNode = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc");

                if (objectDescNode != null)
                {
                    // Form
                    String val = getAttributeValue(objectDescNode, "form");
                    if (val != null)
                        writeRow(sb, "Form", val);

                    // Material
                    Node _node = XML.getRelativeNode(objectDescNode, "supportDesc/support/material");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, manuscriptView);
                        writeRow(sb, "Material", val);
                    }

                    // Watermark
                    _node = XML.getRelativeNode(objectDescNode, "supportDesc/support/watermark");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, manuscriptView);
                        writeRow(sb, "Watermark", val);
                    }

                    // Extent
                    _node = XML.getRelativeNode(objectDescNode, "supportDesc/extent/measure");
                    if (_node != null)
                    {
                        val = getAttributeValue(_node, "quantity");
                        if (val !=  null)
                            writeRow(sb, "Extent", val);
                    }

                    // Dimensions
                    _node = XML.getRelativeNode(objectDescNode, "supportDesc/extent/dimensions");
                    if (_node != null)
                    {
                        val = get_dimensions(_node, null);
                        writeRow(sb, "Dimension", val);
                    }

                    // Foliation
                    NodeList list = XML.getRelativeNodeList(objectDescNode, "supportDesc/foliation");
                    if (list != null)
                    {
                        int len = list.getLength();
                        for (int i=0; i<len; i++)
                        {
                            _node = list.item(i);
                            val = get_foliation(_node, null, manuscriptView);
                            writeRow(sb, "Foliation", val);
                        }
                    }

                    // Collation
                    _node = XML.getRelativeNode(objectDescNode, "supportDesc/collation");
                    if (_node != null)
                    {
                        val = get_foliation(_node, null, manuscriptView);           // same as foliation?
                        writeRow(sb, "Collation", val);
                    }

                    // Condition
                    _node = XML.getRelativeNode(objectDescNode, "supportDesc/condition");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, manuscriptView);           // same as foliation?
                        writeRow(sb, "Condition", val);
                    }

                    // Binding
                    Node bindingDescNode = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.bindingDesc");
                    if (bindingDescNode != null)
                    {
                        list = XML.getRelativeNodeList(bindingDescNode, "binding");
                        if (list != null)
                        {
                            int len = list.getLength();
                            for (int i=0; i<len; i++)
                            {
                                _node = list.item(i);
                                val = get_seal_or_bindingDesc(_node, null, manuscriptView);
                                writeRow(sb, "Binding", val);
                            }
                        }
                    }

                    // Seal
                    Node sealDescNode = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.sealDesc");
                    if (sealDescNode != null)
                    {
                        list = XML.getRelativeNodeList(bindingDescNode, "seal");
                        if (list != null)
                        {
                            int len = list.getLength();
                            for (int i=0; i<len; i++)
                            {
                                _node = list.item(i);
                                val = get_seal_or_bindingDesc(_node, null, manuscriptView);
                                writeRow(sb, "Seal", val);
                            }
                        }
                    }


                    // Accompanying matter
                    Node accMatNode = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.accMat");
                    if (accMatNode != null)
                    {
                        val = getMarkedText(accMatNode, getLanguage(accMatNode, null), manuscriptView);
                        writeRow(sb, "Accompanying material", val);
                    }

                    // Layout
                    list = XML.getRelativeNodeList(objectDescNode, "layoutDesc/layout");
                    if (list != null)
                    {
                        int len = list.getLength();
                        for (int i=0; i<len; i++)
                        {
                            _node = list.item(i);
                            val = getMarkedText(_node, null, manuscriptView);           // same as foliation?
                            writeRow(sb, "Layout", val);
                        }
                    }
                }
                if (sb.length() > 0)
                {
                    physicalDescription.append("<h3>Physical description</h3>\n");
                    physicalDescription.append("<table>\n");
                    physicalDescription.append(sb.toString());
                    physicalDescription.append("</table>\n");
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_physDesc: %s", ex.getMessage()));
            }
        }
    }


    // Hands section
    class Handler_handDesc implements ElementHandler
    {

        public void process(String path, boolean manuscriptView)
        {
            try
            {
                String val = null;
                StringBuilder sb = new StringBuilder();
                Node node = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc");
                if (node != null)
                {
                    Node _node = XML.getRelativeNode(node, "summary");
                    if (_node != null)
                    {
                        val = getMarkedText(_node, null, manuscriptView);
                        writeRow(sb, "Summary", val);
                    }
                    NodeList list = XML.getRelativeNodeList(node, "handNote");
                    if (list != null)
                    {
                        int len = list.getLength();
                        for (int i=0; i<len; i++)
                        {
                            _node = list.item(i);
                            String _lang = getLanguage(_node, null);
                            val = getMarkedText(_node, _lang, manuscriptView);
                            writeRow(sb, String.format("Hand %d", i+1), val);
                        }
                    }

                    node = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.additions");
                    if (node != null)
                    {
                        val = getMarkedText(node, null, manuscriptView);
                        writeRow(sb, "Additions", val);
                    }

                    if (sb.length() > 0)
                    {
                        hands.append("<h3>Hands</h3>\n");
                        hands.append("<table>\n");
                        hands.append(sb.toString());
                        hands.append("</table>");
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_handDesc: %s", ex.getMessage()));
            }
        }

    }


    // Decoration section
    class Handler_decoDesc implements ElementHandler
    {
        public void process(String path, boolean manuscriptView)
        {
            try
            {
                String val = null;
                StringBuilder sb = new StringBuilder();

                Node node = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='color']");
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), manuscriptView);
                    writeRow(sb, "Color", val);
                }

                node = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='border']");
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), manuscriptView);
                    writeRow(sb, "Border", val);
                }

                node = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.decoDesc.decoNote[type='illustration']");
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), manuscriptView);
                    writeRow(sb, "Illustration", val);
                }

                if (sb.length() > 0)
                {
                    decoration.append("<h3>Decoration</h3>\n");

                    decoration.append("<table>\n");

                    decoration.append(sb.toString());

                    decoration.append("</table>");
                }
                
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_decoDesc", ex.getMessage()));
            }
        }
    }


    // History section
    class Handler_history implements ElementHandler
    {
        public void process(String path, boolean manuscriptView)
        {
            try
            {
                String val = null;

                StringBuilder sb = new StringBuilder();

                Node node = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.origin");
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), manuscriptView);
                    writeRow(sb, "Origin", val);
                }

                node = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.provenance");
                if (node != null)
                {
                    val = getMarkedText(node, getLanguage(node, null), manuscriptView);
                    writeRow(sb, "Provenance", val);
                }

                node = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.acquisition");
                if (node != null)
                {
                    biblBreak = false;          // this is assuming no nested acquisitions
                    val = getMarkedText(node, getLanguage(node, null), manuscriptView);
                    biblBreak = true;
                    writeRow(sb, "Acquisition", val);
                }

                if (sb.length() > 0)
                {
                    history.append("<h3>History</h3>\n");

                    history.append("<table>\n");
                    history.append(sb.toString());
                    history.append("</table>");
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_history: %s", ex.getMessage()));
            }
        }
    }


    // Subject headings section
    class Handler_profileDesc implements ElementHandler
    {
        public void process(String path, boolean manuscriptView)
        {
            try
            {
                Node _node = null;
                String val = null;
                StringBuilder sb = new StringBuilder();
                Node node = pi.getXML("TEI.teiHeader.profileDesc");
                if (node != null)
                {
                    NodeList list = XML.getRelativeNodeList(node, "textClass/keywords[@scheme='#LCSH']/list/item");
                    if (list != null)
                    {
                        val = null;
                        _node = null;
                        int len = list.getLength();
                        for (int i=0; i<len; i++)
                        {
                            _node = list.item(i);
                            val = getMarkedText(_node, getLanguage(_node, null), manuscriptView);
                            writeRow(sb, "SubjectLC", val);
                        }
                    }

                    _node = XML.getRelativeNode(node, "textClass/catRef");
                    if (_node != null)
                    {
                        String target = getAttributeValue(_node, "target");
                        if (target != null)
                        {
                            Regex re = new Regex(target, "#(.*)$");
                            if (re.find())
                            {
                                target = re.getGroup(1);
                                target = classifications.get(target);
                            }
                            else
                                target = null;
                        }
                        if (target == null)
                            target = "";

                        String scheme = getAttributeValue(_node, "scheme");
                        if (scheme != null)
                            scheme = String.format("(%s)", scheme);
                        else
                            scheme = "";

                        if (target.length() > 0 || scheme.length() > 0)
                        {
                            val = String.format("%s %s", target, scheme);
                            writeRow(sb, "SubjectSL", val);
                        }
                    }
                }

                if (sb.length() > 0)
                {
                    subjectHeadings.append("<h3>Subject headings</h3>\n");
                    subjectHeadings.append("<table>\n");
                    subjectHeadings.append(sb.toString());
                    subjectHeadings.append("</table>");
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_profileDesc: %s", ex.getMessage()));
            }
        }
    }


    // Facsimile section
    class Handler_facsimile implements ElementHandler
    {
        public void process(String path, boolean manuscriptView)
        {
            try
            {
                StringBuilder sb = new StringBuilder();
                Node node = pi.getXML("TEI.facsimile");
                if (node != null)
                {
                    NodeList list = XML.getRelativeNodeList(node, "graphic");
                    int len = list.getLength();
                    for (int i=0; i<len; i++)
                    {
                        Node _node = list.item(i);
                        String url = getAttributeValue(_node, "url");
                        if (url != null)
                        {
                            newLine(sb);
                            sb.append(String.format("<a href='%s'>%s</a>\n", url, url));
                        }
                    }
                    if (sb.toString().trim().length() > 0)
                    {
                        facsimile.append("<h3>Facsimile</h3>\n");
                        facsimile.append(sb.toString());
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_facsimile: %s", ex.getMessage()));
            }
        }
    }

    // Transcription section
    class Handler_transcription implements ElementHandler
    {
        public void process(String path, boolean manuscriptView)
        {
            try
            {
                StringBuilder sb = new StringBuilder();
                Node node = pi.getXML("TEI.text");
                if (node != null)
                {
                    String lang = "sa-Latn-x-KH";
                    if (node.hasAttributes())       // override KH if xml:lang specified
                    {
                        Node attlang = node.getAttributes().getNamedItem("xml:lang");
                        if (attlang != null)
                            lang = attlang.getNodeValue();
                    }
                    sb.append(getMarkedText(node, lang, manuscriptView));
                    if (sb.toString().trim().length() > 0)
                    {
                        transcription.append("<h3>Transcription</h3>\n");
                        transcription.append(sb.toString());
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_transcription: %s", ex.getMessage()));
            }
        }
    }


    // Last revision
    class Handler_recordHist implements ElementHandler
    {
        public void process(String path, boolean manuscriptView)
        {
            try
            {
                Node node = pi.getXML("TEI.teiHeader.fileDesc.sourceDesc.msDesc.additional.adminInfo.recordHist");
                if (node != null)
                {
                    NodeList list = XML.getRelativeNodeList(node, "change");
                    if (list != null)
                    {
                        int len = list.getLength();
                        if (len > 0)
                        {
                            Node _node = list.item(len - 1);
                            _node = XML.getRelativeNode(_node, "date");
                            if (_node != null)
                            {
                                changes.append("<table>");
                                writeRowHeader(changes, "Record revised", _node.getTextContent());
                                changes.append("</table>");
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                displayError(String.format("Error in Handler_recordHist: %s", ex.getMessage()));
            }
        }

    }

}
