package siva;

import java.io.BufferedWriter;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.FSM;
import org.sanskritlibrary.ResourceInputStream;
import org.sanskritlibrary.TextFormatter;
import org.sanskritlibrary.XML;
import org.xml.sax.Attributes;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;
import org.w3c.dom.Node;

public class PathInfo
{
    static private Hashtable<String, Integer> allpaths = new Hashtable<String, Integer>();
    static private Hashtable<String, AttributeValues> attvalues = new Hashtable<String, AttributeValues>();
    static private Hashtable<String, Boolean>infile = new Hashtable<String, Boolean>();
    static private Hashtable<String, Boolean>emptyElements = new Hashtable<String, Boolean>();
    static Hashtable<String, TagPaths> tagnames = new Hashtable<String, TagPaths>();

    ResourceInputStream ris;
    boolean ofInterest = false;
    StringBuilder currentKey = new StringBuilder();
    StringBuilder currentPath = new StringBuilder();
    StringBuilder currentBuffer = null;
    String containerPath = null;
    Hashtable<String, ElementInfo> paths = new Hashtable<String, ElementInfo>();
    Hashtable<String, String> keys = new Hashtable<String, String>();
    Hashtable<String, Boolean> containers = new Hashtable<String, Boolean>();
    Hashtable<String, ElementHandler> handlers = new Hashtable<String, ElementHandler>();
    int debug = 0;
    Hashtable<String, FSM[]> fsmh = null;

    public void setResourceInputStream(ResourceInputStream ris)
    {
        this.ris = ris;
    }

    public void setFSMH(Hashtable<String, FSM[]> fsmh)
    {
        this.fsmh = fsmh;
    }
    

    public void put(String path)
    {
        paths.put(path, new ElementInfo());
    }

    public void key(String path, String att)
    {
        keys.put(path, att);
    }

    public void container(String path)
    {
        if (paths.get(path) == null)
            paths.put(path, new ElementInfo());
        containers.put(path, true);
    }

    public void handler(String path, ElementHandler eh)
    {
        handlers.put(path, eh);
    }


    public void pushElement(String qName, Attributes atts)
    {
        String delim = "";
        if (currentKey.length() > 0)
            delim = ".";

        currentKey.append(String.format("%s%s", delim, qName));         // used to look up values in keys, containers and handlers hash tables.
        if (currentKey.toString().equals("TEI.text.body.div"))
            debug = debug++ - 1;
        String key = "";
        String attkey = keys.get(currentKey.toString());
        if (attkey != null)                     // need to append the value of the key attribute to the path
        {
            String value = atts.getValue(attkey);

            if (value != null)
                key = String.format("[%s='%s']", attkey, value);
            else
                key = String.format("[%s='*']", attkey);        // key attribute is missing, use '*'
        }
        else
            key = "";

        currentPath.append(String.format("%s%s%s", delim, qName, key));

        if (containerPath != null)
        {
            if (paths.get(currentPath.toString()) != null)
                System.out.format("The path %s is inside a designated container (%s)\n", currentPath, containerPath.toString());
            copyTag(qName, atts);
        }
        else if (isOfInterest())
        {
            ofInterest = true;
            containerPath = currentPath.toString();
            currentBuffer = paths.get(currentPath.toString()).buffer;   // this buffer will be used for all children
            pushAttributes(atts);           // in case any attributes are also of interest
            copyTag(qName, atts);
        }
    }

    public void append(String s)
    {
        if (containerPath != null)
        {
            currentBuffer.append(fixup(s));
        }
    }


    private String fixup(String s)
    {
        if (s.contains("&"))
            s = s.replaceAll("&", "&amp;");

        if (s.contains("<"))
            s = s.replaceAll("<", "&lt;");

        if (s.contains(">"))
            s = s.replaceAll(">", "&gt;");

        return s;
    }


    public void popElement(String qName)
    {
        // Do post processing, if any

        if (currentKey.toString().equals("TEI.teiHeader.fileDesc.sourceDesc.msDesc.history.origin.p"))
            debug = debug++ - 1;

        if (containerPath != null)          // are we currently in a container?
            copyEndTag(qName);

        ElementHandler eh = handlers.get(currentPath.toString());
        if (eh != null)
            eh.process(currentPath.toString(), true);         // note this may not be an item of interest but is the end tag of an element that contains elements of interest.

        if (containerPath != null && containerPath.equals(currentPath.toString()))       // is this the end of the container?
            containerPath = null;

        // Pop the last tag off of the current key
        int n = currentKey.lastIndexOf(".");
        if (n != -1)
            currentKey.setLength(n);
        else
            currentKey.setLength(0);

        // Pop the last tag off of the current path
        n = currentPath.lastIndexOf(".");
        if (n != -1)
            currentPath.setLength(n);
        else
            currentPath.setLength(0);

        ofInterest = isOfInterest();
        if (ofInterest)
            currentBuffer = paths.get(currentPath.toString()).buffer;
        else if (containerPath == null)
            currentBuffer = null;
    }


    public String getLastTag()
    {
        int idx = currentKey.lastIndexOf(".");
        if (idx == -1)
            return currentKey.toString();
        else
            return currentKey.substring(idx+1);
    }

    public void rememberFile(String fileName)
    {
        infile.put(fileName, true);
    }

    public void rememberEmptyElement(String element)
    {
        emptyElements.put(element, true);
    }

    public void recordStatistics(String qName, String path, Attributes atts, String fileName)
    {
        Integer cnt = allpaths.get(path);
        if (cnt == null)
            cnt = new Integer(0);
        allpaths.put(path, new Integer(cnt.intValue() + 1));

        TagPaths tp = tagnames.get(qName);
        if (tp == null)
        {
            tp = new TagPaths(qName);
            tagnames.put(qName, tp);
        }
        tp.count++;
        if (!tp.paths.contains(path))
            tp.paths.add(path);

        int len = atts.getLength();
        for (int i=0; i<len; i++)
        {
            String attname = atts.getQName(i);
            AttributeValues av = attvalues.get(attname);
            if (av ==  null)
            {
                av = new AttributeValues(attname);
                attvalues.put(attname, av);
            }
            av.count++;
            String value = atts.getValue(i);
            ValueInfo vi = av.values.get(value);
            if (vi == null)
            {
                vi = new ValueInfo(value, fileName);
                av.values.put(value, vi);
            }
            vi.count++;
        }
    }


    static public void writeStatistics(BufferedWriter bw) throws Exception
    {
        writeFileNames(bw);
        writeAllPaths(bw);
        writeTagNames(bw);
        writeAttributeValues(bw);
    }


    static public void writeEmptyElements(BufferedWriter bw) throws Exception
    {
        Enumeration<String> en = emptyElements.keys();
        ArrayList<String> a = new ArrayList<String>();
        while(en.hasMoreElements())
        {
            a.add(en.nextElement());
        }

        Collections.sort(a);

        bw.write("\n\n**** In Files\n");

        for (String key : a)
        {
            bw.write(String.format("%s\n", key));
        }
    }

    static private void writeFileNames(BufferedWriter bw) throws Exception
    {
        Enumeration<String> en = infile.keys();
        ArrayList<String> a = new ArrayList<String>();
        while(en.hasMoreElements())
        {
            a.add(en.nextElement());
        }

        Collections.sort(a);

        bw.write("\n\n**** In Files\n");

        for (String key : a)
        {
            bw.write(String.format("%s\n", key));
        }
    }

    static private void writeAllPaths(BufferedWriter bw) throws Exception
    {
        bw.write("\n\n**** All Paths\n");
        Enumeration<String> en = allpaths.keys();
        ArrayList<PathCount> a = new ArrayList<PathCount>();
        while (en.hasMoreElements())
        {
            String path = en.nextElement();
            Integer cnt = allpaths.get(path);
            a.add(new PathCount(path, cnt));
        }

        Collections.sort(a, new PathCount());

        for (PathCount tc : a)
        {
            bw.write(String.format("%s (%d)\n", tc.path, tc.count));
        }        
    }
    
    static private void writeTagNames(BufferedWriter bw) throws Exception
    {
        bw.write("\n\n**** All Tag Names\n");

        Enumeration<String> en = tagnames.keys();
        ArrayList<TagPaths> a = new ArrayList<TagPaths>();
        while (en.hasMoreElements())
        {
            String tagname = en.nextElement();
            a.add(tagnames.get(tagname));
        }

        Collections.sort(a, new TagPaths());

        for (TagPaths tp : a)
        {
            bw.write(String.format("%s (%d)\n", tp.tagName, tp.count));
            Collections.sort(tp.paths);
            for (String path : tp.paths)
            {
                bw.write(String.format("\t%s\n", path));
            }
        }                
    }
    
    static private void writeAttributeValues(BufferedWriter bw) throws Exception
    {
        bw.write("\n\n**** All Attribute Names/Values\n");

        Enumeration<String> en = attvalues.keys();
        ArrayList<AttributeValues> a = new ArrayList<AttributeValues>();
        while (en.hasMoreElements())
        {
            String key = en.nextElement();
            AttributeValues av = attvalues.get(key);
            a.add(av);
        }

        Collections.sort(a, new AttributeValues());

        for (AttributeValues av : a)
        {
            bw.write(String.format("%s (%d)\n", av.name, av.count));

            int i = 0;
            en = av.values.keys();
            while (en.hasMoreElements())
            {
                ValueInfo vi = av.values.get(en.nextElement());
                bw.write(String.format("\t%s (%d) @%s\n", vi.value, vi.count, vi.firstReference));
            }
            bw.write("\n");
        }                
    }

    private void copyTag(String qName, Attributes atts)
    {
        currentBuffer.append(String.format("<%s", qName));
        int len = atts.getLength();
        for (int i=0; i<len; i++)
        {
            currentBuffer.append(String.format(" %s='%s'", atts.getQName(i), atts.getValue(i)));
        }
        currentBuffer.append(">");
    }

    private void copyEndTag(String qName)
    {
        currentBuffer.append(String.format("</%s>", qName));
    }

    private boolean isOfInterest()
    {
        return paths.get(currentPath.toString()) != null;
    }

    // Save each attribute of interest in the buffer
    private void pushAttributes(Attributes atts)
    {
        int len = atts.getLength();
        int pathlen = currentPath.length();
        for (int i=0; i<len; i++)
        {
            currentPath.append(String.format(":%s", atts.getQName(i)));
            ElementInfo ei = paths.get(currentPath.toString());
            if (ei != null)                  // is this an attribute of interest?
                ei.buffer.append(atts.getValue(i));
            currentPath.setLength(pathlen);
        }
    }

    public String getTrimmedAttributeValue(String path)
    {
        ElementInfo ei = paths.get(path);
        if (ei == null)
            return String.format("%s not defined", path);

        String value = ei.buffer.toString();
        ElementInfo temp = paths.get(path + "!xml:lang");
        if (temp != null)
        {
            String lang = temp.buffer.toString();
            value = transcode(value, lang);
        }
        ei.buffer.setLength(0);

        return value;
    }


    public Node getXML(String path) throws Exception
    {
        Node node = null;

        ElementInfo ei = paths.get(path);
        if (ei != null)
        {
            if (ei.buffer.toString().trim().length() > 0)
            {
                StringReader sr = new StringReader(ei.buffer.toString());
                XML xml = new XML(sr);
                node = xml.getNode("/").getFirstChild();
            }
        }

        return node;
    }


    public String transcode(String value, String lang)
    {
        try
        {
            String map = null;
            if (lang.endsWith("SLP1"))
                map = String.format("*:slp1->%s", ParseTEI.target);
            else if (lang.endsWith("KH"))
                map = String.format("*:hk->%s", ParseTEI.target);
            if (map != null)
            {
                TransformMap tm = new TransformMap(map, ris, fsmh);
                value = WebServices.transformString(value, tm);
                value = TextFormatter.charToEntity(value);
            }
        }
        catch (Exception ex)
        {
            System.out.format("TransformMap failed with %s\n", ex.getMessage());
        }

        return value;
    }

    public void clear(String path)
    {
        ElementInfo ei = paths.get(path);
        if (ei != null)
            ei.buffer.setLength(0);
    }


    public void dump(String path)
    {
        System.out.print(path);
        System.out.print(": ");
        ElementInfo ei = paths.get(path);
        if (ei == null)
            System.out.println("null");
        else if (ei.buffer.length() == 0)
            System.out.println("none");
        else
           System.out.println(ei.buffer);
    }
}